#############################################################################
# utils.tcl Copyright 2004, 2005 r0lZ
# Procedures related to the Macros and Info menus.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

namespace eval ::macros {
namespace export macros_menu info_menu
namespace import ::readdvd::*

#############################################################################
## macros_menu
## build the Macros menu
proc macros_menu {menuname} {
	menu $menuname
	$menuname add command -label "Kill PGC Playback" -command {::macros::kill_pgc_playback} -accelerator "Ctrl+K"
	bind . <Control-k> ::macros::kill_pgc_playback
	$menuname add command -label "Enable all operations" -command {::macros::enable_all_operations}
	$menuname add command -label "Reset Used GPRMs to 0" -command {::macros::reset_used_gprms}
	$menuname add separator
	$menuname add command -label "Import First-Play (startup) Clip" -command {::macros::import_fp_clip}
	$menuname add command -label "Import Title's Intro Clip" -command {::macros::import_title_intro_clip}
	$menuname add command -label "Import Title's Closing Clip" -command {::macros::import_title_closing_clip}
	$menuname add separator
	$menuname add command -label "DVDShrink Remote Buttons" -command {::macros::title_but_next_title}
}

#############################################################################
## info_menu
## build the Info menu
proc info_menu {menuname} {
	menu $menuname
	$menuname add command -label "PGC" -command {::macros::show_current_pgc_as_text}
	$menuname add separator
	$menuname add command -label "Title Play Map table" -command ::srptinfo::srptinfo_dump
	$menuname add separator
	$menuname add command -label "DVD commands" -command ::main::dump_DVD
	$menuname add command -label "PGCs params and Cell Tables" -command ::pgced::dump_all_pgcs
	if {$::tcl_platform(platform) != "macintosh"} {
		$menuname add command -label "Menu Buttons commands" -command {::menubuttons::show_menu_buttons_info}
	}
	$menuname add command -label "Go to calling command" -command {::utils::pgc_crossrefs} -accelerator "Ctrl+G"
	bind . <Control-g> ::utils::pgc_crossrefs
	$menuname add command -label "Calls cross references" -command {::utils::show_crossrefs} -accelerator "Ctrl+H"
	bind . <Control-h> ::utils::show_crossrefs
	$menuname add command -label "Streams" -command ::streams::get_streams_info
	$menuname add separator
	$menuname add command -label "Find Unused GPRMs" -command {::macros::find_unused_gprms false}
}


#############################################################################
## Import VTS Title and automatically play it when the DVD play for the first time (PF-PGC)
proc import_fp_clip {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set utiltitle "Import first-play clip"

	# find a free GPRM
	set freegprm [lindex [find_unused_gprms true] end]
	if {$freegprm == ""} {
		set txt "This macro need an unused GPRM to work."
		set txt "$txt\nThere are no remaining free GPRMs!\nSorry."
		tk_messageBox -title $utiltitle -icon warning -parent . -message $txt
		return
	}
	set freegprm [format "%02X" $freegprm]

	### GUI
	if {![info exists ::macros::startupclip_stilltime]} {
		set ::macros::startupclip_stilltime -1		;# still time value: 0, # secs, or 255
		set ::macros::startupclip_stilltime_secs 10	;# used only in GUI
	}

	if [winfo exists .importclip] {destroy .importclip}
	set w .importclip
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w $utiltitle
	wm transient $w .
	wm protocol $w WM_DELETE_WINDOW {set ::macros::startupclip_done 0}
	bind $w <Escape> "set ::macros::startupclip_done 0"

	set txt "This macro will import a new titleset title,"
	set txt "$txt\nand modify the DVD to let the imported clip"
	set txt "$txt\nautomatically play when the DVD is starting."
	label $w.lab -text $txt
	pack  $w.lab -padx 2

	frame $w.f1 -bd 2 -relief groove
	pack  $w.f1 -padx 4 -pady 2 -fill x
	set txt "You will be prompted for the titleset to import when you select"
	set txt "$txt\nthe Continue button.  Select a VTS with only ONE title!"
	label $w.f1.lab -text $txt -anchor w -justify left
	pack  $w.f1.lab -side left

	frame $w.f2 -bd 2 -relief groove
	pack  $w.f2 -padx 4 -pady 2 -fill x
	set txt "If the imported clip is a still frame,"
	set txt "$txt\nplease select his Still Time (pause duration)"
	label $w.f2.lab -text $txt -anchor w -justify left
	pack  $w.f2.lab -side top -anchor w

	frame $w.f2.f
	pack  $w.f2.f -fill x
	radiobutton $w.f2.f.b2 -text "Display still frame" \
			-variable ::macros::startupclip_stilltime -value -1 -anchor w
	pack  $w.f2.f.b2 -side left
	entry $w.f2.f.e -textvariable ::macros::startupclip_stilltime_secs -width 4 \
			-validate all -validatecommand "::ced::value_entry_validate 0 254 {%d} {%P}"
	pack  $w.f2.f.e -side left
	label $w.f2.f.l -text "seconds (max 254)"
	pack  $w.f2.f.l -side left
	radiobutton $w.f2.b1 -text "Don't pause: it's an animated clip" \
			-variable ::macros::startupclip_stilltime -value 0 -anchor w
	pack  $w.f2.b1 -fill x

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -anchor e
	button $w.fb.ok -text "Continue" -width 10 -command {set ::macros::startupclip_done 1}
	pack   $w.fb.ok -side left -padx 2
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::macros::startupclip_done 0}
	pack   $w.fb.cancel -side left -padx 2

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr [winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)]
	set wy [expr [winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	wm geometry $w +$wx+$wy
	wm deiconify $w
	grab set $w

	tkwait variable ::macros::startupclip_done
	grab release $w
	destroy $w
	if {!$::macros::startupclip_done} return

	set startupclip_stilltime $::macros::startupclip_stilltime
	if {$startupclip_stilltime == -1} {
		set startupclip_stilltime $::macros::startupclip_stilltime_secs
		if {$startupclip_stilltime == ""} {set startupclip_stilltime 0}
	}

	#################### here we go ##################
	show_msg "Importing Clip to be automatically played..."

	### import the new VTS with the intro clip
	set startupclip_title_num [::utils::import_VTST]
	if {$startupclip_title_num == 0} return

	show_msg "Modifying DVD commands..."

	### Modify the FP-PGC pre commands to jump to our newly created VMGM menu
	### get the original FP-PGC pre cmds
	set cmds $::pgcs(0,0)
	set idx2 [expr [lsearch $cmds {* post}] - 1]
	# we are adding 4 commands, so increment the GOTOs by 4 in the original commands
	for {set i 1} {$i <= $idx2} {incr i} {
		set cmd [lindex $cmds $i]
		set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
		if {$cmdchk == "001" || $cmdchk == "003"} {
			set ln [expr 0X[string range $cmd 21 22] + 4]
			lset cmds $i "[string range $cmd 0 19] [format %02X $ln]"
		}
	}
	set ::pgcs(0,0) [linsert $cmds 1 \
			"00 A1 00 $freegprm 00 01 00 04" "71 00 00 $freegprm 00 01 00 00" \
			"30 02 00 00 00 [format "%02X" $startupclip_title_num] 00 00" \
			 {00 00 00 00 00 00 00 00}]
	 ;# If <free gprm> == 1 Goto 4; Set <free GPRM> = 1; Jump to title <new_title>; NOP

	### Modify the new Clip's post commands to jump back to the FP-PGC
	set ::pgcs($::pgcs(numvts),1) {{* pre} {* post} {30 08 00 00 01 00 00 00} {* cell}}

	### Allow to use the Next Chapter button to skip to the end of the imported clip
	set_int ::pgcs($::pgcs(numvts),1,header) 0x9C 1 2

	### create a new Root menu for the imported title that jump back to the FP-PGC
	set cmds {"* pre" "30 06 00 00 00 00 00 00" "* post" "* cell"} ;# JumpSS FP
	::utils::new_root_menu $::pgcs(numvts) true
	set ::menus($::pgcs(numvts),1,1) $cmds

	### Set the imported clip still time
	set_int ::pgcs($::pgcs(numvts),1,cellpbtable) \
			[expr (($::pgcs($::pgcs(numvts),1,numcells) - 1) * 24) + 2] \
			$startupclip_stilltime 1

	### refresh interface
	::main::init_tree_menu
	::main::display_pgc pgc 0 0

	show_msg "Clip successfully appended as Title $startupclip_title_num."
	return
}

#############################################################################
## Import VTS Title and automatically play it before current title playback
proc import_title_intro_clip {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set utiltitle "Import title intro clip"

	### remember currently selected title
	set indexlst [split [lindex $::editor_cmds_from 1] ","]
	set origtitle_vts [lindex $indexlst 0]	;# VTS of original title
	set origtitle_ttn [lindex $indexlst 1]	;# TTN of original title

	if {[lindex $::editor_cmds_from 0] == "menu" || $origtitle_vts == 0} {
		set txt "You must select the VTST Title before"
		set txt "$txt\nwhich you want to play the clip"
		set txt "$txt\nbefore calling this macro."
		tk_messageBox -title $utiltitle -icon error -parent . -message $txt
		return
	}

	# find a free GPRM
	set freegprm [lindex [find_unused_gprms true] end]
	if {$freegprm == ""} {
		set txt "This macro need an unused GPRM to work."
		set txt "$txt\nThere are no remaining free GPRMs!\nSorry."
		tk_messageBox -title $utiltitle -icon warning -parent . -message $txt
		return
	}
	set freegprm [format "%02X" $freegprm]

	### GUI
	if {![info exists ::macros::startclip_stilltime]} {
		set ::macros::startclip_stilltime -1		;# still time value: 0, # secs, or 255
		set ::macros::startclip_stilltime_secs 10	;# used only in GUI
	}

	if [winfo exists .importclip] {destroy .importclip}
	set w .importclip
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w $utiltitle
	wm transient $w .
	wm protocol $w WM_DELETE_WINDOW {set ::macros::startclip_done 0}
	bind $w <Escape> "set ::macros::startclip_done 0"

	set txt "This macro will import a new titleset title,"
	set txt "$txt\nand modify the DVD to let the imported clip"
	set txt "$txt\nautomatically play before the currently selected title"
	set txt "$txt\n([$::widget(main_cmd_lab) cget -text])."
	label $w.lab -text $txt
	pack  $w.lab -padx 2

	frame $w.f1 -bd 2 -relief groove
	pack  $w.f1 -padx 4 -pady 2 -fill x
	set txt "You will be prompted for the titleset to import when you select"
	set txt "$txt\nthe Continue button.  Select a VTS with only ONE title!"
	label $w.f1.lab -text $txt -anchor w -justify left
	pack  $w.f1.lab -side left

	frame $w.f2 -bd 2 -relief groove
	pack  $w.f2 -padx 4 -pady 2 -fill x
	set txt "If the imported clip is a still frame,"
	set txt "$txt\nplease select his Still Time (pause duration)"
	label $w.f2.lab -text $txt -anchor w -justify left
	pack  $w.f2.lab -side top -anchor w

	frame $w.f2.f
	pack  $w.f2.f -fill x
	radiobutton $w.f2.f.b2 -text "Display still frame" \
			-variable ::macros::startclip_stilltime -value -1 -anchor w
	pack  $w.f2.f.b2 -side left
	entry $w.f2.f.e -textvariable ::macros::startclip_stilltime_secs -width 4 \
			-validate all -validatecommand "::ced::value_entry_validate 0 254 {%d} {%P}"
	pack  $w.f2.f.e -side left
	label $w.f2.f.l -text "seconds (max 254)"
	pack  $w.f2.f.l -side left
	radiobutton $w.f2.b1 -text "Don't pause: it's an animated clip" \
			-variable ::macros::startclip_stilltime -value 0 -anchor w
	pack  $w.f2.b1 -fill x

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -anchor e
	button $w.fb.ok -text "Continue" -width 10 -command {set ::macros::startclip_done 1}
	pack   $w.fb.ok -side left -padx 2
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::macros::startclip_done 0}
	pack   $w.fb.cancel -side left -padx 2

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr [winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)]
	set wy [expr [winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	wm geometry $w +$wx+$wy
	wm deiconify $w
	grab set $w

	tkwait variable ::macros::startclip_done
	grab release $w
	destroy $w
	if {!$::macros::startclip_done} return

	set startclip_stilltime $::macros::startclip_stilltime
	if {$startclip_stilltime == -1} {
		set startclip_stilltime $::macros::startclip_stilltime_secs
		if {$startclip_stilltime == ""} {set startclip_stilltime 0}
	}

	#################### here we go ##################
	show_msg "Importing Clip to be automatically played..."

	# Title number of the original title
	set origtitle_titlenum $::pgcs($origtitle_vts,$origtitle_ttn,Title)

	### import the new VTS with the intro clip
	set startclip_title_num [::utils::import_VTST]
	if {$startclip_title_num == 0} return

	show_msg "Modifying DVD commands..."

	### create a new VMGM PGC that jumps to our new clip
	set newcmd "30 02 00 00 00 [format "%02X" $startclip_title_num] 00 00"	;# Jump_TT <new title>
	set cmds [list {* pre} $newcmd {* post} {* cell}]
	if {$::menus(0,numlus) == 0} {
		# we need to create a new menu for the VMGM
		::utils::new_title_menu true
		::utils::set_menu_type 0 1 1 0	;# change the new PGC menu type from Title to none
		set ::menus(0,1,1) $cmds
		set new_vmgm_pgc_jump_new 1
	} else {
		# create new PGC
		set new_vmgm_pgc_jump_new [clip_create_VMG_menu $cmds]
	}

	### create a new VMGM PGC that jumps to the original title
	set newcmd1 "71 00 00 $freegprm 00 01 00 00"	;# Set <free GPRM> = 1
	set newcmd2 "30 02 00 00 00 [format "%02X" $origtitle_titlenum] 00 00"	;# Jump_TT <orig title>
	set cmds [list {* pre} $newcmd1 $newcmd2 {* post} {* cell}]
	set new_vmgm_pgc_jump_orig [clip_create_VMG_menu $cmds]


	### Modify the original Title's pre commands to jump to our newly created VMGM menu
	### get the original title's pre cmds
	set cmds $::pgcs($origtitle_vts,$origtitle_ttn)
	set idx2 [expr [lsearch $cmds {* post}] - 1]
	# we are adding 4 commands, so increment the GOTOs by 4 in the original commands
	for {set i 1} {$i <= $idx2} {incr i} {
		set cmd [lindex $cmds $i]
		set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
		if {$cmdchk == "001" || $cmdchk == "003"} {
			set ln [expr 0X[string range $cmd 21 22] + 4]
			lset cmds $i "[string range $cmd 0 19] [format %02X $ln]"
		}
	}
	set digits01 [format "%02X" [expr $new_vmgm_pgc_jump_new / 0X100]]
	set digits23 [format "%02X" [expr $new_vmgm_pgc_jump_new & 0XFF]]
	set ::pgcs($origtitle_vts,$origtitle_ttn) [linsert $cmds 1 \
			"00 A1 00 $freegprm 00 01 00 03" "30 08 $digits01 $digits23 01 C0 00 00" \
			"71 00 00 $freegprm 00 00 00 00" {00 00 00 00 00 00 00 00}]
	 ;# If <free gprm> == 1 Goto 3; CallSS-VMGM-pgc <new_1>; Set <free GPRM> = 0; NOP

	### Modify the new Clip's commands to jump to the new VMGM PGC 2
	set cmds {{* pre} {* post}}
	set digits01 [format "%02X" [expr $new_vmgm_pgc_jump_orig / 0X100]]
	set digits23 [format "%02X" [expr $new_vmgm_pgc_jump_orig & 0XFF]]
	lappend cmds "30 08 $digits01 $digits23 01 C0 00 00" ;# CallSS-VMGM-pgc <new_2>
	lappend cmds {* post}

	set ::pgcs($::pgcs(numvts),1) $cmds

	### Allow to use the Next Chapter button to skip to the end of the imported clip
	set_int ::pgcs($::pgcs(numvts),1,header) 0x9C 1 2

	# create a new Root menu for the imported title, if the original title has a root menu
	set original_root_menu_pgc 0
	if {$::menus($origtitle_vts,numlus) > 0} {
		for {set i 1} {$i <= $::menus($origtitle_vts,1,numpgcs)} {incr i} {
			if {$::menus($origtitle_vts,1,$i,mask) == 131} {
				# create a new VMGM PGC to jump to the original Root menu
				set cmd "30 06 00 [format "%02X" $origtitle_ttn] [format "%02X" $origtitle_vts] 83 00 00"
				;# JumpSS to Root menu of <new VTS> <new TTN>
				set cmds [list "* pre" $cmd "* post" "* cell"]
				set root_jump_pgc [clip_create_VMG_menu $cmds]

				# and create the imported VTS's root menu
				::utils::new_root_menu $::pgcs(numvts) true

				set digits01 [format "%02X" [expr $::menus(0,1,numpgcs) / 0X100]]
				set digits23 [format "%02X" [expr $::menus(0,1,numpgcs) & 0XFF]]
				set cmds [list "* pre" "30 06 $digits01 $digits23 00 C0 00 00" "* post" "* cell"]
				;# CallSS-VMGM-pgc <new root jump pgc>
				set ::menus($::pgcs(numvts),1,1) $cmds

				break
			}
		}
	}

	### Set the imported clip still time
	set_int ::pgcs($::pgcs(numvts),1,cellpbtable) \
			[expr (($::pgcs($::pgcs(numvts),1,numcells) - 1) * 24) + 2] \
			$startclip_stilltime 1

	### refresh interface
	::main::init_tree_menu
	::main::display_pgc pgc $origtitle_vts $origtitle_ttn

	show_msg "Clip successfully appended as Title $startclip_title_num."
	return
}

#############################################################################
## Import VTS Title and automatically play it when current title playback is finished
proc import_title_closing_clip {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set utiltitle "Import title closing clip"

	### remember currently selected title
	set indexlst [split [lindex $::editor_cmds_from 1] ","]
	set origtitle_vts [lindex $indexlst 0]	;# VTS of original title
	set origtitle_ttn [lindex $indexlst 1]	;# TTN of original title

	if {[lindex $::editor_cmds_from 0] == "menu" || $origtitle_vts == 0} {
		set txt "You must select the VTST Title at the end"
		set txt "$txt\nof which you want to play the clip before"
		set txt "$txt\ncalling this macro."
		show_warning -title $utiltitle -icon error -parent . -message $txt
		return
	}

	### GUI
	if {![info exists ::macros::endclip_stilltime]} {
		set ::macros::endclip_stilltime -1		;# still time value: 0, # secs, or 255
		set ::macros::endclip_finish_mode 3		;# 0=exit, 1=FP-PGC, 2=Clone 3=VTS menu
		set ::macros::endclip_stilltime_secs 10	;# used only in GUI
	}

	if [winfo exists .importclip] {destroy .importclip}
	set w .importclip
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w $utiltitle
	wm transient $w .
	wm protocol $w WM_DELETE_WINDOW {set ::macros::endclip_done 0}
	bind $w <Escape> "set ::macros::endclip_done 0"

	set txt "This macro will import a new titleset title,"
	set txt "$txt\nand modify the DVD to let the imported clip"
	set txt "$txt\nautomatically play when the currently selected title"
	set txt "$txt\n([$::widget(main_cmd_lab) cget -text]) has finished playing."
	set txt "$txt\nThis may be used to show an \"Insert Disc 2\" still frame,"
	set txt "$txt\nor to add a featurette from a bonus disc after the main title."
	label $w.lab -text $txt
	pack  $w.lab -padx 2

	frame $w.f1 -bd 2 -relief groove
	pack  $w.f1 -padx 4 -pady 2 -fill x
	set txt "You will be prompted for the titleset to import when you select"
	set txt "$txt\nthe Continue button.  Select a VTS with only ONE title!"
	label $w.f1.lab -text $txt -anchor w -justify left
	pack  $w.f1.lab -side left

	frame $w.f2 -bd 2 -relief groove
	pack  $w.f2 -padx 4 -pady 2 -fill x
	set txt "If the imported clip is a still frame,"
	set txt "$txt\nplease select his Still Time (pause duration)"
	label $w.f2.lab -text $txt -anchor w -justify left
	pack  $w.f2.lab -side top -anchor w

	frame $w.f2.f
	pack  $w.f2.f -fill x
	radiobutton $w.f2.f.b2 -text "Display still frame" \
			-variable ::macros::endclip_stilltime -value -1 -anchor w \
			-command "::macros::validate_endclip_buttons $w"
	pack  $w.f2.f.b2 -side left
	entry $w.f2.f.e -textvariable ::macros::endclip_stilltime_secs -width 4 \
			-validate all -validatecommand "::ced::value_entry_validate 0 254 {%d} {%P}"
	pack  $w.f2.f.e -side left
	label $w.f2.f.l -text "seconds (max 254)"
	pack  $w.f2.f.l -side left
	radiobutton $w.f2.b3 -text "Display still frame forever" \
			-variable ::macros::endclip_stilltime -value 255 -anchor w \
			-command "::macros::validate_endclip_buttons $w"
	pack  $w.f2.b3 -fill x
	radiobutton $w.f2.b1 -text "Don't pause: it's an animated clip" \
			-variable ::macros::endclip_stilltime -value 0 -anchor w \
			-command "::macros::validate_endclip_buttons $w"
	pack  $w.f2.b1 -fill x

	frame $w.f3 -bd 2 -relief groove
	pack  $w.f3 -padx 4 -pady 2 -fill x
	set txt "When the end-clip has finished playing,"
	set txt "$txt\nwhat do you want to do?"
	label $w.f3.lab -text $txt -anchor w -justify left
	pack  $w.f3.lab -side top -anchor w
	radiobutton $w.f3.b0 -text "Return to the Root Menu of the original title." \
			-variable ::macros::endclip_finish_mode -value 3 -anchor w
	pack  $w.f3.b0 -fill x
	radiobutton $w.f3.b1 -text "Restart the DVD from the beginning. (FP-PGC)" \
			-variable ::macros::endclip_finish_mode -value 1 -anchor w
	pack  $w.f3.b1 -fill x
	radiobutton $w.f3.b2 -text "End the DVD playback. (Exit)" \
			-variable ::macros::endclip_finish_mode -value 0 -anchor w
	pack  $w.f3.b2 -fill x
	set txt "Try to Clone the original title behavior."
	set txt "$txt\nWarning: Experimental feature!"
	set txt "$txt\nCheck carefully the DVD before burning!"
	radiobutton $w.f3.b3 -text $txt -justify left \
			-variable ::macros::endclip_finish_mode -value 2 -anchor w
	pack  $w.f3.b3 -fill x


	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -anchor e
	button $w.fb.ok -text "Continue" -width 10 -command {set ::macros::endclip_done 1}
	pack   $w.fb.ok -side left -padx 2
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::macros::endclip_done 0}
	pack   $w.fb.cancel -side left -padx 2

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr [winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)]
	set wy [expr [winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	wm geometry $w +$wx+$wy
	wm deiconify $w
	grab set $w

	### get the original title's post cmds
	set cmds $::pgcs($origtitle_vts,$origtitle_ttn)
	set idx1 [expr [lsearch $cmds {* post}] + 1]
	set idx2 [expr [lsearch $cmds {* cell}] - 1]
	if {$idx2 >= $idx1} {
		set origtitle_post_cmds [lrange $cmds $idx1 $idx2]
	} else {
		set origtitle_post_cmds ""
	}

	# find root menu PGC in original VTS
	set original_root_menu_pgc 0
	if {$::menus($origtitle_vts,numlus) > 0} {
		for {set i 1} {$i <= $::menus($origtitle_vts,1,numpgcs)} {incr i} {
			if {$::menus($origtitle_vts,1,$i,mask) == 131} {
				set original_root_menu_pgc $i
				break
			}
		}
	}

	# check if it is possible to 'clone the original title behavior'
	set cmds $origtitle_post_cmds
	set clone_ok true
	# convert VTST commands to VMGM commands
	if $clone_ok {
		set converted_origcmds [list "* pre"]
		foreach cmd $cmds {
			set cmd [string toupper $cmd]
			# check for SetSTN and SetNVTMR (illegal in the VMG)
			if {[lsearch -regexp $cmds {[45][12] .. .. .. .. .. .. ..}] >= 0} {
				set clone_ok false
				break
			}
			# check for links (can't be processed)
			if {[lsearch -regexp $cmds {20 .. .. .. .. .. .. ..}] >= 0} {
				set clone_ok false
				break
			}
			# CallSS-FP, -> JumpSS-FP and CallSS-VMGM-pgc -> JumpSS-VMGM-pgc
			if {[lsearch -regexp [list $cmd] {30 .8 .. .. .. [0C]. .. ..}] >= 0} {
				set cmd "[string range $cmd 0 3]6[string range $cmd 5 11]00[string range $cmd 14 end]"
			}
			# CallSS_VTSM -> JumpSS_VTSM
			if {[lsearch -regexp [list $cmd] {30 .8 .. .. .. 8. .. ..}] >= 0} {
				set cmd "[string range $cmd 0 3]6 00 [format "%02X" $origtitle_ttn] [format "%02X" $origtitle_vts] [string range $cmd 15 end]"
			}
			# CallSS-VMGM-menu, -> LinkPGCN (to menu pgc)
			if {[lsearch -regexp [list $cmd] {30 .8 .. .. .. 42 .. ..}] >= 0} {
				set pgc 0
				for {set i 1} {$i <= $::menus(0,1,numpgcs)} {incr i} {
					if {$::menus(0,1,$i,mask) == 130} {
						set pgc $i
						break
					}
				}
				if {$pgc == 0} {
					set clone_ok false
					break
				}
				set digit1  [format  "%1X" [expr $pgc / 256]]
				set digit23 [format "%02X" [expr $pgc & 0XFF]]
				set cmd "20 [string range $cmd 3 3]4 00 [string range $cmd 18 19] 00 [string range $cmd 21 22] 0$digit1 $digit23"
			}
			# JumpVTS_TT (ttn) -> JumpTT (title)
			if {[lsearch -regexp [list $cmd] {30 .3 .. .. .. .. .. ..}] >= 0} {
				set ttn [expr "0X[string range $cmd 15 16]"]
				if {! [info exists ::pgcs($origtitle_vts,$ttn,Title)]} {
					set clone_ok false
					break
				} else {
					set title $::pgcs($origtitle_vts,$ttn,Title)
					set cmd "[string range $cmd 0 3]2[string range $cmd 5 14][format "%02X" $title][string range $cmd 17 end]"
				}
			}
			# JumpVTS_PTT (ptt) -> JumpTT (only if ptt==1)
			if {[lsearch -regexp [list $cmd] {30 .5 .. .. .. .. .. ..}] >= 0} {
				set ptt [expr "0X[string range $cmd 7 7][string range $cmd 9 10]"]
				if {$ptt > 1} {
					set clone_ok false
					break
				}
				set ttn [expr "0X[string range $cmd 15 16]"]
				if {! [info exists ::pgcs($origtitle_vts,$ttn,Title)]} {
					set clone_ok false
					break
				} else {
					set title $::pgcs($origtitle_vts,$ttn,Title)
					set cmd "[string range $cmd 0 3]2[string range $cmd 5 14][format "%02X" $title][string range $cmd 17 end]"
				}
			}
			lappend converted_origcmds $cmd
		}
		lappend converted_origcmds "* post" "* cell"
	}
	if {!$clone_ok} {
		set ::macros::clone_on_state "disabled"
		if {$::macros::endclip_finish_mode == 2} {
			set ::macros::endclip_finish_mode 3
		}
	} else {
		set ::macros::clone_on_state "normal"
	}
	if {$original_root_menu_pgc == 0} {
		set ::macros::root_on_state "disabled"
		if {$::macros::endclip_finish_mode == 3} {
			set ::macros::endclip_finish_mode 1
		}
	} else {
		set ::macros::root_on_state "normal"
	}

	validate_endclip_buttons $w
	if {$::macros::endclip_stilltime_secs == ""} {
		set ::macros::endclip_stilltime_secs 10
	}


	tkwait variable ::macros::endclip_done
	grab release $w
	destroy $w
	if {!$::macros::endclip_done} return

	set endclip_stilltime $::macros::endclip_stilltime
	if {$endclip_stilltime == -1} {
		set endclip_stilltime $::macros::endclip_stilltime_secs
		if {$endclip_stilltime == ""} {set endclip_stilltime 0}
	}
	set endclip_finish_mode $::macros::endclip_finish_mode
	if {$endclip_stilltime == 255} {
		set endclip_finish_mode 0
	}

	#################### here we go ##################
	# Variables initialized at this point:
	# origtitle_vts: VTS of original title
	# origtitle_ttn: TTN of original title
	# endclip_stilltime: still time value: 0, # secs, or 255
	# endclip_finish_mode: 0=exit, 1=FP-PGC, 2=Clone, 3=Root menu
	#
	# New variables to be initialized later:
	# origtitle_titlenum: Title number of the original title
	# endclip_title_num: Title number of new clip
	# new_vmgm_pgc_num: PGC number of the newly created VMGM PGC
	# origtitle_post_cmds: Original post commands of the original title

	show_msg "Importing Clip to be automatically played..."

	# Title number of the original title
	set origtitle_titlenum [subst \$::pgcs($origtitle_vts,$origtitle_ttn,Title)]

	### import the new VTS with the end clip
	set endclip_title_num [::utils::import_VTST]
	if {$endclip_title_num == 0} return

	### create a new VMGM PGC that jumps to our new clip
	show_msg "Modifying DVD commands..."
	set newcmd "30 02 00 00 00 [format "%02X" $endclip_title_num] 00 00"	;# Jump_TT
	set cmds [list {* pre} $newcmd {* post} {* cell}]
	if {$::menus(0,numlus) == 0} {
		# we need to create a new menu for the VMGM
		::utils::new_title_menu true
		::utils::set_menu_type 0 1 1 0	;# change the new PGC menu type from Title to none
		set ::menus(0,1,1) $cmds
		set new_vmgm_pgc_num 1
	} else {
		# create new PGC
		set new_vmgm_pgc_num [clip_create_VMG_menu $cmds]
	}

	### Modify the original Title's post commands to jump to our newly created VMGM menu
	set cmds $::pgcs($origtitle_vts,$origtitle_ttn)
	set idx1 [expr [lsearch $cmds {* post}] + 1]
	set idx2 [expr [lsearch $cmds {* cell}] - 1]
	# we are adding 2 commands, so increment the GOTOs by 2 in the original commands
	for {set i $idx1} {$i <= $idx2} {incr i} {
		set cmd [lindex $cmds $i]
		set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
		if {$cmdchk == "001" || $cmdchk == "003"} {
			set ln [expr 0X[string range $cmd 21 22] + 2]
			lset cmds $i "[string range $cmd 0 19] [format %02X $ln]"
		}
	}
	set digits01 [format "%02X" [expr $new_vmgm_pgc_num / 0X100]]
	set digits23 [format "%02X" [expr $new_vmgm_pgc_num & 0XFF]]
	set ::pgcs($origtitle_vts,$origtitle_ttn) [linsert $cmds $idx1 \
			"30 08 $digits01 $digits23 01 C0 00 00" \
			{00 00 00 00 00 00 00 00}]	;# CallSS-VMGM-pgc ; NOP


	### Modify the new Clip's commands according to the finish mode
	set cmds {{* pre} {* post}}
	if {$endclip_finish_mode < 2} {
		# reset all used GPRMs to 0
		set unusedgprms [find_unused_gprms true]
		for {set r 0} {$r < 16} {incr r} {
			if {[lsearch $unusedgprms $r] == -1} {
				lappend cmds "71 00 00 [format "%02X" $r] 00 00 00 00"
			}
		}
	}
	# add finish command
	switch $endclip_finish_mode {
		0 {	# add 'exit'
			lappend cmds "30 01 00 00 00 00 00 00"
		}
		1 {	# add 'CallSS-FP, resume cell 1'
			lappend cmds "30 08 00 00 01 00 00 00"
		}
		2 -
		3 {	# clone original cmds OR Return to root menu
			# add CallSS-VMGM-pgc <new_pgc_num>, resume cell 1
			set new_vmgm_pgc_num2 [expr $new_vmgm_pgc_num + 1]
			set digits01 [format "%02X" [expr $new_vmgm_pgc_num2 / 0X100]]
			set digits23 [format "%02X" [expr $new_vmgm_pgc_num2 & 0XFF]]
			lappend cmds "30 08 $digits01 $digits23 01 C0 00 00"

			# if clone mode, create a new VMGM PGC for the original commands
			if {$endclip_finish_mode == 2} {
				clip_create_VMG_menu $converted_origcmds
			}
		}
	}
	lappend cmds {* cell}
	set ::pgcs($::pgcs(numvts),1) $cmds

	### Allow to use the Next Chapter button to skip to the end of the imported clip
	set_int ::pgcs($::pgcs(numvts),1,header) 0x9C 1 2

	# create a new Root menu for the imported title, if the original title has a root menu
	if {$original_root_menu_pgc > 0} {
		# create a new VMGM PGC to jump to the original Root menu
		set cmd "30 06 00 [format "%02X" $origtitle_ttn] [format "%02X" $origtitle_vts] 83 00 00"
		set cmds [list "* pre" $cmd "* post" "* cell"]
		set root_jump_pgc [clip_create_VMG_menu $cmds]

		# and create the imported VTS's root menu
		::utils::new_root_menu $::pgcs(numvts) true

		set digits01 [format "%02X" [expr $::menus(0,1,numpgcs) / 0X100]]
		set digits23 [format "%02X" [expr $::menus(0,1,numpgcs) & 0XFF]]
		set cmds [list "* pre" "30 06 $digits01 $digits23 00 C0 00 00" "* post" "* cell"]
		set ::menus($::pgcs(numvts),1,1) $cmds
	}

	### Set the imported clip still time
	set_int ::pgcs($::pgcs(numvts),1,cellpbtable) \
			[expr (($::pgcs($::pgcs(numvts),1,numcells) - 1) * 24) + 2] \
			$endclip_stilltime 1

	### refresh interface
	::main::init_tree_menu
	::main::display_pgc pgc $origtitle_vts $origtitle_ttn

	show_msg "Clip successfully appended as Title $endclip_title_num."
	return
}
proc validate_endclip_buttons {w} {
	switch -- $::macros::endclip_stilltime {
		-1 {
			$w.f2.f.e configure -state normal
			$w.f3.b0  configure -state $::macros::root_on_state
			$w.f3.b1  configure -state normal
			$w.f3.b2  configure -state normal
			$w.f3.b3  configure -state $::macros::clone_on_state
			if {$::macros::endclip_stilltime_secs == ""} {
				set ::macros::endclip_stilltime_secs 10
			}
		}
		0 {
			$w.f2.f.e configure -state disabled
			$w.f3.b0  configure -state $::macros::root_on_state
			$w.f3.b1  configure -state normal
			$w.f3.b2  configure -state normal
			$w.f3.b3  configure -state $::macros::clone_on_state
		}
		255 {
			$w.f2.f.e configure -state disabled
			$w.f3.b0  configure -state disabled
			$w.f3.b1  configure -state disabled
			$w.f3.b2  configure -state disabled
			$w.f3.b3  configure -state disabled
		}
	}
}
proc clip_create_VMG_menu {cmds} {
	for {set lunum 1} {$lunum <= $::menus(0,numlus)} {incr lunum} {
		incr ::menus(0,$lunum,numpgcs)
		set pgcnum $::menus(0,$lunum,numpgcs)
		if {$lunum == 1} {
			set new_pgc_num $pgcnum
		}
		set ::menus(0,$lunum,$pgcnum,mask) 0
		set ::menus(0,$lunum,$pgcnum,masktext) ""
		set ::menus(0,$lunum,$pgcnum) $cmds
		set ::menus(0,$lunum,$pgcnum,header) [string repeat \000 236]
		if [info exists ::pgcs(0,1,header)] {
			set_int ::menus(0,$lunum,$pgcnum,header) 4 \
					[expr [get_int ::pgcs(0,1,header) 4 4] & 0XC0] 4
		} else {
			set_int ::menus(0,$lunum,$pgcnum,header) 4 64 4
		}
		set ::menus(0,$lunum,$pgcnum,numprograms) 0
		set ::menus(0,$lunum,$pgcnum,programmap) ""
		set ::menus(0,$lunum,$pgcnum,numcells) 0
		set ::menus(0,$lunum,$pgcnum,cellpbtable) ""
		set ::menus(0,$lunum,$pgcnum,cellpostable) ""
		set ::raw_sectors(0,VMGM_PGCI_UT) "$::raw_sectors(0,VMGM_PGCI_UT)[string repeat \000 8]"
	}
	return $new_pgc_num
}


#############################################################################
## add title button to skip to the next title on a DVD reauthored with DVDShrink
proc title_but_next_title {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set utiltitle "Navigation buttons for DVDShrink reauthored movies"
	if {!$::config(expert_mode)} {
		set txt "This macro work ONLY on DVDShrink reauthored DVDs!"
		set txt "$txt\n"
		set txt "$txt\nThe macro will modify the commands of (almost) every PGCs,"
		set txt "$txt\nand create the VMGM Title menu PGC, so that the user can"
		set txt "$txt\nuse the 'Title' button of the remote to skip to the next title."
		set txt "$txt\n"
		set txt "$txt\nThanks goes to 2COOL for the idea.  See this tread:"
		set txt "$txt\nhttp://forum.doom9.org/showthread.php?threadid=61238"
		set txt "$txt\n"
		set txt "$txt\nThe macro will also allow to use the 'Prev' and 'Next Chapter'"
		set txt "$txt\nbuttons to be used in respectively the first and last chapter"
		set txt "$txt\nof a title, and the 'Return' button to go back to the beginning"
		set txt "$txt\nof the current title, regardless of the current chapter."
		set txt "$txt\nThe last option allow to automatically loop back to the beginning"
		set txt "$txt\nof the DVD when the last title has finished playing."
		set txt "$txt\n"
		set txt "$txt\nWARNING: Only some basic checks are made to ensure that"
		set txt "$txt\nthe DVD is in the correct DVDShrink reauthored format."
		set txt "$txt\nUse this macro only with DVDShrink reauthored DVDs!"
		set txt "$txt\nALWAYS backup the DVD IFOs before using this macro!"
		set txt "$txt\n"
		set txt "$txt\nContinue ?"
		set r [tk_messageBox -title $utiltitle -icon question -message $txt \
				-parent . -type yesno -default no]
		if {$r == "no"} return
	}

	### init GUI defaults to reflect current state
	if {$::menus(0,numlus)==1 && $::menus(0,1,numpgcs)==1} {
		set ::macros::dvdshrinkbuttons_title 1
	} else {
		set ::macros::dvdshrinkbuttons_title 0
	}

	set ::macros::dvdshrinkbuttons_prev [get_int ::pgcs(1,1,header) 0x9E 2]
	if {$::macros::dvdshrinkbuttons_prev > 1} {set ::macros::dvdshrinkbuttons_prev 1}

	set ::macros::dvdshrinkbuttons_next [get_int ::pgcs(1,1,header) 0x9C 2]
	if {$::macros::dvdshrinkbuttons_next > 1} {set ::macros::dvdshrinkbuttons_next 1}

	set ::macros::dvdshrinkbuttons_return [get_int ::pgcs(1,1,header) 0xA0 2]
	if {$::macros::dvdshrinkbuttons_return > 1} {set ::macros::dvdshrinkbuttons_return 1}

	if {[lsearch [get_type_cmds post ::pgcs($::pgcs(numvts),1)] \
			{30 08 00 00 01 00 00 00}] >= 0} {	;# CallSS FP
		set ::macros::dvdshrinkbuttons_loop 1
	} else {
		set ::macros::dvdshrinkbuttons_loop 0
	}

	### GUI
	if [winfo exists .dvdshrinkbuttons] {destroy .dvdshrinkbuttons}
	set w .dvdshrinkbuttons
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w $utiltitle
	wm transient $w .
	wm protocol $w WM_DELETE_WINDOW {set ::macros::dvdshrinkbuttons_done 0}
	bind $w <Escape> "set ::macros::dvdshrinkbuttons_done 0"

	set txt "This macro allow to use the remote's buttons to navigate in DVDs reauthored with DVDShrink."
	label $w.lab -text $txt
	pack  $w.lab -padx 2 -pady 4

	frame $w.f1 -bd 2 -relief groove
	pack  $w.f1 -padx 4 -pady 2 -fill x
	checkbutton $w.f1.b2 -text "When in chapter 1, allow the 'Prev Chapter' button to go back to the beginning of the current title?" \
			-variable ::macros::dvdshrinkbuttons_prev
	pack  $w.f1.b2 -anchor w
	checkbutton $w.f1.b3 -text "When in last chapter, allow the 'Next Chapter' button to skip to the next title?" \
			-variable ::macros::dvdshrinkbuttons_next
	pack  $w.f1.b3 -anchor w
	checkbutton $w.f1.b6 -text "Always allow the 'Return' (GoUp) button to go back to the beginning of the current title?" \
			-variable ::macros::dvdshrinkbuttons_return
	pack  $w.f1.b6 -anchor w
	checkbutton $w.f1.b1 -text "Always allow the 'Title Menu' button to skip to the next title?" \
			-variable ::macros::dvdshrinkbuttons_title
	pack  $w.f1.b1 -anchor w
	checkbutton $w.f1.b4 -text "At the end of the last title, automatically loop back to the first title?" \
			-variable ::macros::dvdshrinkbuttons_loop
	pack  $w.f1.b4 -anchor w
	button $w.f1.b5 -text "Check all" -width 10 -command {
		set ::macros::dvdshrinkbuttons_title  1
		set ::macros::dvdshrinkbuttons_return 1
		set ::macros::dvdshrinkbuttons_prev   1
		set ::macros::dvdshrinkbuttons_next   1
		set ::macros::dvdshrinkbuttons_loop   1
	}
	pack  $w.f1.b5 -anchor w -padx 4 -pady 4


	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -anchor e
	button $w.fb.ok -text "Continue" -width 10 -command {set ::macros::dvdshrinkbuttons_done 1}
	pack   $w.fb.ok -side left -padx 2
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::macros::dvdshrinkbuttons_done 0}
	pack   $w.fb.cancel -side left -padx 2

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr [winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)]
	set wy [expr [winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	wm geometry $w +$wx+$wy
	wm deiconify $w
	grab set $w

	tkwait variable ::macros::dvdshrinkbuttons_done
	grab release $w
	destroy $w
	if {!$::macros::dvdshrinkbuttons_done} return

	if $::macros::dvdshrinkbuttons_title {
		### create or replace the VMGM menu
		set r "yes"
		if {$::menus(0,numlus)>0} {
			set txt "WARNING: There is already a VMGM Menu!"
			set txt "$txt\nIf you continue, it will be REPLACED by a new one!"
			set txt "$txt\n"
			set txt "$txt\nContinue anyway ?"
			set r [tk_messageBox -title $utiltitle -icon warning -message $txt \
					-parent . -type yesno -default yes]
		}
		if {$r == "yes"} {
			set ::raw_sectors(0,VMGM_PGCI_UT) \
					[readFile $::config(apppath)/lib/data/vmgm_pgci_ut_wo_vob.sec binary]

			parse_PGCI_UT ::raw_sectors(0,VMGM_PGCI_UT) 0 ::menus
			::main::init_tree_menu

			### for each VTS except last one, replace the pre command "Set gprm(0) = 0"
			### by "Set gprm(0) = <current VTS number>"
			for {set vts 1} {$vts < $::pgcs(numvts)} {incr vts} {
				set cmds [get_type_cmds pre ::pgcs($vts,1)]
				set idx [lsearch -regexp $cmds {71 00 00 00 .. .. 00 00}]
				if {$idx == -1} {
					set cmds [linsert $cmds 0 "71 00 00 00 00 [format %02X $vts] 00 00"]
				} else {
					lset cmds $idx "71 00 00 00 00 [format %02X $vts] 00 00"
				}
				set_type_cmds pre ::pgcs($vts,1) $cmds
			}
		}
	} else {
		### delete the VMGM menu
		set r "yes"
		if {$::menus(0,numlus)>0} {
			set txt "WARNING: The entire VMGM Menu will be DELETED!"
			set txt "$txt\n"
			set txt "$txt\nContinue anyway ?"
			set r [tk_messageBox -title $utiltitle -icon warning -message $txt \
					-parent . -type yesno -default yes]
			if {$r == "no"} return
		}
		if {$r == "yes"} {
			set ::raw_sectors(0,VMGM_PGCI_UT) ""
			set ::menus(0,numlus) 0
			::main::init_tree_menu

			### for each VTS except last one, replace the pre command
			### "Set gprm(0) = <current VTS number>" by "Set gprm(0) = 0"
			for {set vts 1} {$vts < $::pgcs(numvts)} {incr vts} {
				set cmds [get_type_cmds pre ::pgcs($vts,1)]
				set idx [lsearch -regexp $cmds {71 00 00 00 .. .. 00 00}]
				if {$idx != -1} {
					lset cmds $idx "71 00 00 00 00 00 00 00"
				}
				set_type_cmds pre ::pgcs($vts,1) $cmds
			}
		}
	}

	if $::macros::dvdshrinkbuttons_loop {
		### Last VTS: replace Exit by CallSS FP
		set vts $::pgcs(numvts)
		set cmds [get_type_cmds post ::pgcs($vts,1)]
		if {$cmds == ""} {
			set cmds {{30 08 00 00 01 00 00 00}}
		} else {
			set idx [lsearch $cmds {30 01 00 00 00 00 00 00}]	;# Exit
			if {$idx >= 0} {
				lset cmds $idx "30 08 00 00 01 00 00 00"	;# CallSS FP
			}
		}
		set_type_cmds post ::pgcs($vts,1) $cmds
	} else {
		### Last VTS: replace CallSS-FP by Exit
		set vts $::pgcs(numvts)
		set cmds [get_type_cmds post ::pgcs($vts,1)]
		if {$cmds == ""} {
			set cmds {{30 01 00 00 00 00 00 00}}
		} else {
			set idx [lsearch $cmds {30 08 00 00 01 00 00 00}]	;# CallSS FP
			if {$idx >= 0} {
				lset cmds $idx "30 01 00 00 00 00 00 00"	;# Exit
			}
		}
		set_type_cmds post ::pgcs($vts,1) $cmds
	}

	### Allow to use the Prev button, when pressed during Chapter 1) to go back to the
	### beginning of the title, and the Next Button to go directly to the end of the Title
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		if $::macros::dvdshrinkbuttons_prev {
			set_int ::pgcs($vts,1,header) 0x9E 1 2
		} else {
			set_int ::pgcs($vts,1,header) 0x9E 0 2
		}
		if $::macros::dvdshrinkbuttons_next {
			set_int ::pgcs($vts,1,header) 0x9C 1 2
		} else {
			set_int ::pgcs($vts,1,header) 0x9C 0 2
		}
		if $::macros::dvdshrinkbuttons_return {
			set_int ::pgcs($vts,1,header) 0xA0 1 2
		} else {
			set_int ::pgcs($vts,1,header) 0xA0 0 2
		}
	}

	### util done.  Refresh listbox.
	refresh_current_pgc
	::main::dvd_modified true false

	# if there is an old VMGM VOB, delete it or move it to backup
	::utils::delete_vob_with_confirm "VIDEO_TS.VOB" true
}

#############################################################################
## enable navigation buttons on the remote
proc nav_buttons {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	if {![info exists ::macros::navbuttons_prev]} {
		set ::macros::navbuttons_prev   1
		set ::macros::navbuttons_next   1
		set ::macros::navbuttons_return 1
		set ::macros::navbuttons_menu	1
		set ::macros::navbuttons_menu_opt 1
		set ::macros::navbuttons_domain 1
	}

	### GUI
	set w .navbuttons
	if [winfo exists $w] {destroy $w}
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w "Enable Remote Navigation Buttons"
	wm transient $w .
	wm protocol $w WM_DELETE_WINDOW {set ::macros::navbuttons_done 0}
	bind $w <Escape> "set ::macros::navbuttons_done 0"

	frame $w.f1 -bd 2 -relief groove
	pack  $w.f1 -padx 4 -pady 4 -fill x
	checkbutton $w.f1.b2 -text "When in chapter 1, allow the 'Prev Chapter' button\nto go back to the beginning of the current PGC?" \
			-variable ::macros::navbuttons_prev -justify left
	pack  $w.f1.b2 -anchor w

	checkbutton $w.f1.b3 -text "When in last chapter, allow the 'Next Chapter' button\nto skip to the end of the PGC?" \
			-variable ::macros::navbuttons_next -justify left
	pack  $w.f1.b3 -anchor w

	checkbutton $w.f1.b6 -text "In the Title domain, allow the 'Return' (GoUp) button\nto go back to the beginning of the current PGC?" \
			-variable ::macros::navbuttons_return -justify left
	pack  $w.f1.b6 -anchor w

	checkbutton $w.f1.b4 -text "In the Menu domain, allow the 'Return' (GoUp) button" \
			-variable ::macros::navbuttons_menu -justify left \
			-command "if \$::macros::navbuttons_menu {set state normal} else {set state disabled}; $w.f1.f4.f.r1 configure -state \$state;  $w.f1.f4.f.r2 configure -state \$state"
	pack  $w.f1.b4 -anchor w
	frame $w.f1.f4
	pack  $w.f1.f4 -fill x
	frame $w.f1.f4.pad -width 20
	pack  $w.f1.f4.pad -side left
	frame $w.f1.f4.f
	pack  $w.f1.f4.f -fill x
	radiobutton $w.f1.f4.f.r1 -text "...to go back to the beginning of the current PGC?" \
			-variable ::macros::navbuttons_menu_opt -value 1 -justify left -bd 1
	pack  $w.f1.f4.f.r1 -anchor w
	radiobutton $w.f1.f4.f.r2 -text "...to resume to the last visited Title?" \
			-variable ::macros::navbuttons_menu_opt -value 0 -justify left -bd 1
	pack  $w.f1.f4.f.r2 -anchor w
	if $::macros::navbuttons_menu {set state normal} else {set state disabled}
	$w.f1.f4.f.r1 configure -state $state
	$w.f1.f4.f.r2 configure -state $state

	frame $w.f2 -bd 2 -relief groove
	pack  $w.f2 -padx 4 -pady 2 -fill x
	radiobutton $w.f2.r1 -variable ::macros::navbuttons_domain -value 1 \
			-text "Apply to current PGC only"
	pack  $w.f2.r1 -anchor w
	radiobutton $w.f2.r2 -variable ::macros::navbuttons_domain -value 2 \
			-text "Apply to all PGCs of the current domain"
	pack  $w.f2.r2 -anchor w
	radiobutton $w.f2.r3 -variable ::macros::navbuttons_domain -value 3 \
			-text "Apply to all Titles of the DVD"
	pack  $w.f2.r3 -anchor w
	radiobutton $w.f2.r4 -variable ::macros::navbuttons_domain -value 4 \
			-text "Apply to all PGCs of the VMGM menu"
	pack  $w.f2.r4 -anchor w

	set txt "Notes: The buttons behavior are changed only when not"
	set txt "$txt\nalready used for another purpose, and are not changed"
	set txt "$txt\nfor 'dummy' menu PGCs (without video)."
	set txt "$txt\nThe 'Next Chapter' button is enabled only if there are post"
	set txt "$txt\ncommands in the PGC."
	set txt "$txt\nEnabling the buttons on menus may lead to strange results!"
	label $w.lab -text $txt -justify left
	pack  $w.lab -fill x -anchor w -padx 4

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -anchor e
	button $w.fb.app -text "Apply" -width 10 -command {set ::macros::navbuttons_done 2}
	pack   $w.fb.app -side left -padx 2
	button $w.fb.ok  -text "OK" -width 10 -command {set ::macros::navbuttons_done 1}
	pack   $w.fb.ok  -side left -padx 2
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::macros::navbuttons_done 0}
	pack   $w.fb.cancel -side left -padx 2

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr [winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)]
	set wy [expr [winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	wm geometry $w +$wx+$wy
	wm deiconify $w

	set done 0
	while {! $done} {
		set ::macros::navbuttons_done -1
		tkwait variable ::macros::navbuttons_done
		if {$::macros::navbuttons_done==0} {
			destroy $w
			return
		}
		if {$::macros::navbuttons_done==1} {
			destroy $w
			set done 1
		}

		set curtype [lindex $::editor_cmds_from 0]
		set curvts  [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
		if {($curtype == "menu" && $::macros::navbuttons_domain!=3) \
				|| $::macros::navbuttons_domain==4} {
			set curlu  [lindex [split [lindex $::editor_cmds_from 1] ","] 1]
			set curpgc [lindex [split [lindex $::editor_cmds_from 1] ","] 2]
			switch $::macros::navbuttons_domain {
				1 {
					set startvts $curvts
					set endvts   $curvts
					set startlu  $curlu
					set endlu    $curlu
					set startpgc $curpgc
					set endpgc   $curpgc
				}
				2 {
					set startvts $curvts
					set endvts   $curvts
					set startlu  1
					set endlu    $::menus($curvts,numlus)
					set startpgc 1
					set endpgc   -1
				}
				4 {
					set startvts 0
					set endvts   0
					set startlu  1
					set endlu    $::menus(0,numlus)
					set startpgc 1
					set endpgc   -1
				}
			}
			for {set vts $startvts} {$vts <= $endvts} {incr vts} {
				for {set lu $startlu} {$lu <= $endlu} {incr lu} {
					if {$endpgc == -1} {
						set endpgc2 $::menus($vts,$lu,numpgcs)
					} else {
						set endpgc2 $endpgc
					}
					for {set pgc $startpgc} {$pgc <= $endpgc2} {incr pgc} {
						if {$::menus($vts,$lu,$pgc,numprograms)>0} {
							if {$::macros::navbuttons_prev   && \
									[get_int ::menus($vts,$lu,$pgc,header) 0x9E 2]==0} {
								set_int ::menus($vts,$lu,$pgc,header) 0x9E $pgc 2
							}
							if {$::macros::navbuttons_next   && \
									[get_int ::menus($vts,$lu,$pgc,header) 0x9C 2]==0 &&
									[num_type_cmds post ::menus($vts,$lu,$pgc)]>0} {
								set_int ::menus($vts,$lu,$pgc,header) 0x9C $pgc 2
							}
							if {$::macros::navbuttons_menu && \
									[get_int ::menus($vts,$lu,$pgc,header) 0xA0 2]==0} {
								if $::macros::navbuttons_menu_opt {
									set_int ::menus($vts,$lu,$pgc,header) 0xA0 $pgc 2
								} else {
									set_int ::menus($vts,$lu,$pgc,header) 0xA0 0xFFFF 2
								}
							}
						}
					}
				}
			}
		} else {
			set curpgc [lindex [split [lindex $::editor_cmds_from 1] ","] 1]
			switch $::macros::navbuttons_domain {
				1 {
					set startvts $curvts
					set endvts   $curvts
					set startpgc $curpgc
					set endpgc   $curpgc
				}
				2 {
					set startvts $curvts
					set endvts   $curvts
					set startpgc 1
					set endpgc   $::pgcs($curvts,numpgcs)
				}
				3 {
					set startvts 1
					set endvts   $::pgcs(numvts)
					set startpgc 1
					set endpgc   -1
				}
			}
			for {set vts $startvts} {$vts <= $endvts} {incr vts} {
				if {$endpgc == -1} {
					set endpgc2 $::pgcs($vts,numpgcs)
				} else {
					set endpgc2 $endpgc
				}
				for {set pgc $startpgc} {$pgc <= $endpgc2} {incr pgc} {
					if {$::macros::navbuttons_prev   && \
							[get_int ::pgcs($vts,$pgc,header) 0x9E 2]==0} {
						set_int ::pgcs($vts,$pgc,header) 0x9E $pgc 2
					}
					if {$::macros::navbuttons_next   && \
							[get_int ::pgcs($vts,$pgc,header) 0x9C 2]==0 &&
							[num_type_cmds post ::pgcs($vts,$pgc)]>0} {
						set_int ::pgcs($vts,$pgc,header) 0x9C $pgc 2
					}
					if {$::macros::navbuttons_return && \
							[get_int ::pgcs($vts,$pgc,header) 0xA0 2]==0} {
						set_int ::pgcs($vts,$pgc,header) 0xA0 $pgc 2
					}
				}
			}
		}
		### util done.  Refresh listbox.
		refresh_current_pgc
		::main::dvd_modified true false
	}
	return
}


#############################################################################
## Kill PGC Playback
## Modify the PGC pre commands so that the PGC video is not played anymore.
proc kill_pgc_playback {args} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}

	if {$args != ""} {
		set pgcvar $args
	} else {
		set pgcvar ::[lindex $::editor_cmds_from 0]s([lindex $::editor_cmds_from 1])
		set numcellcmds [num_type_cmds cell $pgcvar]

		if {!$::config(expert_mode)} {
			set txt "This macro will inhibit the video playback"
			set txt "$txt\nof the current PGC."
			set txt "$txt\nThe VOB file will not be modified."
			set txt "$txt\n"
			if {$numcellcmds > 0} {
				set txt "$txt\nWarning: there are cell commands in this PGC!"
				set txt "$txt\nThe cell commands will not be executed anymore."
				set txt "$txt\n"
			}
			set txt "$txt\nContinue ?"
			set r [tk_messageBox -title "Kill PGC Playback" -icon warning -message $txt \
					-parent . -type yesno -default yes]
			if {$r == "no"} return
		}
		if {$numcellcmds > 0} {
			show_msg "Kill PGC Playback: The cell commands will not be executed anymore!"
		}
	}

	# get pre-commands, and add a NOP command at the end, if needed and possible.
	set cmds [get_type_cmds pre $pgcvar]
	if {[lindex $cmds end]!={00 00 00 00 00 00 00 00}} {
		if {[expr [num_type_cmds pre $pgcvar]+[num_type_cmds post $pgcvar]+[num_type_cmds cell $pgcvar]] < 128 ||
				[num_type_cmds post $pgcvar]==0} {
			lappend cmds {00 00 00 00 00 00 00 00}	;# NOP
		}
	}
	set gotocmd [llength $cmds]	;# we will change the Breaks to Gotos to this line number

	set oktoblank true

	# search for Break commands
	set idxs [lsearch -regexp -all $cmds {00 .2 .. .. .. .. .. ..}]	;# Break
	foreach idx $idxs {
		set bcmd [lindex $cmds $idx]
		set gcmd [string range $bcmd 0 3]1[string range $bcmd 5 20][format %02X $gotocmd]
		lset cmds $idx $gcmd
	}
	# search for Link*Cell and Link*PG commands
	# replace them with gotos only if the Highlight parameter is NOT present.
	set idxs [lsearch -regexp -all $cmds {20 .1 .. .. .. .. .. 0[1-7]}]	;# Link*Cell and Link*PG
	foreach idx $idxs {
		set bcmd [lindex $cmds $idx]
		if {[string range $bcmd 18 19] == "00"} {
			set gcmd "00 [string range $bcmd 3 3]1 00 [string range $bcmd 9 16] 00 [format %02X $gotocmd]"
			lset cmds $idx $gcmd
		} else {
			set oktoblank false
		}
	}
	# search for LinkPTT, LinkPGN and LinkCN commands
	# replace them with gotos only if the Highlight parameter is NOT present.
	set idxs [lsearch -regexp -all $cmds {20 .[5-7] .. .. .. .. .. ..}]	;# LinkPTT, LinkPGN and LinkCN
	foreach idx $idxs {
		set bcmd [lindex $cmds $idx]
		if {[string range $bcmd 18 19] == "00"} {
			set gcmd "00 [string range $bcmd 3 3]1 00 [string range $bcmd 9 16] 00 [format %02X $gotocmd]"
			lset cmds $idx $gcmd
		} else {
			set oktoblank false
		}
	}

	# add post-commands after the NOP
	set postcmds [get_type_cmds post $pgcvar]
	# increment GOTOs
	set idxs [lsearch -regexp -all $postcmds {00 .[13] .. .. .. .. .. ..}]	;# GOTOs
	foreach idx $idxs {
		set cmd [lindex $postcmds $idx]
		set ln [expr 0x[lindex $cmd end]+$gotocmd]
		lset cmd end [format "%02X" $ln]
		lset postcmds $idx $cmd
	}
	# append post-cmds to pre-cmds
	set cmds [concat $cmds $postcmds]

	# replace pre-commands
	set_type_cmds pre $pgcvar $cmds

	# remove post commands if the total number of commands > 128
	if {[expr [num_type_cmds pre $pgcvar]+[num_type_cmds post $pgcvar]+[num_type_cmds cell $pgcvar]] > 128} {
		set_type_cmds post $pgcvar {}
		show_msg "Kill Playback WARNING: post commands removed!  (Max 128 commands in a PGC)"
		bell
	}

	# reset PGC Still Time to 0
	if {[lindex $::editor_cmds_from 0] == "menu"} {
		set_int ::menus([lindex $::editor_cmds_from 1],header) 0xA3 0 1
		for {set cell 1} {$cell <= $::menus([lindex $::editor_cmds_from 1],numcells)} {incr cell} {
			set cellidx [expr ($cell-1)*24]
			# still time = 0
			set_int ::menus([lindex $::editor_cmds_from 1],cellpbtable) [expr $cellidx+2] 0 1
		}
	} else {
		set_int ::pgcs([lindex $::editor_cmds_from 1],header) 0xA3 0 1
		for {set cell 1} {$cell <= $::pgcs([lindex $::editor_cmds_from 1],numcells)} {incr cell} {
			set cellidx [expr ($cell-1)*24]
			# still time = 0
			set_int ::pgcs([lindex $::editor_cmds_from 1],cellpbtable) [expr $cellidx+2] 0 1
		}
	}

	# remove Cells and Programs if type is menu.
	if {! $oktoblank} {
		show_msg "Kill Playback WARNING: Link to Cell, PG or PTT command with highlight button not changed to GoTo."
		bell
	}
	if {$args == ""} {
		refresh_current_pgc

		if {[lindex $::editor_cmds_from 0] == "menu" && \
				$::menus([lindex $::editor_cmds_from 1],numcells) > 0 &&
				$oktoblank} {
			set txt "You killed a Menu PGC playback."
			set txt "$txt\n"
			set txt "$txt\nDo you want to remove also the"
			set txt "$txt\nPrograms and Cells from the PGC?"
			set txt "$txt\nThis will allow to easily strip the"
			set txt "$txt\nVOB ID with IfoEdit."
			set r [tk_messageBox -title "Kill PGC Playback" -icon question -message $txt \
					-parent . -type yesno -default yes]
			if {$r == "yes"} {
				set ::menus([lindex $::editor_cmds_from 1],numprograms) 0
				set ::menus([lindex $::editor_cmds_from 1],programmap) ""
				set ::menus([lindex $::editor_cmds_from 1],numcells) 0
				set ::menus([lindex $::editor_cmds_from 1],cellpbtable) ""
				set ::menus([lindex $::editor_cmds_from 1],cellpostable) ""

				::main::init_tree_menu_no_move
			}
		}
	}
	::main::dvd_modified true false
}

#############################################################################
## Enable all operations for the current PGC
proc enable_all_operations {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set type  [lindex $::editor_cmds_from 0]
	set index [lindex $::editor_cmds_from 1]

	# clear PUOs
	set_int ::${type}s($index,header) 8 0 4

	# if it is a Title PGC, clear also the PUOs in the TT_SRPT table
	if {$type == "pgc" && [info exists ::${type}s($index,Title)]} {
		set data $::raw_sectors(0,VMG_TT_SRPT)
		set title [subst \$::${type}s($index,Title)]
		set numtitles [get_int data 0 2]
		if {$numtitles >= $title} {
			set t_idx [expr 8+(($title-1)*12)]
			set typebyte [expr [get_int data $t_idx 1] & 0xFC]
			set_int ::raw_sectors(0,VMG_TT_SRPT) $t_idx $typebyte 1
		}
	}

	# enable navigation buttons
	set txt ""
	if {[subst \$::${type}s($index,numprograms)] > 0} {
		set pgcnum [lindex [split $index ","] end]
		if {[get_int ::${type}s($index,header) 0x9C 2] == 0} {
			set cmds [subst \$::${type}s($index)]
			if {[num_type_cmds "post" cmds] > 0} {
				set_int ::${type}s($index,header) 0x9C $pgcnum 2
			}
		}
		if {[get_int ::${type}s($index,header) 0x9E 2] == 0} {
			set_int ::${type}s($index,header) 0x9E $pgcnum 2
		}
		if {$type == "menu" && [get_int ::${type}s($index,header) 0xA0 2] == 0} {
			set_int ::${type}s($index,header) 0xA0 65535 2
			set txt " and Prev, Next and GoUp"
		} else {
			set txt " and Prev and Next"
		}
		set txt "$txt navigation buttons enabled"
	}

	show_msg "All Prohibited User Operations cleared$txt."
	::main::dvd_modified true false
	return
}

#############################################################################
## Completely remove the parental managment from a DVD
proc remove_parental_managment {} {
	set utiltitle "Remove Parental Managment"
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	if {!$::config(expert_mode)} {
		set txt "This utility will remove the parental managment:"
		set txt "$txt\ndelete the VMG_PTL_MAIT table, reset all Parental ID"
		set txt "$txt\nmasks in the VMG_TT_SRPT and the VTS_PGCITI tables,"
		set txt "$txt\nand change the SetTmpPML commands to GOTOs.\n"
		set txt "$txt\nThe DVD will therefore ignore the parental level"
		set txt "$txt\nof the player.\n"
		set txt "$txt\nAre you sure ?"
		set r [tk_messageBox -title $utiltitle -icon warning -message $txt \
				-parent . -type yesno -default no]
		if {$r == "no"} return
	}

	# remove the VMG_PTL_MAIT table
	if {$::raw_sectors(0,VMG_PTL_MAIT) != ""} {
		set ::raw_sectors(0,VMG_PTL_MAIT) ""
		set log "The VMG_PTL_MAIT table has been erased."
	} else {
		set log "The VMG_PTL_MAIT table was not defined."
	}

	# Reset all PM Masks for all titles in VMG_TT_SRPT to 0.
	set numtitles [get_int ::raw_sectors(0,VMG_TT_SRPT) 0 2]
	for {set t 1} {$t <= $numtitles} {incr t} {
		set_int ::raw_sectors(0,VMG_TT_SRPT) [expr $t*12] 0 2
	}
	set log "${log}\nParental ID masks for $numtitles Titles cleared in VMG_TT_SRPT table."


	# Reset the Category Parental ID mask for all PGCs in all VTS_PGCI to 0.
	set numchanges 0
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		set numpgcs [get_int ::raw_sectors($vts,VTS_PGCI) 0 2]
		for {set pgc 1} {$pgc <= $numpgcs} {incr pgc} {
			set idx [expr 2+($pgc*8)]
			set_int ::raw_sectors($vts,VTS_PGCI) $idx 0 2
		}
		incr numchanges $numpgcs
	}
	set log "${log}\nParental ID masks of $numchanges PGCs cleared in $::pgcs(numvts) VTS_PGCITI tables."

	# Change all SetTmpPML to Goto (to same line).
	set numchanges 0
	for {set lu 1} {$lu <= $::menus(0,numlus)} {incr lu} {
		for {set pgc 1} {$pgc <= $::menus(0,$lu,numpgcs)} {incr pgc} {
			set idxs [lsearch -regexp -all $::menus(0,$lu,$pgc) {00 .3 .. .. .. .. .. ..}]	;# SetTmpPML
			foreach idx $idxs {
				set scmd [lindex $::menus(0,$lu,$pgc) $idx]
				set gcmd [string range $scmd 0 3]1[string range $scmd 5 17]00[string range $scmd 20 end]
				lset ::menus(0,$lu,$pgc) $idx $gcmd
				incr numchanges
			}
		}
	}
	switch $numchanges {
		0 {set log "${log}\nNo SetTmpPML command found in the VMGM."}
		1 {set log "${log}\nOne SetTmpPML command changed to GoTo in the VMGM."}
		default {set log "${log}\n$numchanges SetTmpPML commands changed to GoTo in the VMGM."}
	}

	# Change all instances of SPRM(13) by constant 15: left to user.

	# Done. Refresh interface
	::main::dvd_modified true false
	::main::display_pgc pgc 0 0
	set ::main::searchstring "sprm(13:"

	set txt "Done:\n$log\n"
	set txt "${txt}You will never be prompted for a password anymore!\n\n"
	set txt "${txt}To be sure that the Parental level is never used in the DVD,\n"
	set txt "${txt}you should search for any occurence of sprm(13:Parental level)\n"
	set txt "${txt}and change the sprm 13 in the command by the constant value 15.\n"
	set txt "${txt}(The search field is alerady initialized with the string \"sprm(13:\")\n"
	set txt "${txt}For example, change \"Set gprm(4) =(mov) sprm(13:Parental level)\"\n"
	set txt "${txt}to \"Set gprm(4) =(mov) 15\".\n"

	set r [tk_messageBox -title $utiltitle -icon info -message $txt \
			-parent . -type ok]
	return
}

#############################################################################
## Find Unused GPRMs
proc find_unused_gprms {quiet} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}

	for {set i 0} {$i < 16} {incr i} {
		set used($i) 0
	}

	# FP PGC
	set r [get_used_gprms ::pgcs(0,0)]
	for {set i 0} {$i < 16} {incr i} {
		incr used($i) [lindex $r $i]
	}

	# Titles
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		for {set ttn 1} {$ttn <= $::pgcs($vts,numpgcs)} {incr ttn} {
			set r [get_used_gprms ::pgcs($vts,$ttn)]
			for {set i 0} {$i < 16} {incr i} {
				incr used($i) [lindex $r $i]
			}
		}
	}

	# Menus
	for {set vts 0} {$vts <= $::menus(numvts)} {incr vts} {
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set r [get_used_gprms ::menus($vts,$lu,$pgc)]
				for {set i 0} {$i < 16} {incr i} {
					incr used($i) [lindex $r $i]
				}
			}
		}
	}

	# Menu buttons
	for {set vts 0} {$vts <= $::menus(numvts)} {incr vts} {
		set butcmds {}
		if {[info exists ::buttons::butinfo($vts,errorcode)] && \
				$::buttons::butinfo($vts,errorcode) == 0} {
			for {set vid 1} {$vid <= $::buttons::butinfo($vts,numvobids)} {incr vid} {
				for {set cid 1} {$cid <= $::buttons::butinfo($vts,$vid,numcellids)} {incr cid} {
					if {[info exists ::buttons::butinfo($vts,$vid,$cid,numbuts)] && \
								$::buttons::butinfo($vts,$vid,$cid,numbuts) > 0} {
						for {set group 1} {$group <= $::buttons::butinfo($vts,$vid,$cid,numgroups)} {incr group} {
							for {set but 1} {$but <= $::buttons::butinfo($vts,$vid,$cid,numbuts)} {incr but} {
								lappend butcmds $::buttons::butinfo($vts,$vid,$cid,$group,$but,command)
							}
						}
					}
				}
			}
		}
		set r [get_used_gprms butcmds]
		for {set i 0} {$i < 16} {incr i} {
			incr used($i) [lindex $r $i]
		}
	}
	unset butcmds

	set txt ""
	set lst ""
	set usedtxt ""
	for {set i 0} {$i < 16} {incr i} {
		if {$used($i) == 0} {
			lappend lst $i
			set txt "$txt$i, "
		} else {
			set usedtxt "$usedtxt\ngprm($i) : $used($i)"
		}
	}
	if {$txt==""} {
		set txt "All GPRMs are used!"
	} else {
		set txt "Unused GPRM numbers:\n[string trim $txt ", "]."
	}
	if {!$quiet} {
		show_msg $txt
		if {$::tcl_platform(platform) != "macintosh"} {
			set txt "$txt\n\nUsage count (including menubuttons):$usedtxt"
		} else {
			set txt "$txt\n\nUsage count (EXCLUDING menubuttons):$usedtxt"
		}
		tk_messageBox -title "Find Unused GPRMs" -icon info -message $txt -parent .
	}
	return $lst
}
proc get_used_gprms {cmds_upvar} {
	upvar $cmds_upvar cmds
	for {set i 0} {$i < 16} {incr i} {
		set used($i) 0
	}
	foreach cmd $cmds {
		if {[string range $cmd 0 0] == "*"} continue
		set d [::disasm::disasm $cmd]
		set idx [string first "gprm(" $d]
		while {$idx >= 0} {
			incr idx 5
			scan [string range $d $idx end] "%d" g
			incr used($g)
			set idx [string first "gprm(" $d $idx]
		}
	}
	set allused ""
	for {set i 0} {$i < 16} {incr i} {
		lappend allused $used($i)
	}
	return $allused
}

#############################################################################
## Search for all used GPRMs in the DVD and insert the commands to reset them
## after the current line
proc reset_used_gprms {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	if {!$::config(expert_mode)} {
		set txt "This macro will search for all used GPRMs"
		set txt "$txt\nand insert the commands to reset them to 0"
		set txt "$txt\nAFTER the current selection."
		set txt "$txt\nThe GOTOs lines are renumbered if needed."
		set txt "$txt\n"
		set txt "$txt\nContinue ?"
		set r [tk_messageBox -title "Reset Used GPRMs" -icon question -message $txt \
				-parent . -type yesno -default yes]
		if {$r == "no"} return
	}

	# find all used GPRMs
	set unusedgprms [find_unused_gprms true]
	set usedcmds ""
	for {set u 0} {$u < 16} {incr u} {
		if {[lsearch $unusedgprms $u] == -1} {
			lappend usedcmds "71 00 00 [format "%02X" $u] 00 00 00 00"
		}
	}
	set numused [llength $usedcmds]
	if {$numused == 0} return

	# split pre/post/cell commands
	set idxs [::main::get_selection_indexes]
	set idx1 [lindex $idxs 0]
	set idx2 [lindex $idxs 1]
	if {$idx1 == "-"} {
		bell
		return ""
	}
	set cmds [subst \$::[lindex $::editor_cmds_from 0]s([lindex $::editor_cmds_from 1])]
	set line 0
	set lines(pre)  ""
	set lines(post) ""
	set lines(cell) ""
	foreach c $cmds {
		if {[string range $c 0 0] == "*"} {
			set type [lindex $c 1]
			set numcmds($type) 0
		} else {
			incr numcmds($type)
			lappend lines($type) $c
		}
		if {$line == $idx2} {
			set instype $type
			set insln $numcmds($type)
		}
		incr line
	}

	set cmds $lines($instype)

	# renumber GOTO lines
	for {set idx 0} {$idx < [llength $cmds]} {incr idx} {
		set cmd [lindex $cmds $idx]
		set cmdchk "[string range $cmd 0 1][string range $cmd 4 4]"
		if {$cmdchk == "001" || $cmdchk == "003"} {
			set ln [expr 0x[lindex $cmd end]]
			if {$ln > $insln} {
				incr ln $numused
				lset cmd end [format "%02X" $ln]
				lset cmds $idx $cmd
			}
		}
	}

	# rebuild full PGC commands
	set cmds1 [lrange $cmds 0 [expr $insln-1]]
	set cmds2 [lrange $cmds $insln end]
	set lines($instype) [concat $cmds1 $usedcmds $cmds2]
	set ::[lindex $::editor_cmds_from 0]s([lindex $::editor_cmds_from 1]) \
			[concat {{* pre}} $lines(pre) {{* post}} $lines(post) {{* cell}} $lines(cell)]

	refresh_current_pgc
	::main::dvd_modified true false
}

#############################################################################
## Show current PGC as text
proc show_current_pgc_as_text {} {
	if {$::scratchbook_active} {
		set title "Scratchbook"
		set txt ""
	} else {
		if {! [info exists ::dvddir]} {
			bell
			return ""
		}
		set title "[$::widget(main_cmd_lab) cget -text] -[$::widget(main_cmd_lab2) cget -text]"
		set txt "$title\n\n"
	}
	foreach i $::editor_cmds_list {
		if $::config(dump_hex_codes) {
			set txt "$txt$i\n"
		} else {
			if {[string range $i 0 0] == "*"} {
				set txt "$txt$i\n"
			} else {
				set txt "$txt[string range $i 25 end]\n"
			}
		}
	}
	if [info exists ::dvddir] {
		set type [lindex $::editor_cmds_from 0]
		set index [lindex $::editor_cmds_from 1]

		if {$type == "menu" && !$::scratchbook_active} {
			set vts [lindex [split $index ","] 0]
			set txt "$txt\n[::menubuttons::dump_pgc_menu_buttons $vts \
					$::menus($index,numcells) $::menus($index,cellpostable) "" $::config(dump_hex_codes)]"
		}

		if {!$::scratchbook_active} {
			set pgcheader    [subst \$::${type}s($index,header)]
			set numprograms  [subst \$::${type}s($index,numprograms)]
			set programmap   [subst \$::${type}s($index,programmap)]
			set numcells     [subst \$::${type}s($index,numcells)]
			set cellpbtable  [subst \$::${type}s($index,cellpbtable)]
			set cellpostable [subst \$::${type}s($index,cellpostable)]
			if {$type == "pgc" && [info exists ::${type}s($index,pttlist)]} {
				set pttlist [subst \$::${type}s($index,pttlist)]
			} else {
				set pttlist ""
			}
			set vtsnum [lindex [split $index ","] 0]

			set txt "$txt\n\n[::pgced::dump_pgc \
					$type $index $pgcheader $numprograms $programmap \
					$numcells $cellpbtable $cellpostable $pttlist]"
		}
	}

	::main::showtext .curpgc $title $txt
}

proc explore_dvd {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	exec -- explorer.exe [file nativename $::dvddir] &
}

}	;# end of namespace ::macros
