#############################################################################
# toolbar.tcl Copyright 2004, 2005 r0lZ
# toolbar and balloon tooltips procs.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

namespace eval ::toolbar {

proc init_toolbar {mode} {

	if [winfo exists .mf.tb.toolbar] {destroy .mf.tb.toolbar}

	if {$mode == "toggle"} {
		if $::toolbar::toolbar_opened {
			set ::toolbar::toolbar_opened false
		} else {
			set ::toolbar::toolbar_opened true
		}
		set mode $::toolbar::toolbar_opened
	}

	if {! $mode} {
		.mf.tb configure -height 1
		return
	}

	set iconsdir "$::config(apppath)/lib/icons/sweetness"

	frame .mf.tb.toolbar -bd 2 -relief groove
	pack  .mf.tb.toolbar -anchor w -padx 2

	set icons {
		{open_dvd {::main::Read_New_DVD} "Open DVD:\nRead DVD IFOs from disk"}
		{save_dvd {::main::Write_DVD false} "Save DVD:\nSave DVD IFOs to disk"}
		{reopen_dvd {::main::ReRead_DVD} "Reopen DVD:\nUndo all edits since last save"}
		{new_dvd {::utils::new_DVD} "New DVD:\nCreate a new DVD basis"}
		{-}
		{create_backup {::main::Create_Backup false} "Create backup:\nCopy the IFOs in DVD folder\nto backup folder"}
		{restore_backup {::main::Restore_Backup} "Restore backup:\nRestore the IFOs from backup\nand reload the DVD"}
		{-}
		{open_scratchbook {::main::open_scratchbook} "Open scratchbook:\nOpen a previously\nsaved scratchbook"}
		{save_scratchbook {::main::save_scratchbook} "Save scratchbook:\nSave current scratchbook\nto file"}
		{-}
		{check_dvd_size {::writedvd::DVD_size} "Check DVD+/-R size:\nCheck if the current DVD\nfits on a DVD-5"}
		{-}
		{set_provider_id {::utils::set_provider_ID} "Set Provider ID:\nSet or change the provider ID"}
		{set_dvd_text {::utils::set_General_Name_GUI} "Set DVD-TEXT General Name:\nSet the DVD-TEXT General Name\nand Movie Name of the DVD"}
		{region_code {::utils::set_region_code} "Region code:\nChange or clear\nthe DVD region code"}
		{title_play_map_table {::srptinfo::srptinfo} "Title Play Map table:\nShow the DVD table of titles\n(and edit the table's PUOs)"}
		{-}
		{streams_attributes {::streams::set_streams_attribs .} "Domain streams attributes:\nShow/change the streams attributes\nof the current domain"}
		{-}
		{enable_all_operations {::macros::enable_all_operations} "Enable all operations:\nClear all PUOs and enable\nthe navigation buttons of\nthe current PGC"}
		{kill_pgc_playback {::macros::kill_pgc_playback} "Kill PGC playback:\nInhibit the playback\nof the current PGC"}
		{-}
		{new_pgc {::utils::new_pgc false false} "New PGC:\nCreate a new PGC\nin the current domain"}
		{-}
		{find_unused_gprms {::macros::find_unused_gprms false} "Find unused GPRMs:\nSearch all PGCs\nfor unused GPRMs"}
		{calculator_converter {::calc::calculator .calc} "Tcl calculator/converter:\nSimple Tcl calculator and\nhex/octal/language code\nconverter"}
		{-}
		{trace {if $::tr::tracemode {set ::tr::tracemode 0} else {set ::tr::tracemode 1}; ::tr::dvd_trace} "Trace mode:\nStep through commands,\nwatching register changes"}
	}

	if {$::tcl_platform(platform) == "windows"} {
		set icons [linsert $icons 4 {explore_dvd_folder {::macros::explore_dvd} "Explore DVD folder:\nOpen DVD folder in explorer"}]
		lappend icons {preview {::preview::preview_current_pgc_2} "Preview current PGC"}
	}
	if {$::tcl_platform(platform) != "macintosh"} {
		set icons [linsert $icons 23 {show_menu_buttons {::menubuttons::display_current_menu_buttons info} "Show menu buttons:\nShow the buttons of\nthe current menu PGC"}]
	}

	set padframe 0
	foreach icon $icons {
		if {$icon == "-"} {
			incr padframe
			pack [frame .mf.tb.toolbar.pad$padframe -width 2 -bd 1 -relief sunken] -fill y -side left
		} else {
			set imagename	[lindex $icon 0]
			set cmd			[lindex $icon 1]
			set tooltip		[lindex $icon 2]
			image create photo ::toolbar::$imagename -file [file join $iconsdir ${imagename}.gif]
			button .mf.tb.toolbar.$imagename -image ::toolbar::$imagename -background white \
					-bd 1 -relief flat -overrelief raised -command $cmd
			pack   .mf.tb.toolbar.$imagename -side left
			::balloon::set_balloon .mf.tb.toolbar.$imagename $tooltip
		}
	}
}

}

##############################################################################
# $Id: balloon.tcl,v 1.3 1997/04/18 02:00:38 stewart Exp $
#
# balloon.tcl - procedures used by balloon help
#
# Copyright (C) 1996-1997 Stewart Allen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##############################################################################

namespace eval ::balloon {
namespace export set_balloon

bind ::balloon::tt(balloon) <Enter> {
	set ::balloon::tt(balloon,set) 0
	set ::balloon::tt(balloon,first) 1
	set ::balloon::tt(balloon,id) [after 500 {catch {::balloon::balloon %W $::balloon::tt(balloon,%W)}}]
}

bind ::balloon::tt(balloon) <Button> {
	set ::balloon::tt(balloon,first) 0
	::balloon::kill_balloon
}

bind ::balloon::tt(balloon) <Leave> {
	set ::balloon::tt(balloon,first) 0
	::balloon::kill_balloon
}

bind ::balloon::tt(balloon) <Motion> {
	if {$::balloon::tt(balloon,set) == 0} {
		after cancel $::balloon::tt(balloon,id)
		set ::balloon::tt(balloon,id) [after 500 {catch {::balloon::balloon %W $::balloon::tt(balloon,%W)}}]
	}
}

proc set_balloon {target message} {
	set ::balloon::tt(balloon,$target) $message
	bindtags $target "[bindtags $target] ::balloon::tt(balloon)"
}

proc kill_balloon {} {
	catch { after cancel $::balloon::tt(balloon,id) }
	if {[winfo exists .balloon] == 1} {
		destroy .balloon
	}
	set ::balloon::tt(balloon,set) 0
}

proc balloon {target message} {
	if {$::balloon::tt(balloon,first) == 1} {
		set ::balloon::tt(balloon,first) 2
		set x [winfo rootx $target]
		set y [expr [winfo rooty $target]+[winfo height $target]]
		toplevel .balloon -bg black
		wm overrideredirect .balloon 1
		if {$::tcl_platform(platform) == "windows"} {
			label .balloon.l \
				-text $message -relief flat -bg SystemInfoBackground \
				-fg SystemInfoText -padx 2 -pady 0 -anchor w -justify left
		} else {
			label .balloon.l \
				-text $message -relief flat -bg #FFFFC0 \
				-fg #000000 -padx 2 -pady 0 -anchor w -justify left
		}
		pack .balloon.l -side left -padx 1 -pady 1
		wm geometry .balloon +${x}+${y}
		set ::balloon::tt(balloon,set) 1
	}
}

}

