#############################################################################
# utils.tcl Copyright 2004, 2005 r0lZ
# Procedures related to the Utilities menu.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

namespace eval ::utils {
namespace export utils_menu
namespace import ::readdvd::*

#############################################################################
## utils_menu
## build the Utilities menu
proc utils_menu {menuname} {
	menu $menuname
	$menuname add command -label "Title Play Map table" -command ::srptinfo::srptinfo
	$menuname add command -label "Domain Streams Attributes" -command {::streams::set_streams_attribs .}
	$menuname add separator
	$menuname add command -label "Region Code" -command {::utils::set_region_code}
	$menuname add command -label "Filter Prohibited User Operations" -command {::utils::filter_puos}
	$menuname add command -label "Enable Navigation Buttons" -command {::macros::nav_buttons}
	$menuname add command -label "Remove Parental Managment" -command {::macros::remove_parental_managment}
	$menuname add separator
	$menuname add command -label "Blank out all PGCs in Domain" -command {::utils::blank_out_PGCS_GUI}
	$menuname add command -label "New Menu" -command {::utils::new_menu false}
	$menuname add command -label "Import Menu" -command {::utils::import_menu}
	$menuname add command -label "Remove Menu" -command {::utils::remove_menu}
	$menuname add command -label "Set Menu Type" -command {::utils::set_menu_type_GUI}
	if {$::tcl_platform(platform) != "macintosh"} {
		$menuname add command -label "Show Menu Buttons" \
				-command {::menubuttons::display_current_menu_buttons info} -accelerator "Ctrl+M"
		bind . <Control-m> {::menubuttons::display_current_menu_buttons info}
	}
	$menuname add command -label "Swap menu LUs" -command {::utils::swap_lus}
	$menuname add command -label "Copy LU cmds to other LUs" -command {::utils::copy_lu_to_lus}
	$menuname add command -label "New dummy PGC" -command {::utils::new_pgc false false}
	$menuname add command -label "Clone current PGC" -command {::utils::clone_pgc}
	$menuname add command -label "New 'Play All' Title" -command {::pgced::create_playall_pgc}
	$menuname add separator
	$menuname add command -label "Set Provider ID" -command ::utils::set_provider_ID
	$menuname add command -label "Set DVD-TEXT General Name" -command ::utils::set_General_Name_GUI
	$menuname add command -label "Remove DVD-TEXT data" -command ::utils::remove_dvd_text_data
}

#############################################################################
## build the PGC Selector's context sensitive popup menus
proc selector_popup_menu_fp {menuname} {
	if [winfo exists $menuname] {destroy $menuname}
	menu $menuname
	if $::config(tearoff_menus) {set itemnum 1} else {set itemnum 0}
	set ::main::last_fp_menu_selected $itemnum
	$menuname add command -label "Edit PGC" \
			-command "set ::main::last_fp_menu_selected $itemnum; ::main::action_Edit_PGC"
	incr itemnum
	$menuname add command -label "Info: PGC" \
			-command "set ::main::last_fp_menu_selected $itemnum; ::macros::show_current_pgc_as_text"
	incr itemnum
	$menuname add command -label "Go to calling command" \
			-command "set ::main::last_fp_menu_selected $itemnum; ::utils::pgc_crossrefs"
	incr itemnum
	$menuname add separator
	incr itemnum
	$menuname add command -label "New Menu" \
			-command "set ::main::last_fp_menu_selected $itemnum; ::utils::new_menu false"
	incr itemnum
	$menuname add command -label "Import Menu" \
			-command "set ::main::last_fp_menu_selected $itemnum; ::utils::import_menu"
	incr itemnum
	$menuname add separator
	incr itemnum
	$menuname add command -label "Import First-Play Clip" \
			-command "set ::main::last_fp_menu_selected $itemnum; ::macros::import_fp_clip"
	incr itemnum
	$menuname add separator
	incr itemnum
	$menuname add command -label "Toggle PGC pre/post breakpoint" \
			-command "set ::main::last_fp_menu_selected $itemnum; ::tr::toggle_breakpoint prepost"
	incr itemnum
	$menuname add command -label "Toggle PGC playback breakpoint" \
			-command "set ::main::last_fp_menu_selected $itemnum; ::tr::toggle_breakpoint playback"
}
proc selector_popup_menu_menus {menuname} {
	if [winfo exists $menuname] {destroy $menuname}
	menu $menuname
	if $::config(tearoff_menus) {set itemnum 1} else {set itemnum 0}
	set ::main::last_menus_menu_selected $itemnum

	if {$::tcl_platform(platform) == "windows"} {
		$menuname add command -label "Preview PGC" -command \
				"set ::main::last_menus_menu_selected $itemnum; ::preview::preview_current_pgc_2"
		incr itemnum
		$menuname add separator
		incr itemnum
	}
	$menuname add command -label "Edit PGC" \
			-command "set ::main::last_menus_menu_selected $itemnum; ::main::action_Edit_PGC"
	incr itemnum
	$menuname add command -label "Info: PGC" \
			-command "set ::main::last_menus_menu_selected $itemnum; ::macros::show_current_pgc_as_text"
	incr itemnum
	$menuname add command -label "Go to calling command" \
			-command "set ::main::last_menus_menu_selected $itemnum; ::utils::pgc_crossrefs"
	incr itemnum
	$menuname add separator
	incr itemnum
	$menuname add command -label "Domain Streams Attributes" \
			-command "set ::main::last_menus_menu_selected $itemnum; ::streams::set_streams_attribs ."
	incr itemnum
	$menuname add separator
	incr itemnum
	$menuname add command -label "Blank out all Menu PGCs" \
			-command "set ::main::last_menus_menu_selected $itemnum; ::utils::blank_out_PGCS_GUI"
	incr itemnum
	$menuname add command -label "New Menu" \
			-command "set ::main::last_menus_menu_selected $itemnum; ::utils::new_menu false"
	incr itemnum
	$menuname add command -label "Import Menu" \
			-command "set ::main::last_menus_menu_selected $itemnum; ::utils::import_menu"
	incr itemnum
	$menuname add command -label "Remove Menu" \
			-command "set ::main::last_menus_menu_selected $itemnum; ::utils::remove_menu"
	incr itemnum
	$menuname add command -label "Set Menu Type" \
			-command "set ::main::last_menus_menu_selected $itemnum; ::utils::set_menu_type_GUI"
	if {$::tcl_platform(platform) != "macintosh"} {
		incr itemnum
		$menuname add command -label "Show Menu Buttons" \
				-command "set ::main::last_menus_menu_selected $itemnum; ::menubuttons::display_current_menu_buttons info"
	}
	incr itemnum
	$menuname add command -label "Swap menu LUs" \
			-command "set ::main::last_menus_menu_selected $itemnum; ::utils::swap_lus"
	incr itemnum
	$menuname add command -label "Copy LU cmds to other LUs" \
			-command "set ::main::last_menus_menu_selected $itemnum; ::utils::copy_lu_to_lus"
	incr itemnum
	$menuname add command -label "New dummy PGC" \
			-command "set ::main::last_menus_menu_selected $itemnum; ::utils::new_pgc false false"
	incr itemnum
	$menuname add command -label "Clone PGC" \
			-command "set ::main::last_menus_menu_selected $itemnum; ::utils::clone_pgc"
	incr itemnum
	$menuname add separator
	incr itemnum
	$menuname add command -label "Kill PGC Playback" \
			-command "set ::main::last_menus_menu_selected $itemnum; ::macros::kill_pgc_playback"
	incr itemnum
	$menuname add command -label "Enable all operations" \
			-command "set ::main::last_menus_menu_selected $itemnum; ::macros::enable_all_operations"
	incr itemnum
	$menuname add separator
	incr itemnum
	$menuname add command -label "Toggle PGC pre/post breakpoint" \
			-command "set ::main::last_menus_menu_selected $itemnum; ::tr::toggle_breakpoint prepost"
	incr itemnum
	$menuname add command -label "Toggle PGC playback breakpoint" \
			-command "set ::main::last_menus_menu_selected $itemnum; ::tr::toggle_breakpoint playback"
}
proc selector_popup_menu_titles {menuname} {
	if [winfo exists $menuname] {destroy $menuname}
	menu $menuname
	if $::config(tearoff_menus) {set itemnum 1} else {set itemnum 0}
	set ::main::last_titles_menu_selected $itemnum

	if {$::tcl_platform(platform) == "windows"} {
		$menuname add command -label "Preview PGC" -command \
				"set ::main::last_titles_menu_selected $itemnum; ::preview::preview_current_pgc_2"
		incr itemnum
		$menuname add separator
		incr itemnum
	}
	$menuname add command -label "Edit PGC" \
			-command "set ::main::last_titles_menu_selected $itemnum; ::main::action_Edit_PGC"
	incr itemnum
	$menuname add command -label "Info: PGC" \
			-command "set ::main::last_titles_menu_selected $itemnum; ::macros::show_current_pgc_as_text"
	incr itemnum
	$menuname add command -label "Go to calling command" \
			-command "set ::main::last_titles_menu_selected $itemnum; ::utils::pgc_crossrefs"
	incr itemnum
	$menuname add separator
	incr itemnum
	$menuname add command -label "Domain Streams Attributes" \
			-command "set ::main::last_titles_menu_selected $itemnum; ::streams::set_streams_attribs ."
	incr itemnum
	$menuname add separator
	incr itemnum
	$menuname add command -label "Blank out all VTST Titles" \
			-command "set ::main::last_titles_menu_selected $itemnum; ::utils::blank_out_PGCS_GUI"
	incr itemnum
	$menuname add command -label "New Menu" \
			-command "set ::main::last_titles_menu_selected $itemnum; ::utils::new_menu false"
	incr itemnum
	$menuname add command -label "Import Menu" \
			-command "set ::main::last_titles_menu_selected $itemnum; ::utils::import_menu"
	incr itemnum
	$menuname add command -label "New dummy PGC" \
			-command "set ::main::last_titles_menu_selected $itemnum; ::utils::new_pgc false false"
	incr itemnum
	$menuname add command -label "Clone PGC" \
			-command "set ::main::last_titles_menu_selected $itemnum; ::utils::clone_pgc"
	incr itemnum
	$menuname add separator
	incr itemnum
	$menuname add command -label "Kill PGC Playback" \
			-command "set ::main::last_titles_menu_selected $itemnum; ::macros::kill_pgc_playback"
	incr itemnum
	$menuname add command -label "Enable all operations" \
			-command "set ::main::last_titles_menu_selected $itemnum; ::macros::enable_all_operations"
	incr itemnum
	$menuname add separator
	incr itemnum
	$menuname add command -label "Import Title's Intro Clip" \
			-command "set ::main::last_titles_menu_selected $itemnum; ::macros::import_title_intro_clip"
	incr itemnum
	$menuname add command -label "Import Title's Closing Clip" \
			-command "set ::main::last_titles_menu_selected $itemnum; ::macros::import_title_closing_clip"
	incr itemnum
	$menuname add command -label "New 'Play All' Title" \
			-command "set ::main::last_titles_menu_selected $itemnum; ::pgced::create_playall_pgc"
	incr itemnum
	$menuname add separator
	incr itemnum
	$menuname add command -label "Toggle PGC pre/post breakpoint" \
			-command "set ::main::last_titles_menu_selected $itemnum; ::tr::toggle_breakpoint prepost"
	incr itemnum
	$menuname add command -label "Toggle PGC playback breakpoint" \
			-command "set ::main::last_titles_menu_selected $itemnum; ::tr::toggle_breakpoint playback"
}

proc selector_popup_menu_notpgc {menuname} {
	if [winfo exists $menuname] {destroy $menuname}
	menu $menuname
	if $::config(tearoff_menus) {set itemnum 1} else {set itemnum 0}
	set ::main::last_notpgc_menu_selected $itemnum
	$menuname add command -label "Clear 'PGC played' highlights" \
			-command "set ::main::last_notpgc_menu_selected $itemnum; ::main::init_tree_menu_no_move"
	incr itemnum
	$menuname add command -label "Hide PGC Selector pane" \
			-command {::main::selector toggle ::main::labelslist ::main::cmdslist}
	incr itemnum
}


#############################################################################
## utility procedures

#############################################################################
## Change the DVD Region Code
proc set_region_code {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}

	set regions {
		{USA, Canada, U.S. Territories}
		{Europe, Japan, South Africa, Middle East}
		{Southeast Asia and East Asia (including Hong Kong)}
		{Australia, New Zealand, Latin America, Pacific Islands}
		{Eastern Europe, Russia, India, Africa, North Korea, Mongolia}
		{China}
		{reserved}
		{Special international venues (airplanes, cruise ships...)}
	}

	set code [expr [get_int ::raw_sectors(0,VMGM_MAT) 0X23 1] ^ 0XFF]

	# init Search flag
	if {![info exists ::utils::regcode_search]} {
		set ::utils::regcode_search 1
	}

	if [winfo exists .regioncode] {destroy .regioncode}
	set w .regioncode
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w "Region Code"
	wm transient $w .
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	wm protocol $w WM_DELETE_WINDOW {set ::utils::regioncode_done 0}
	bind $w <Escape> "set ::utils::regioncode_done 0"

	frame $w.f -bd 2 -relief groove
	pack  $w.f -padx 4 -pady 3 -fill x
	label $w.f.lab -text "Allowed regions:"
	pack  $w.f.lab -anchor w
	for {set i 0} {$i < 8} {incr i} {
		set bitmask [expr round(pow(2,$i))]
		set ::utils::regcode_$i [expr $code & $bitmask]
		checkbutton $w.f.c$i -text "[expr $i+1] (val:$bitmask) [lindex $regions $i]" \
				-variable ::utils::regcode_$i -onvalue [expr $bitmask] -bd $::config(min_bd) -anchor w
		pack $w.f.c$i -fill x -anchor w
	}
	button $w.f.cl -text "  Region Free  " \
		-command {for {set i 0} {$i < 8} {incr i} {set ::utils::regcode_$i [expr round(pow(2,$i))]}}
	pack   $w.f.cl -padx 2 -pady 4 -anchor w

	frame $w.f2 -bd 2 -relief groove
	pack  $w.f2 -padx 4 -pady 3 -fill x

	set txt "The region code of your player may be tested by programmation."
	set txt "$txt\nIf it doesn't match the DVD zone, a jump to a PGC that will inhibit the"
	set txt "$txt\nplayback or return to the menu is usually done."
	set txt "$txt\nYou may try to replace the sprm(20) reference(s) by the constant value"
	set txt "$txt\n(between parenthesis above) matching the original DVD region code,"
	set txt "$txt\nor replace the command by NOP, or modify the goto or the jump target."
	label $w.f2.lab -text $txt -anchor w -justify left
	pack  $w.f2.lab -pady 2 -fill x -anchor w

	set txt "Search for \"sprm(20:player region code)\" in the DVD commands?"
	checkbutton $w.f2.s -text $txt -variable ::utils::regcode_search -anchor w -bd $::config(min_bd)
	pack  $w.f2.s -pady 2 -fill x

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -anchor e
	button $w.fb.ok -text "OK" -width 10 -command {set ::utils::regioncode_done 1}
	pack   $w.fb.ok -side left -padx 2
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::utils::regioncode_done 0}
	pack   $w.fb.cancel -side left -padx 2

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr [winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)]
	set wy [expr [winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	wm geometry $w +$wx+$wy
	wm deiconify $w
	grab set $w

	tkwait variable ::utils::regioncode_done

	if {$::utils::regioncode_done} {
		set code 0
		for {set i 0} {$i < 8} {incr i} {
			set bitmask [expr round(pow(2,$i))]
			set code [expr $code | [subst \$::utils::regcode_$i]]
		}
		set_int ::raw_sectors(0,VMGM_MAT) 0X23 [expr $code ^ 0XFF] 1
		if $::utils::regcode_search {
			set oldsearchstring $::main::searchstring
			set oldsearchcase   $::main::searchcase
			set oldsearchregexp $::main::searchregexp
			set oldcmds $::editor_cmds_from
			set oldselection [::main::get_selection_indexes]

			set ::main::searchstring  "sprm(20:player region code)"
			set ::main::searchcase    1
			set ::main::searchregexp  0
			::main::display_pgc pgc 0 0
			set r [::main::find_in_dvd -forwards]
			if {$r == 0} {
				set ::main::searchstring $oldsearchstring
				set ::main::searchcase   $oldsearchcase
				set ::main::searchregexp $oldsearchregexp
				set args [split [lindex $oldcmds 1] ","]
				::main::display_pgc [lindex $oldcmds 0] \
						[lindex $args 0] [lindex $args 1] [lindex $args 2]
				::main::set_listbox_selection [lindex $oldselection 0] [lindex $oldselection 1]
				show_msg "No \"sprm(20:player region code)\" found in the DVD commands."
			} else {
				show_msg "\"sprm(20:player region code)\" found!"
			}
		}
	}

	grab release $w
	destroy $w

	return
}

#############################################################################
## Filter prohibited user operations
proc filter_puos {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set utiltitle "Filter Prohibited User Operations"

	if {!$::config(expert_mode)} {
		set txt "This utility will selectively clear the PUOs"
		set txt "$txt\nin all PGCs of the VMGM Menus,"
		set txt "$txt\nin all PGCs of the VTSM Menus,"
		set txt "$txt\nin all PGCs of the VTST Titles,"
		set txt "$txt\nand in the VMG Title Play Map Table."
		set txt "$txt\n"
		set txt "$txt\nContinue ?"
		set r [tk_messageBox -title $utiltitle -icon question -message $txt \
				-parent . -type yesno -default no]
		if {$r == "no"} return
	}

	# VMG Menus PUOs
	set puos 0
	for {set lu 1} {$lu <= $::menus(0,numlus)} {incr lu} {
		for {set pgc 1} {$pgc <= $::menus(0,$lu,numpgcs)} {incr pgc} {
			set puos [expr $puos | [get_int ::menus(0,$lu,$pgc,header) 8 4]]
		}
	}
	if {$puos == 0} {
		show_warning -title $utiltitle -icon info -parent . \
				-message "No Prohibited User Operations found in the VMGM Menus."
	} else {
		set ::pgced::puos $puos
		set filterpuos [::pgced::puos_editor "Filter PUOs for all VMGM Menus :\nClear the PUOs you want to clear,\nSet the PUOs you don't want to change." .]
		if {$filterpuos != $puos} {
			::main::dvd_modified true false
			for {set lu 1} {$lu <= $::menus(0,numlus)} {incr lu} {
				for {set pgc 1} {$pgc <= $::menus(0,$lu,numpgcs)} {incr pgc} {
					set_int ::menus(0,$lu,$pgc,header) 8 \
							[expr $filterpuos & [get_int ::menus(0,$lu,$pgc,header) 8 4]] 4
				}
			}
		}
	}

	# VTS Menus PUOs
	set puos 0
	for {set vts 1} {$vts <= $::menus(numvts)} {incr vts} {
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				set puos [expr $puos | [get_int ::menus($vts,$lu,$pgc,header) 8 4]]
			}
		}
	}
	if {$puos == 0} {
		show_warning -title $utiltitle -icon info -parent . \
				-message "No Prohibited User Operations found in the VTSM Menus."
	} else {
		set ::pgced::puos $puos
		set filterpuos [::pgced::puos_editor "Filter PUOs for all VTSM Menus :\nClear the PUOs you want to clear,\nSet the PUOs you don't want to change." .]
		if {$filterpuos != $puos} {
			::main::dvd_modified true false
			for {set vts 1} {$vts <= $::menus(numvts)} {incr vts} {
				for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
					for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
						set_int ::menus($vts,$lu,$pgc,header) 8 \
								[expr $filterpuos & [get_int ::menus($vts,$lu,$pgc,header) 8 4]] 4
					}
				}
			}
		}
	}

	# Titles PUOs
	set puos 0
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		for {set ttn 1} {$ttn <= $::pgcs($vts,numpgcs)} {incr ttn} {
			set puos [expr $puos | [get_int ::pgcs($vts,$ttn,header) 8 4]]
		}
	}
	if {$puos == 0} {
		show_warning -title $utiltitle -icon info -parent . \
				-message "No Prohibited User Operations found in the VTST Title's PGCs."
	} else {
		set ::pgced::puos $puos
		set filterpuos [::pgced::puos_editor "Filter PUOs for all VTST Titles :\nClear the PUOs you want to clear,\nSet the PUOs you don't want to change." .]
		if {$filterpuos != $puos} {
			::main::dvd_modified true false
			for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
				for {set ttn 1} {$ttn <= $::pgcs($vts,numpgcs)} {incr ttn} {
					set_int ::pgcs($vts,$ttn,header) 8 \
							[expr $filterpuos & [get_int ::pgcs($vts,$ttn,header) 8 4]] 4
				}
			}
		}
	}

	# VMG_TT_SRPT PUOs
	set numtitles [get_int ::raw_sectors(0,VMG_TT_SRPT) 0 2]
	set puos 0
	for {set t 1} {$t <= $numtitles} {incr t} {
		set t_idx [expr 8+(($t-1)*12)]
		set puos [expr $puos | ([get_int ::raw_sectors(0,VMG_TT_SRPT) $t_idx 1] & 3)]
	}
	if {$puos == 0} {
		show_warning -title $utiltitle -icon info -parent . \
				-message "No Prohibited User Operations found in the Title Play Map Table."
	} else {
		set filterpuos $puos
		if {[expr $puos & 1]} {
			set txt "At least one Title has the \"Time Play or Search\" prohibited"
			set txt "$txt\nin the VMG Title Play Map Table."
			set txt "$txt\n"
			set txt "$txt\nClear this prohibition for all Titles ?"
			set r [tk_messageBox -title $utiltitle -icon question -message $txt \
					-parent . -type yesno -default no]
			if {$r == "yes"} {set filterpuos [expr $filterpuos & 2]}
		}
		if {[expr $puos & 2]} {
			set txt "At least one Title has the \"PTT Play or Search\" prohibited"
			set txt "$txt\nin the VMG Title Play Map Table."
			set txt "$txt\n"
			set txt "$txt\nClear this prohibition for all Titles ?"
			set r [tk_messageBox -title $utiltitle -icon question -message $txt \
					-parent . -type yesno -default no]
			if {$r == "yes"} {set filterpuos [expr $filterpuos & 1]}
		}
		if {$puos != $filterpuos} {
			::main::dvd_modified true false
			set mask [expr $filterpuos | 0XFC]
			for {set t 1} {$t <= $numtitles} {incr t} {
				set t_idx [expr 8+(($t-1)*12)]
				set byte [expr [get_int ::raw_sectors(0,VMG_TT_SRPT) $t_idx 1] & $mask]
				set_int ::raw_sectors(0,VMG_TT_SRPT) $t_idx $byte 1
			}
		}
	}
}

#############################################################################
## Import VTS Titles
proc import_VTST {} {
	if {! [info exists ::dvddir]} {
		bell
		return 0
	}
	set utiltitle "Import VTST"

	set sourceifofile [tk_getOpenFile -initialdir $::config(importmenuinitialdir) \
		-filetypes {{"VTS IFO files" {*_0.IFO}}} \
		-title "Select the VTS IFO file with the title(s) to import"]
	if {$sourceifofile == ""} {return 0}
	if {[file tail $sourceifofile] == "VIDEO_TS.IFO"} {
		show_warning -title $utiltitle -icon error -parent . \
				-message "You must select a VTS IFO (VTS_**_*.IFO)"
		return 0
	}
	set sourceifofile [file nativename $sourceifofile]
	set ::config(importmenuinitialdir) [file dirname $sourceifofile]

	# read the VTS
	set sourcevtsnum [readdvdfile $sourceifofile ::imported_sectors ::imported_pgcs ::imported_menus ::imported_ttns]
	if {$sourcevtsnum == -1} {return 0}
	if {$sourcevtsnum == 0} {
		show_warning -title $utiltitle -icon error -parent . \
				-message "You must select a VTS IFO (VTS_**_*.IFO)"
		# cleanup
		unset ::imported_sectors ::imported_pgcs ::imported_menus
		return 0
	}

	# How many TTNs are in the imported VTS?
	set num_new_ttns [get_int ::imported_sectors($sourcevtsnum,VTS_PTT_SRPT) 0 2]
	if {$num_new_ttns == 0} {
		show_warning -title $utiltitle -icon error -parent . \
				-message "Error: there are no titles defined in VTS file\n\"$sourceifofile\""
		# cleanup
		unset ::imported_sectors ::imported_pgcs ::imported_menus
		return 0
	}

	# read the source VIDEO_TS.IFO to get the VMG_TT_SRPT table.
	set vmgfile [file join [file dirname $sourceifofile] "VIDEO_TS.IFO"]
	set src_vmg_tt_srpt ""
	if [file exists $vmgfile] {
		if {[readdvdfile $vmgfile ::ivmg_sectors ::ivmg_pgcs ::ivmg_menus ::ivmg_ttns] == 0} {
			set src_vmg_tt_srpt $::ivmg_sectors(0,VMG_TT_SRPT)
			unset ::ivmg_sectors ::ivmg_pgcs ::ivmg_menus
		}
	}
	if {$src_vmg_tt_srpt == ""} {
		set txt "WARNING:\nCan't find or load the source VMG (VIDEO_TS.IFO)!\n"
		set txt "$txt\nThe VMG Table of Titles will be"
		set txt "$txt\ninitialized with default values."
		set r [tk_messageBox -title $utiltitle -icon warning -parent . -type okcancel \
				-default cancel -message $txt]
		if {$r == "cancel"} {
			# cleanup
			unset ::imported_sectors ::imported_pgcs ::imported_menus
			return 0
		}
	}

	### Go modify the DVD!
	# init the new VTS sectors
	set newvtsnum [expr $::pgcs(numvts) + 1]
	::readdvd::parse_VTS_PTT_SRPT ::imported_sectors($sourcevtsnum,VTS_PTT_SRPT) $newvtsnum ::imported_pgcs ::imported_ttns
	set ::raw_sectors($newvtsnum,VTSI_MAT) \
			$::imported_sectors($sourcevtsnum,VTSI_MAT)
	set ::raw_sectors($newvtsnum,VTS_PTT_SRPT) \
			$::imported_sectors($sourcevtsnum,VTS_PTT_SRPT)
	set ::raw_sectors($newvtsnum,VTS_PGCI) \
			$::imported_sectors($sourcevtsnum,VTS_PGCI)
	set ::raw_sectors($newvtsnum,VTS_TMAPTI) \
			$::imported_sectors($sourcevtsnum,VTS_TMAPTI)
	set ::raw_sectors($newvtsnum,VTS_C_ADT) \
			$::imported_sectors($sourcevtsnum,VTS_C_ADT)
	set ::raw_sectors($newvtsnum,VTS_VOBU_ADMAP) \
			$::imported_sectors($sourcevtsnum,VTS_VOBU_ADMAP)

	# clear the menu sectors
	set ::raw_sectors($newvtsnum,VTSM_PGCI_UT) ""
	set ::raw_sectors($newvtsnum,VTSM_C_ADT) ""
	set ::raw_sectors($newvtsnum,VTSM_VOBU_ADMAP) ""

	# cleanup
	unset ::imported_sectors ::imported_pgcs ::imported_menus

	# Init variables and parse the new TTNs
	set ::pgcs(numvts)  $newvtsnum
	set ::menus(numvts) $newvtsnum
	set_int ::raw_sectors(0,VMGM_MAT) 62 $newvtsnum 2
	set ::menus($newvtsnum,numlus) 0
	::readdvd::parse_VTS_PGCI ::raw_sectors($newvtsnum,VTS_PGCI) $newvtsnum ::pgcs
	::readdvd::parse_VTS_PTT_SRPT ::raw_sectors($newvtsnum,VTS_PTT_SRPT) $newvtsnum ::pgcs ::ttns

	# Add new Titles to the TT_SRPT Title Table.
	# If the original table was read, copy the relevent entries,
	# otherwise, set default values (except for PTTs value)
	if {$src_vmg_tt_srpt != ""} {
		set data $::raw_sectors(0,VMG_TT_SRPT)
		set old_num_titles [get_int data 0 2]
		set src_num_titles [get_int src_vmg_tt_srpt 0 2]
		set num_new_entries 0
		for {set t 1} {$t <= $src_num_titles} {incr t} {
			set startidx [expr ($t * 12) - 4]
			set endidx   [expr $startidx + 11]
			if {$endidx > [string length $src_vmg_tt_srpt]} break
			set entry [string range $src_vmg_tt_srpt $startidx $endidx]
			set vtsn [get_int entry 6 1]
			if {$vtsn == $sourcevtsnum} {
				set_int entry 6 $newvtsnum 1
				set data "$data$entry"
				incr num_new_entries
			}
		}
		set new_num_titles [expr $old_num_titles + $num_new_entries]
		set_int data 0 $new_num_titles 2
		set_int data 4 [expr [string length $data] -1] 4
		set ::raw_sectors(0,VMG_TT_SRPT) $data
	} else {
		set data $::raw_sectors(0,VMG_TT_SRPT)
		set old_num_titles [get_int data 0 2]
		set new_num_titles [expr $old_num_titles + $num_new_ttns]
		set_int data 0 $new_num_titles 2
		set vtssrpt $::raw_sectors($newvtsnum,VTS_PTT_SRPT)
		for {set t 1} {$t <= $num_new_ttns} {incr t} {
			# TODO: Set real values instead of Type=Sequential, Angles=1, parental management=0
			# get number of PTTs from the VTS_PTT_SRPT table.
			set entry "\074\001\000\000\000\000\000\000\000\000\000\000"
			# set number of chapters from the VTS_PTT_SRPT
			if {$vtssrpt != ""} {
				set numttus [get_int vtssrpt 0 2]
				set idx [get_int vtssrpt [expr 4 + ($t * 4)] 4]
				set pgcn [get_int vtssrpt $idx 2]
				set num_ptts 0
				while {$pgcn == $t} {
					incr num_ptts
					incr idx 4
					if {$idx >= [string length $vtssrpt]} break
					set pgcn [get_int vtssrpt $idx 2]
				}
			}
			set_int entry 2 $num_ptts 2
			set_int entry 6 $newvtsnum 1
			set_int entry 7 $t 1
			set data "$data$entry"
		}
		set_int data 4 [expr [string length $data] -1] 4
		set ::raw_sectors(0,VMG_TT_SRPT) $data
	}

	# get the table of titles from the VMG_TT_SRPT again
	# (to be able to display the title number in the PGC selectors)
	for {set t 1} {$t <= $new_num_titles} {incr t} {
		set t_idx [expr 14+(($t-1)*12)]
		set ::pgcs([get_int data $t_idx 1],$::ttns([get_int data $t_idx 1],[get_int data [expr $t_idx+1] 1]),Title) $t
	}

	# Reinit GUI
	::main::init_tree_menu
	::main::display_pgc pgc $newvtsnum 1
	::main::dvd_modified true false

	# copy or move the TTN VOBs, if user wants so.
	set destvobbasename [file join $::dvddir "VTS_[format %02i $newvtsnum]_"]
	set sourcevobbasename [string trimright [file rootname $sourceifofile] "0"]
	if {$destvobbasename != $sourcevobbasename} {
		set numvobs 0
		set vobsize 0
		for {set i 1} {$i < 10} {incr i} {
			if {! [file exists "${sourcevobbasename}$i.VOB"]} break
			set numvobs $i
			set vobsize [expr $vobsize+([file size "${sourcevobbasename}$i.VOB"]/1024.0)]
		}
		set vobkbsize [expr round($vobsize*100)/100.0]
		set vobmbsize [expr round(($vobkbsize/1024.0)*100)/100.0]
		set vobgbsize [expr round(($vobmbsize/1024.0)*100)/100.0]
		if {$vobmbsize > 1} {
			if {$vobgbsize > 1} {
				set sizemsg "($vobgbsize GB.)"
			} else {
				set sizemsg "($vobmbsize MB.)"
			}
		} else {
			set sizemsg "($vobkbsize KB.)"
		}
		if {$numvobs > 0} {
			if {$numvobs > 1} {
				set pluriel "s"
				set vobfilesmsg "$numvobs TTN VOB files"
			} else {
				set pluriel ""
				set vobfilesmsg "TTN VOB file"
			}
			set txt "New VTS $newvtsnum ($num_new_ttns title"
			if {$num_new_ttns > 1} {set txt "${txt}s"}
			set txt "$txt) imported OK.\n"
			set txt "$txt\nImport the $vobfilesmsg $sizemsg ?"
			if {[file exists ${destvobbasename}1.VOB] || [file exists ${destvobbasename}0.IFO]} {
				set txt "$txt\n\nWARNING: There are already some VTS $newvtsnum"
				set txt "$txt\nIFO/VOB files in your DVD folder."
				set txt "$txt\nIF YOU CONTINUE, THEY WILL BE DELETED FIRST!"
			}
			set r [tk_dialog .confirm "PgcEdit" $txt "question" 0 \
					"  Copy VOB$pluriel  " "  Move VOB$pluriel  " "Don't import VOB$pluriel"]
			if {$r != 2} {
				for {set i 1} {$i < 10} {incr i} {
					if [file exists "${destvobbasename}$i.VOB"] {
						file delete "${destvobbasename}$i.VOB"
					}
				}
			}
			update idletasks
			switch $r {
				0 {	# copy VOBs
					for {set i 1} {$i <= $numvobs} {incr i} {
						show_msg "Copying VOB $i/$numvobs.  Please wait."
						update idletasks
						file copy "${sourcevobbasename}$i.VOB" "${destvobbasename}$i.VOB"
					}
				}
				1 {	# move VOBs
					show_msg "Moving VOBs."
					for {set i 1} {$i <= $numvobs} {incr i} {
						file rename "${sourcevobbasename}$i.VOB" "${destvobbasename}$i.VOB"
					}
				}
			}
		}
	}
	show_msg ""
	return [expr $old_num_titles+1]
}


#############################################################################
## Blank out all VTST Titles: GUI
proc blank_out_PGCS_GUI {} {
	set utiltitle "Blank out all PGCs of current domain"
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set type   [lindex $::editor_cmds_from 0]
	if {$type == "menu"} {
		set type "Menus"
		set vtsi_mat_idx 0x100
	} else {
		set type "Titles"
		set vtsi_mat_idx 0x200
	}
	set vtsnum [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
	if {$vtsnum == 0} {
		set vtsi_table_name "VMGM_MAT"
	} else {
		set vtsi_table_name "VTSI_MAT"
	}
	if {$vtsnum == 0 && $type == "Titles"} {bell; return}
	if {[expr [get_int ::raw_sectors($vtsnum,$vtsi_table_name) $vtsi_mat_idx 2] & 0x3000] == 0} {
		set standard "ntsc"
	} else {
		set standard "pal"
	}

	# check if the menu has video.  If not, abort!
	if {[lindex $::editor_cmds_from 0] == "menu"} {
		set hasvideo false
		for {set lu 1} {$lu <= $::menus($vtsnum,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vtsnum,$lu,numpgcs)} {incr pgc} {
				if {$::menus($vtsnum,$lu,$pgc,numcells)>0} {
					set hasvideo true
					break
				}
			}
		}
		if {!$hasvideo} {
			set r [tk_messageBox -title $utiltitle -icon question -parent . -type yesno -default yes \
					-message "This menu has no video.\n\nDo you want to remove the VOB file\n(and all IFO's references to it)?"]
			if {$r == "yes"} {
				set ::utils::blankout_movevob true
				blank_out_PGCS_remove_menu $vtsnum
			}
			return
		}
	}

	set current_aspect [expr [get_int ::raw_sectors($vtsnum,$vtsi_table_name) $vtsi_mat_idx 2] & 0x0C00]
	set restxt "Current video standard:  [string toupper $standard]"
	if {$current_aspect == [expr 0x0C00]} {
		set restxt "$restxt  16:9"
	} else {
		set restxt "$restxt  4:3"
	}

	if {![info exists ::utils::blankout_type]} {
		set ::utils::blankout_type "black"
		set ::utils::blankout_killpb 1
		set ::utils::blankout_clip "pause"
		set ::utils::blankout_clip_celltime 5
		set ::utils::blankout_usemenu false
		set ::utils::blankout_movevob true
	}

	set w .blankout
	if [winfo exists $w] {destroy $w}
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w $utiltitle
	wm transient $w .
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	wm protocol $w WM_DELETE_WINDOW {set ::utils::blankout_done 0}
	bind $w <Escape> "set ::utils::blankout_done 0"
	#bind $w <Return> "set ::utils::blankout_done 1"

	if {$vtsnum == 0} {set t "VMGM"} else {set t "VTS $vtsnum"}
	label $w.lab -text "Blank out ALL $type of $t"
	pack  $w.lab -padx 6 -pady 4
	if {$type == "Menus"} {
		frame $w.labw -bd 2 -relief ridge
		pack  $w.labw -fill x -padx 4 -pady 1
		label $w.labw.l -text "WARNING: Blanking menus with buttons may lead to infinite loops!" \
				-background pink -height 2
		pack  $w.labw.l -fill both
	}

	# current video resolution
	frame $w.f0 -bd 2 -relief groove
	pack  $w.f0 -fill x -padx 4 -pady 4
	label $w.f0.l -text $restxt -anchor w -height 2
	pack  $w.f0.l -fill x -padx 8

	# main Blank frame
	frame $w.fm -bd 2 -relief groove
	pack  $w.fm -padx 4 -pady 4 -fill x
	label $w.fm.lab -text "Blank out method:" -anchor w
	pack  $w.fm.lab -fill x -anchor w -padx 8

	# Blank out and remove VOB
	if {$type == "Menus"} {
		frame $w.fm.fr -bd 2 -relief groove
		pack  $w.fm.fr -padx 4 -pady 4 -fill x
		radiobutton $w.fm.fr.b -text "Completely remove the menu VOB file." \
				-variable ::utils::blankout_type -value "rem" -anchor w \
				-command "::utils::blank_out_PGCS_validate $w"
		pack  $w.fm.fr.b -fill x -padx 4 -pady 1
		::balloon::set_balloon $w.fm.fr.b "Remove the VOB and all references in the IFOs.\nThis will convert all PGCs to dummies."
	} elseif {$::utils::blankout_type == "rem"} {
		set ::utils::blankout_type "black"
	}

	# Blank out with black frame
	frame $w.fm.f1 -bd 2 -relief groove
	pack  $w.fm.f1 -padx 4 -fill x
	radiobutton $w.fm.f1.b -text "Replace all $type by a black frame. (VOB size: 10KB)" \
			-variable ::utils::blankout_type -value "black" -anchor w \
			-command "::utils::blank_out_PGCS_validate $w"
	pack  $w.fm.f1.b -fill x -padx 4 -pady 1
	::balloon::set_balloon $w.fm.f1.b "Replace the original video by a short black clip."
	radiobutton $w.fm.f1.r1 -text "Kill playback if there are no cell commands in the PGC." \
			-variable ::utils::blankout_killpb -value 1 -pady 0
	pack  $w.fm.f1.r1 -anchor w -padx 24 -pady 0
	::balloon::set_balloon $w.fm.f1.r1 "Convert Post to Pre commands only when it is safe."
	radiobutton $w.fm.f1.r2 -text "Kill playback unconditionally." \
			-variable ::utils::blankout_killpb -value 2 -pady 0
	pack  $w.fm.f1.r2 -anchor w -padx 24 -pady 0
	::balloon::set_balloon $w.fm.f1.r2 "Kill playback even for PGCs with Cell commands.  May be unsafe!"
	radiobutton $w.fm.f1.r3 -text "Don't kill playback." \
			-variable ::utils::blankout_killpb -value 0 -pady 0
	pack  $w.fm.f1.r3 -anchor w -padx 24 -pady 0
	::balloon::set_balloon $w.fm.f1.r3 "Never convert Post to Pre commands.  Some black frames will be played."

	# Replace with clip frame
	frame $w.fm.f2 -bd 2 -relief groove
	pack  $w.fm.f2 -padx 4 -pady 4 -fill x
	set txt "Replace all $type by an imported clip (like a 'feature removed' still frame)."
	set txt "$txt\nThe imported clip MUST have only ONE VOB/Cell ID!"
	radiobutton $w.fm.f2.b -text $txt -justify left \
			-variable ::utils::blankout_type -value "clip" -anchor w \
			-command "::utils::blank_out_PGCS_validate $w"
	pack  $w.fm.f2.b -fill x -padx 4
	::balloon::set_balloon $w.fm.f2.b "Replace all videos by a still frame or an animated clip."
	frame $w.fm.f2.f
	pack  $w.fm.f2.f -fill x -padx 4 -fill x
	frame $w.fm.f2.f.f
	pack  $w.fm.f2.f.f -fill x -padx 20
	radiobutton $w.fm.f2.f.f.b1 -text "Display still frame" \
			-variable ::utils::blankout_clip -value "pause" -anchor w \
			-command "::utils::blank_out_PGCS_validate $w"
	pack  $w.fm.f2.f.f.b1 -fill x -side left
	::balloon::set_balloon $w.fm.f2.f.f.b1 "Pause N seconds after the clip has finished playing."
	entry $w.fm.f2.f.f.e -textvariable ::utils::blankout_clip_celltime -width 4 \
			-validate all -validatecommand "::ced::value_entry_validate 0 254 {%d} {%P}"
	pack  $w.fm.f2.f.f.e -side left
	label $w.fm.f2.f.f.l -text "seconds. (max 254)"
	pack  $w.fm.f2.f.f.l -side left
	radiobutton $w.fm.f2.f.b2 -text "Don't pause: it's an animated clip." \
			-variable ::utils::blankout_clip -value "anim" -anchor w \
			-command "::utils::blank_out_PGCS_validate $w"
	pack  $w.fm.f2.f.b2 -fill x -padx 20
	::balloon::set_balloon $w.fm.f2.f.b2 "Continue immediately when the clip has finished playing."
	radiobutton $w.fm.f2.f.b3 -text "BOV: use the Menu from source file. (For expert users!)\nThe menu MUST be in the FIRST PGC with a video cell,\nand have a 'Continue' button, with LinkTailPGC." \
			-variable ::utils::blankout_clip -value "bov" -anchor w -justify left \
			-command "::utils::blank_out_PGCS_validate $w"
	pack  $w.fm.f2.f.b3 -fill x -padx 20
	::balloon::set_balloon $w.fm.f2.f.b3 "Replace the video by a menu with a 'continue' button."

	# What to do with VOBs? frame
	frame $w.f3 -bd 2 -relief groove
	pack  $w.f3 -padx 4 -fill x
	radiobutton $w.f3.b1 -text "Move original VOB file(s) to backup." \
			-variable ::utils::blankout_movevob -value true -anchor w
	pack  $w.f3.b1 -fill x -padx 4
	::balloon::set_balloon $w.f3.b1 "Keep a backup of the original VOB."
	radiobutton $w.f3.b2 -text "Delete original VOB file(s)." \
			-variable ::utils::blankout_movevob -value false -anchor w
	pack  $w.f3.b2 -fill x -padx 4
	::balloon::set_balloon $w.f3.b2 "Overwrite the original VOB."

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -anchor e
	button $w.fb.ok -text "OK" -width 10 -command {set ::utils::blankout_done 1}
	focus  $w.fb.ok
	pack   $w.fb.ok -side left -padx 2
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::utils::blankout_done 0}
	pack   $w.fb.cancel -side left -padx 2

	blank_out_PGCS_validate $w

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr [winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)]
	set wy [expr [winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	wm geometry $w +$wx+$wy
	wm deiconify $w
	grab set $w

	tkwait variable ::utils::blankout_done
	grab release $w
	destroy $w

	if {$::utils::blankout_done} {
		# do the job

		switch $::utils::blankout_type {
			"rem" {
				blank_out_PGCS_remove_menu $vtsnum
			}
			"black" {
				blank_out_PGCS_blank $vtsnum $type $::utils::blankout_killpb
			}
			"clip" {
				if {$::utils::blankout_clip_celltime == ""} {
					set ::utils::blankout_clip_celltime 5
				}
				switch $::utils::blankout_clip {
					"pause" {blank_out_PGCS_replace $vtsnum $type $::utils::blankout_clip_celltime}
					"anim"  {blank_out_PGCS_replace $vtsnum $type 0}
					"bov"   {blank_out_PGCS_replace $vtsnum $type 255}
				}
			}
		}
	}
	return
}
proc blank_out_PGCS_validate {w} {
	switch $::utils::blankout_type {
		"rem" {
			$w.fm.f2.f.f.b1 configure -state disabled
			$w.fm.f2.f.f.e  configure -state disabled
			$w.fm.f2.f.f.l  configure -state disabled
			$w.fm.f2.f.b2   configure -state disabled
			$w.fm.f2.f.b3   configure -state disabled
			$w.fm.f1.r1     configure -state disabled
			$w.fm.f1.r2     configure -state disabled
			$w.fm.f1.r3     configure -state disabled
		}
		"black" {
			$w.fm.f2.f.f.b1 configure -state disabled
			$w.fm.f2.f.f.e  configure -state disabled
			$w.fm.f2.f.f.l  configure -state disabled
			$w.fm.f2.f.b2   configure -state disabled
			$w.fm.f2.f.b3   configure -state disabled
			$w.fm.f1.r1     configure -state normal
			$w.fm.f1.r2     configure -state normal
			$w.fm.f1.r3     configure -state normal
		}
		"clip" {
			$w.fm.f2.f.f.b1 configure -state normal
			if {$::utils::blankout_clip == "pause"} {
				$w.fm.f2.f.f.e  configure -state normal
				$w.fm.f2.f.f.l  configure -state normal
			} else {
				$w.fm.f2.f.f.e  configure -state disabled
				$w.fm.f2.f.f.l  configure -state disabled
			}
			$w.fm.f2.f.b2   configure -state normal
			$w.fm.f2.f.b3   configure -state normal
			$w.fm.f1.r1     configure -state disabled
			$w.fm.f1.r2     configure -state disabled
			$w.fm.f1.r3     configure -state disabled
		}
	}
	if {$::utils::blankout_clip_celltime == ""} {
		set ::utils::blankout_clip_celltime 5
	}
}
proc blank_out_remove_vobs {vtsnum type} {
	# remove (or move to backup) the original VTST VOBs.
	set backupdir [file join $::dvddir "PgcEdit_backup"]
	if {$type == "Titles"} {
		set destvobbasename [file join $::dvddir "VTS_[format %02i $vtsnum]_"]
		if $::utils::blankout_movevob {
			if {! [file exists $backupdir]} {file mkdir $backupdir}
			set back true
			if [file exists [file join $backupdir "VTS_[format %02i $vtsnum]_1.VOB"]] {
				set r [tk_messageBox -title "PgcEdit: Backup" -icon warning \
					-message "Backup of VOB files already exists!\nOverwrite?" -parent . \
					-type yesno -default no]
				if {$r == "no"} {set back false}
			}
			if $back {
				for {set i 1} {$i < 10} {incr i} {
					catch {file delete [file join $backupdir VTS_[format %02i $vtsnum]_$i.VOB]}
					if [file exists "${destvobbasename}$i.VOB"] {
						file rename -force -- "${destvobbasename}$i.VOB" $backupdir
					}
				}
			} else {
				for {set i 1} {$i < 10} {incr i} {
					catch {file delete "${destvobbasename}$i.VOB"}
				}
			}
		} else {
			for {set i 1} {$i < 10} {incr i} {
				catch {file delete "${destvobbasename}$i.VOB"}
			}
		}
	} else {
		if {$vtsnum == 0} {
			set name "VIDEO_TS.VOB"
		} else {
			set name "VTS_[format %02i $vtsnum]_0.VOB"
		}
		if $::utils::blankout_movevob {
			if {! [file exists $backupdir]} {file mkdir $backupdir}
			set back true
			if [file exists [file join $backupdir $name]] {
				set r [tk_messageBox -title "PgcEdit: Backup" -icon warning \
					-message "Backup of VOB file already exists!\nOverwrite?" -parent . \
					-type yesno -default no]
				if {$r == "no"} {set back false}
			}
			if $back {
				if [file exists [file join $::dvddir $name]] {
					catch {file delete [file join $backupdir $name]}
					file rename -force -- [file join $::dvddir $name] $backupdir
				}
			} else {
				catch {file delete [file join $::dvddir $name]}
			}
		} else {
			catch {file delete [file join $::dvddir $name]}
		}
	}
}

#############################################################################
## Remove Menu VOB
proc blank_out_PGCS_remove_menu {vtsnum} {
	# remove old files
	blank_out_remove_vobs $vtsnum "Menus"

	if {$vtsnum == 0} {
		set vtsi_table_name "VMGM_MAT"
		set tablesubname "VMGM"
	} else {
		set vtsi_table_name "VTSI_MAT"
		set tablesubname "VTSM"
	}
	set totalpbtime "\000\000\000\000"

	set vattr [get_int ::raw_sectors($vtsnum,$vtsi_table_name) 0x100 2]
	if {[expr $vattr & 0x3000] == 0} {
		set_int totalpbtime 0 192 4
	} else {
		set_int totalpbtime 0 64 4
	}
	### Go modify the DVD!

	## patch the current video attributes in IFOs
	#set vattr [expr ($vattr & 0x3FE3) | 0x4000]	;# MPEG 2, VBR, 720
	#set_int ::raw_sectors($vtsnum,$vtsi_table_name) 0x100 $vattr 2

	# set number of audio streams and number of subpics to 0
	set_int ::raw_sectors($vtsnum,$vtsi_table_name) 0x102 0 2
	set_int ::raw_sectors($vtsnum,$vtsi_table_name) 0x154 0 2


	# Fix all cells in all PGCS of current VTS
	for {set lu 1} {$lu <= $::menus($vtsnum,numlus)} {incr lu} {
		for {set pgc 1} {$pgc <= $::menus($vtsnum,$lu,numpgcs)} {incr pgc} {
			set ::menus($vtsnum,$lu,$pgc,numcells) 0
			set ::menus($vtsnum,$lu,$pgc,numprograms) 0
			set ::menus($vtsnum,$lu,$pgc,programmap) ""
			set ::menus($vtsnum,$lu,$pgc,cellpbtable) ""
			set ::menus($vtsnum,$lu,$pgc,cellpostable) ""
			# fix PGC total time
			set ::menus($vtsnum,$lu,$pgc,header) \
					"[string range $::menus($vtsnum,$lu,$pgc,header) 0 3]$totalpbtime[string range $::menus($vtsnum,$lu,$pgc,header) 8 end]"
			# reset PGC Still Time to 0
			set_int ::menus($vtsnum,$lu,$pgc,header) 0xA3 0 1
			# Kill PGC Playback
			::macros::kill_pgc_playback ::menus($vtsnum,$lu,$pgc)
		}
	}

	# remove the VTS_C_ADT and VTS_VOBU_ADMAP tables
	set ::raw_sectors($vtsnum,${tablesubname}_C_ADT) ""
	set ::raw_sectors($vtsnum,${tablesubname}_VOBU_ADMAP) ""

	set ::buttons::butinfo($vtsnum,errorcode) 1
	set ::buttons::butinfo($vtsnum,errormsg) "Can't open menu VOB file."
	set ::buttons::butinfo($vtsnum,numvobids) 0
	catch {
		unset ::buttons::butinfo($vtsnum,1,numcellids)
		unset ::buttons::butinfo($vtsnum,1,1,navpack)
		unset ::buttons::butinfo($vtsnum,1,1,numbuts)
		unset ::buttons::butinfo($vtsnum,1,1,numgroups)
		unset ::buttons::butinfo($vtsnum,1,1,forceselectbutn)
		unset ::buttons::butinfo($vtsnum,1,1,forceactionbutn)
		unset ::buttons::butinfo($vtsnum,1,1,1,grouptype)
		unset ::buttons::butinfo($vtsnum,1,1,1,1,autoaction)
		unset ::buttons::butinfo($vtsnum,1,1,1,1,coords)
		unset ::buttons::butinfo($vtsnum,1,1,1,1,uldr)
		unset ::buttons::butinfo($vtsnum,1,1,1,1,command)
		unset ::buttons::butinfo($vtsnum,1,1,2,grouptype)
		unset ::buttons::butinfo($vtsnum,1,1,3,grouptype)
	}

	# Reinit GUI
	set lunum  [lindex [split [lindex $::editor_cmds_from 1] ","] 1]
	set pgcnum [lindex [split [lindex $::editor_cmds_from 1] ","] 2]
	::main::init_tree_menu_no_move
	::main::display_pgc menu $vtsnum $lunum $pgcnum
	::main::dvd_modified true false
	show_msg "VTS $vtsnum Menu VOB removed OK."
	return
}

#############################################################################
#############################################################################
## Blank out all VTST Titles and replace them with black frame
proc blank_out_PGCS_blank {vtsnum type killpb} {
	# remove old files
	blank_out_remove_vobs $vtsnum $type

	# copy new VOBs
	if {$type == "Menus"} {
		set vtsi_mat_idx 0x100
		if {$vtsnum == 0} {
			set destvobname [file join $::dvddir "VIDEO_TS.VOB"]
			set tablesubname "VMGM"
		} else {
			set destvobname [file join $::dvddir "VTS_[format %02i $vtsnum]_0.VOB"]
			set tablesubname "VTSM"
		}
	} else {
		set vtsi_mat_idx 0x200
		set destvobname [file join $::dvddir "VTS_[format %02i $vtsnum]_1.VOB"]
		set tablesubname "VTS"
	}
	if {$vtsnum == 0} {
		set vtsi_table_name "VMGM_MAT"
	} else {
		set vtsi_table_name "VTSI_MAT"
	}
	set totalpbtime    "\000\000\000\000"
	set standardpbtime "\000\000\000\000"
	set vattr [get_int ::raw_sectors($vtsnum,$vtsi_table_name) $vtsi_mat_idx 2]
	if {[expr $vattr & 0x3000] == 0} {
		set standard "ntsc"
		set_int standardpbtime 0 210 4
	} else {
		set standard "pal"
		set_int standardpbtime 0 82 4
	}

	# patch the current video attributes in IFOs
	set vattr [expr ($vattr & 0x3F0E) | 0x4000]	;# MPEG 2, VBR, Camera, both NTSC CC off.
	set_int ::raw_sectors($vtsnum,$vtsi_table_name) $vtsi_mat_idx $vattr 2

	# set number of audio streams and number of subpics to 0
	set_int ::raw_sectors($vtsnum,$vtsi_table_name) [expr $vtsi_mat_idx + 2] 0 2
	set_int ::raw_sectors($vtsnum,$vtsi_table_name) [expr $vtsi_mat_idx + 0x54] 0 2

	# copy the new blank VOB
	saveFile $destvobname binary [::blankcell::blankCell $vattr 1 1 0]

	### Go modify the DVD!

	# Fix all cells in all PGCS of current VTS
	if {$type == "Menus"} {
		for {set lu 1} {$lu <= $::menus($vtsnum,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vtsnum,$lu,numpgcs)} {incr pgc} {
				if {$standard == "ntsc"} {
					set_int totalpbtime    0 192 4
				} else {
					set_int totalpbtime    0 64 4
				}
				for {set cell 1} {$cell <= $::menus($vtsnum,$lu,$pgc,numcells)} {incr cell} {
					set cellidx [expr ($cell-1)*24]
					;# set cell_type's SCR Discontinuity
					set typeflags [get_int ::menus($vtsnum,$lu,$pgc,cellpbtable) $cellidx 1]
					set_int ::menus($vtsnum,$lu,$pgc,cellpbtable) $cellidx [expr $typeflags | 2] 1
					;# clear restricted flag
					set_int ::menus($vtsnum,$lu,$pgc,cellpbtable) [expr $cellidx+1] 0 1
					# still time = 0
					set_int ::menus($vtsnum,$lu,$pgc,cellpbtable) [expr $cellidx+2] 0 1
					# Set PB time = still PB time in Cell Playback table
					set ::menus($vtsnum,$lu,$pgc,cellpbtable) \
							"[string range $::menus($vtsnum,$lu,$pgc,cellpbtable) 0 [expr $cellidx+3] \
							]$standardpbtime\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\004[ \
							string range $::menus($vtsnum,$lu,$pgc,cellpbtable) [expr $cellidx+24] end]"

					# Use VOB ID 1, Cell ID 1 in Cell Position table
					set idx [expr ($cell-1)*4]
					set_int ::menus($vtsnum,$lu,$pgc,cellpostable) $idx 1 2
					set_int ::menus($vtsnum,$lu,$pgc,cellpostable) [expr $idx+3] 1 1
					# Add cell time to total PGC time, but only if angle 1.
					if {[expr $typeflags & 0x10] == 0 || [expr $typeflags & 0xC0] == 0x40} {
						set totalpbtime [::pgced::AddBCDtime $totalpbtime $standardpbtime]
					}
				}
				# fix PGC total time
				set ::menus($vtsnum,$lu,$pgc,header) \
						"[string range $::menus($vtsnum,$lu,$pgc,header) 0 3]$totalpbtime[string range $::menus($vtsnum,$lu,$pgc,header) 8 end]"
				# reset PGC Still Time to 0
				set_int ::menus($vtsnum,$lu,$pgc,header) 0xA3 0 1

				# if there are no cell commands, or if force kill is on, Kill PGC Playback
				if {$killpb == 1} {
					if {[num_type_cmds cell ::menus($vtsnum,$lu,$pgc)] == 0} {
						::macros::kill_pgc_playback ::menus($vtsnum,$lu,$pgc)
					}
				} elseif {$killpb == 2} {
					::macros::kill_pgc_playback ::menus($vtsnum,$lu,$pgc)
				}
			}
		}
		set ::buttons::butinfo($vtsnum,errorcode) 0
		set ::buttons::butinfo($vtsnum,errormsg) "OK"
		set ::buttons::butinfo($vtsnum,numvobids) 1
		set ::buttons::butinfo($vtsnum,1,numcellids) 1
		set ::buttons::butinfo($vtsnum,1,1,navpack) 0
		set ::buttons::butinfo($vtsnum,1,1,numbuts) 0
		catch {
			unset ::buttons::butinfo($vtsnum,1,1,numgroups)
			unset ::buttons::butinfo($vtsnum,1,1,forceselectbutn)
			unset ::buttons::butinfo($vtsnum,1,1,forceactionbutn)
			unset ::buttons::butinfo($vtsnum,1,1,1,grouptype)
			unset ::buttons::butinfo($vtsnum,1,1,1,1,autoaction)
			unset ::buttons::butinfo($vtsnum,1,1,1,1,coords)
			unset ::buttons::butinfo($vtsnum,1,1,1,1,uldr)
			unset ::buttons::butinfo($vtsnum,1,1,1,1,command)
			unset ::buttons::butinfo($vtsnum,1,1,2,grouptype)
			unset ::buttons::butinfo($vtsnum,1,1,3,grouptype)
		}
	} else {
		for {set pgc 1} {$pgc <= $::pgcs($vtsnum,numpgcs)} {incr pgc} {
			if {$standard == "ntsc"} {
				set_int totalpbtime    0 192 4
			} else {
				set_int totalpbtime    0 64 4
			}
			for {set cell 1} {$cell <= $::pgcs($vtsnum,$pgc,numcells)} {incr cell} {
				set cellidx [expr ($cell-1)*24]
				;# set cell_type's SCR Discontinuity
				set typeflags [get_int ::pgcs($vtsnum,$pgc,cellpbtable) $cellidx 1]
				set_int ::pgcs($vtsnum,$pgc,cellpbtable) $cellidx [expr $typeflags | 2] 1
				;# clear restricted flag
				set_int ::pgcs($vtsnum,$pgc,cellpbtable) [expr $cellidx+1] 0 1
				# still time = 0
				set_int ::pgcs($vtsnum,$pgc,cellpbtable) [expr $cellidx+2] 0 1
				# Set PB time = still PB time in Cell Playback table
				set ::pgcs($vtsnum,$pgc,cellpbtable) \
						"[string range $::pgcs($vtsnum,$pgc,cellpbtable) 0 [expr $cellidx+3] \
						]$standardpbtime\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\004[ \
						string range $::pgcs($vtsnum,$pgc,cellpbtable) [expr $cellidx+24] end]"

				# Use VOB ID 1, Cell ID 1 in Cell Position table
				set idx [expr ($cell-1)*4]
				set_int ::pgcs($vtsnum,$pgc,cellpostable) $idx 1 2
				set_int ::pgcs($vtsnum,$pgc,cellpostable) [expr $idx+3] 1 1
				# Add cell time to total PGC time, but only if angle 1.
				if {[expr $typeflags & 0x10] == 0 || [expr $typeflags & 0xC0] == 0x40} {
					set totalpbtime [::pgced::AddBCDtime $totalpbtime $standardpbtime]
				}
			}
			# fix PGC total time
			set ::pgcs($vtsnum,$pgc,header) \
					"[string range $::pgcs($vtsnum,$pgc,header) 0 3]$totalpbtime[string range $::pgcs($vtsnum,$pgc,header) 8 end]"
			# reset PGC Still Time to 0
			set_int ::pgcs($vtsnum,$pgc,header) 0xA3 0 1

			# if there are no cell commands, or if force kill is on, Kill PGC Playback
			if {$killpb == 1} {
				if {[num_type_cmds cell ::pgcs($vtsnum,$pgc)] == 0} {
					::macros::kill_pgc_playback ::pgcs($vtsnum,$pgc)
				}
			} elseif {$killpb == 2} {
				::macros::kill_pgc_playback ::pgcs($vtsnum,$pgc)
			}
		}
	}

	# replace the VTS_C_ADT and VTS_VOBU_ADMAP tables with blank 1 cell VOB tables
	set ::raw_sectors($vtsnum,${tablesubname}_C_ADT) \
		"\000\001\000\000\000\000\000\023\000\001\001\000\000\000\000\000\000\000\000\004"
	set ::raw_sectors($vtsnum,${tablesubname}_VOBU_ADMAP) \
		"\000\000\000\007\000\000\000\000"

	# Reinit GUI
	#refresh_current_pgc
	::main::init_tree_menu_no_move
	::main::dvd_modified true false
	show_msg "VTS $vtsnum $type blanked OK."
	return
}


#############################################################################
## Replace all VTST Titles by an imported title
proc blank_out_PGCS_replace {vtsnum type cell_still_time} {

	# set destination defaults
	if {$type == "Menus"} {
		set vtsi_mat_idx 0x100
		if {$vtsnum == 0} {
			set destvobname [file join $::dvddir "VIDEO_TS.VOB"]
			set tablesubname "VMGM"
		} else {
			set destvobname [file join $::dvddir "VTS_[format %02i $vtsnum]_0.VOB"]
			set tablesubname "VTSM"
		}
	} else {
		set vtsi_mat_idx 0x200
		set tablesubname "VTS"
	}
	if {$vtsnum == 0} {
		set vtsi_table_name "VMGM_MAT"
	} else {
		set vtsi_table_name "VTSI_MAT"
	}

	# set source defaults
	if {$::utils::blankout_clip == "bov"} {
		set filetypes {{"IFO files" {*.IFO}}}
	} else {
		set filetypes {{"VTS IFO files" {*_0.IFO}}}
	}
	set sourceifofile [tk_getOpenFile -initialdir $::config(importmenuinitialdir) \
		-filetypes $filetypes \
		-title "Select the VTS IFO file with the clip to import"]
	if {$sourceifofile == ""} {return 0}
	if {$::utils::blankout_clip == "bov" && [file tail $sourceifofile] == "VIDEO_TS.IFO"} {
		show_warning -title $utiltitle -icon error -parent . \
				-message "You must select a VTS IFO (VTS_**_*.IFO)"
		return 0
	}
	set sourceifofile [file nativename $sourceifofile]
	set ::config(importmenuinitialdir) [file dirname $sourceifofile]

	# read the VTS
	set sourcevtsnum [readdvdfile $sourceifofile ::imported_sectors ::imported_pgcs ::imported_menus ::imported_ttns]
	if {$sourcevtsnum == -1} {return 0}


	# Process Menu if BOV mode
	if {$::utils::blankout_clip == "bov"} {
		# Find out the first menu PGC with video
		set menupgc 0
		for {set p 1} {$p <= $::imported_menus($sourcevtsnum,1,numpgcs)} {incr p} {
			if {$::imported_menus($sourcevtsnum,1,$p,numcells)>0} {
				set menupgc $p
				break
			}
		}
		if {$menupgc == 0} {
			show_warning -title $utiltitle -icon error -parent . -message "No menu with video found!"
			return 0
		}
		set importedpbtime [string range $::imported_menus($sourcevtsnum,1,$menupgc,cellpbtable) 4 7]
		set importedpbtablepart2 [string range $::imported_menus($sourcevtsnum,1,$menupgc,cellpbtable) 4 23]
		if {$sourcevtsnum == 0} {
			set importedstandard [expr [get_int ::imported_sectors(0,VMGM_MAT) 0x100 2] & 0x3000]
			set source_C_ADT_table $::imported_sectors(0,VMGM_C_ADT)
			set source_VOBU_ADMAP_table $::imported_sectors(0,VMGM_VOBU_ADMAP)
			set source_vtsi_table_name "VMGM_MAT"
		} else {
			set importedstandard [expr [get_int ::imported_sectors($sourcevtsnum,VTSI_MAT) 0x100 2] & 0x3000]
			set source_C_ADT_table $::imported_sectors($sourcevtsnum,VTSM_C_ADT)
			set source_VOBU_ADMAP_table $::imported_sectors($sourcevtsnum,VTSM_VOBU_ADMAP)
			set source_vtsi_table_name "VTSI_MAT"
		}
		set source_vtsi_table_address [expr 0x100]
	} else {
		# How many TTNs are in the imported VTS?
		set num_new_ttns [get_int ::imported_sectors($sourcevtsnum,VTS_PTT_SRPT) 0 2]
		if {$num_new_ttns == 0 || $num_new_ttns > 1} {
			show_warning -title $utiltitle -icon error -parent . \
					-message "Error: you must select a VTS with one and only one title!"
			# cleanup
			unset ::imported_sectors ::imported_pgcs ::imported_menus
			return 0
		}
		set importedpbtime [string range $::imported_pgcs($sourcevtsnum,1,cellpbtable) 4 7]
		set importedpbtablepart2 [string range $::imported_pgcs($sourcevtsnum,1,cellpbtable) 4 23]
		set importedstandard [expr [get_int ::imported_sectors($sourcevtsnum,VTSI_MAT) 0x200 2] & 0x3000]

		set source_C_ADT_table $::imported_sectors($sourcevtsnum,VTS_C_ADT)
		set source_VOBU_ADMAP_table $::imported_sectors($sourcevtsnum,VTS_VOBU_ADMAP)
		set source_vtsi_table_name "VTSI_MAT"
		set source_vtsi_table_address [expr 0x200]
	}
	if {$importedstandard == 0} {	;# if standard of imported clip is NTSC
		set importedstandard "NTSC"
	} else {
		set importedstandard "PAL"
	}

	# copy the TTN VOBs
	show_msg "Copying VOBs..."
	update idletasks
	set destvobbasename [file join $::dvddir "VTS_[format %02i $vtsnum]_"]
	if {$::utils::blankout_clip == "bov"} {
		set sourcevobname "[file rootname $sourceifofile].VOB"
	} else {
		set sourcevobbasename [string trimright [file rootname $sourceifofile] "0"]
		set sourcevobname "${sourcevobbasename}1.VOB"
	}
	if {$type == "Menus"} {
		if {$vtsnum == 0} {
			if {$sourcevobname != [file join $::dvddir "VIDEO_TS.VOB"]} {
				# remove old files
				blank_out_remove_vobs $vtsnum $type
				if [file exists $sourcevobname] {
					if [file exists [file join $::dvddir "VIDEO_TS.VOB"]] {
						file delete [file join $::dvddir "VIDEO_TS.VOB"]
					}
					file copy $sourcevobname [file join $::dvddir "VIDEO_TS.VOB"]
				} else bell
			}
		} else {
			if {$sourcevobname != "${destvobbasename}0.VOB"} {
				# remove old files
				blank_out_remove_vobs $vtsnum $type
				if [file exists $sourcevobname] {
					if [file exists "${destvobbasename}0.VOB"] {
						file delete "${destvobbasename}0.VOB"
					}
					file copy $sourcevobname "${destvobbasename}0.VOB"
				} else bell
			}
		}
	} else {
		if {$::utils::blankout_clip != "bov"} {
			if {"${sourcevobbasename}1.VOB" != "${destvobbasename}1.VOB"} {
				# remove old files
				blank_out_remove_vobs $vtsnum $type
				for {set i 1} {$i < 10} {incr i} {
					if [file exists "${destvobbasename}$i.VOB"] {
						file delete "${destvobbasename}$i.VOB"
					}
					if [file exists "${sourcevobbasename}$i.VOB"] {
						file copy "${sourcevobbasename}$i.VOB" "${destvobbasename}$i.VOB"
					}
				}
			}
		} else {
			if {$sourcevobname != "${destvobbasename}1.VOB"} {
				# remove old files
				blank_out_remove_vobs $vtsnum $type
				for {set i 1} {$i < 10} {incr i} {
					if [file exists "${destvobbasename}$i.VOB"] {
						file delete "${destvobbasename}$i.VOB"
					}
				}
				if [file exists $sourcevobname] {
					file copy $sourcevobname "${destvobbasename}1.VOB"
				}
			}
		}
	}

	# Fix all cells in all PGCS of current VTS
	set totalpbtime "\000\000\000\000"
	if {$type == "Menus"} {
		for {set lu 1} {$lu <= $::menus($vtsnum,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vtsnum,$lu,numpgcs)} {incr pgc} {
				if {$importedstandard == "NTSC"} {
					set_int totalpbtime    0 192 4
				} else {
					set_int totalpbtime    0 64 4
				}
				for {set cell 1} {$cell <= $::menus($vtsnum,$lu,$pgc,numcells)} {incr cell} {
					set cellidx [expr ($cell-1)*24]

					;# set cell_type's SCR Discontinuity
					set typeflags [get_int ::menus($vtsnum,$lu,$pgc,cellpbtable) $cellidx 1]
					set_int ::menus($vtsnum,$lu,$pgc,cellpbtable) $cellidx [expr $typeflags | 2] 1

					;# clear restricted flag
					set_int ::menus($vtsnum,$lu,$pgc,cellpbtable) [expr $cellidx+1] 0 1
					# if first cell or previous cell has cell command, then set the still frame delay,
					# else set still frame delay = 0
					if {$cell == 1 || \
							[get_int ::menus($vtsnum,$lu,$pgc,cellpbtable) [expr $cellidx+3-24] 1] > 0} {
						set_int ::menus($vtsnum,$lu,$pgc,cellpbtable) [expr $cellidx+2] $cell_still_time 1
					} else {
						set_int ::menus($vtsnum,$lu,$pgc,cellpbtable) [expr $cellidx+2] 0 1
					}
					# Set PB time = still PB time in Cell Playback table
					set ::menus($vtsnum,$lu,$pgc,cellpbtable) \
							"[string range $::menus($vtsnum,$lu,$pgc,cellpbtable) 0 [expr $cellidx+3] \
							]$importedpbtablepart2[ \
							string range $::menus($vtsnum,$lu,$pgc,cellpbtable) [expr $cellidx+24] end]"

					# Use VOB ID 1, Cell ID 1 in Cell Position table
					set idx [expr ($cell-1)*4]
					set_int ::menus($vtsnum,$lu,$pgc,cellpostable) $idx 1 2
					set_int ::menus($vtsnum,$lu,$pgc,cellpostable) [expr $idx+3] 1 1

					# Add cell time to total PGC time, but only if angle 1.
					if {[expr $typeflags & 0x10] == 0 || [expr $typeflags & 0xC0] == 0x40} {
						set totalpbtime [::pgced::AddBCDtime $totalpbtime $importedpbtime]
					}
				}
				# fix PGC total time
				set ::menus($vtsnum,$lu,$pgc,header) \
						"[string range $::menus($vtsnum,$lu,$pgc,header) 0 3]$totalpbtime[string range $::menus($vtsnum,$lu,$pgc,header) 8 end]"
				# reset PGC Still Time to 0
				set_int ::menus($vtsnum,$lu,$pgc,header) 0xA3 0 1

				# Clear all PUOs
				set_int ::menus($vtsnum,$lu,$pgc,header) 0x08 0 4
				# Add the NextPGCN and PrevPGCN pointers to be able to use the Next and Prev buttons.
				if {[get_int ::menus($vtsnum,$lu,$pgc,header) 0x9C 2] == 0 &&
						[num_type_cmds post ::menus($vtsnum,$lu,$pgc)] > 0} {
					 set_int ::menus($vtsnum,$lu,$pgc,header) 0x9C $pgc 2
				}
				if {[get_int ::menus($vtsnum,$lu,$pgc,header) 0x9E 2] == 0} {
					 set_int ::menus($vtsnum,$lu,$pgc,header) 0x9E $pgc 2
				}
			}
		}
	} else {
		for {set pgc 1} {$pgc <= $::pgcs($vtsnum,numpgcs)} {incr pgc} {
			if {$importedstandard == "NTSC"} {
				set_int totalpbtime    0 192 4
			} else {
				set_int totalpbtime    0 64 4
			}
			for {set cell 1} {$cell <= $::pgcs($vtsnum,$pgc,numcells)} {incr cell} {
				set cellidx [expr ($cell-1)*24]

				;# set cell_type's SCR Discontinuity
				set typeflags [get_int ::pgcs($vtsnum,$pgc,cellpbtable) $cellidx 1]
				set_int ::pgcs($vtsnum,$pgc,cellpbtable) $cellidx [expr $typeflags | 2] 1
				;# clear restricted flag
				set_int ::pgcs($vtsnum,$pgc,cellpbtable) [expr $cellidx+1] 0 1
				# if first cell or previous cell has cell command, then set the still frame delay,
				# else set still frame delay = 0
				if {$cell == 1 || \
						[get_int ::pgcs($vtsnum,$pgc,cellpbtable) [expr $cellidx+3-24] 1] > 0} {
					set_int ::pgcs($vtsnum,$pgc,cellpbtable) [expr $cellidx+2] $cell_still_time 1
				} else {
					set_int ::pgcs($vtsnum,$pgc,cellpbtable) [expr $cellidx+2] 0 1
				}
				# Set PB time = still PB time in Cell Playback table
				set ::pgcs($vtsnum,$pgc,cellpbtable) \
						"[string range $::pgcs($vtsnum,$pgc,cellpbtable) 0 [expr $cellidx+3] \
						]$importedpbtablepart2[ \
						string range $::pgcs($vtsnum,$pgc,cellpbtable) [expr $cellidx+24] end]"

				# Use VOB ID 1, Cell ID 1 in Cell Position table
				set idx [expr ($cell-1)*4]
				set_int ::pgcs($vtsnum,$pgc,cellpostable) $idx 1 2
				set_int ::pgcs($vtsnum,$pgc,cellpostable) [expr $idx+3] 1 1

				# Add cell time to total PGC time, but only if angle 1.
				if {[expr $typeflags & 0x10] == 0 || [expr $typeflags & 0xC0] == 0x40} {
					set totalpbtime [::pgced::AddBCDtime $totalpbtime $importedpbtime]
				}
			}
			# fix PGC total time
			set ::pgcs($vtsnum,$pgc,header) \
					"[string range $::pgcs($vtsnum,$pgc,header) 0 3]$totalpbtime[string range $::pgcs($vtsnum,$pgc,header) 8 end]"
			# reset PGC Still Time to 0
			set_int ::pgcs($vtsnum,$pgc,header) 0xA3 0 1
			# Clear all PUOs
			set_int ::pgcs($vtsnum,$pgc,header) 0x08 0 4
			# Add the NextPGCN and PrevPGCN pointers to be able to use the Next and Prev buttons.
			if {[get_int ::pgcs($vtsnum,$pgc,header) 0x9C 2] == 0 &&
					[num_type_cmds post ::pgcs($vtsnum,$pgc)] > 0} {
				 set_int ::pgcs($vtsnum,$pgc,header) 0x9C $pgc 2
			}
			if {[get_int ::pgcs($vtsnum,$pgc,header) 0x9E 2] == 0} {
				 set_int ::pgcs($vtsnum,$pgc,header) 0x9E $pgc 2
			}
		}
	}

	### Go modify the DVD!

	# replace the VTS_C_ADT and VTS_VOBU_ADMAP tables with blank imported ones
	set ::raw_sectors($vtsnum,${tablesubname}_C_ADT) $source_C_ADT_table
	set ::raw_sectors($vtsnum,${tablesubname}_VOBU_ADMAP) $source_VOBU_ADMAP_table

	# copy the Video/Audio/Subpic attributes in VTSI_MAT
	set vattr [get_int ::imported_sectors($sourcevtsnum,$source_vtsi_table_name) $source_vtsi_table_address 2]
	set_int ::raw_sectors($vtsnum,$vtsi_table_name) $vtsi_mat_idx $vattr 2

	# copy audio streams attribs
	set numaudio [get_int ::imported_sectors($sourcevtsnum,$source_vtsi_table_name) [expr $source_vtsi_table_address + 2] 2]
	set_int ::raw_sectors($vtsnum,$vtsi_table_name) [expr $vtsi_mat_idx + 2] $numaudio 2
	set audioattribs [string range $::imported_sectors($sourcevtsnum,$source_vtsi_table_name) [expr $source_vtsi_table_address + 4] [expr $source_vtsi_table_address + 67]]
	set ::raw_sectors($vtsnum,$vtsi_table_name) [string range $::raw_sectors($vtsnum,$vtsi_table_name) 0 [expr $vtsi_mat_idx + 3]]$audioattribs[string range $::raw_sectors($vtsnum,$vtsi_table_name) [expr $vtsi_mat_idx + 68] end]

	# copy subpic streams attribs
	set numsubpics [get_int ::imported_sectors($sourcevtsnum,$source_vtsi_table_name) [expr $source_vtsi_table_address + 0x54] 2]
	if {$type == "Menus" && $numsubpics > 1} {set numsubpics 1}
	set_int ::raw_sectors($vtsnum,$vtsi_table_name) [expr $vtsi_mat_idx + 0x54] $numsubpics 2
	set subpicattribs [string range $::imported_sectors($sourcevtsnum,$source_vtsi_table_name) [expr $source_vtsi_table_address + 0x56] [expr $source_vtsi_table_address + 0x55 + ($numsubpics*6)]]
	set ::raw_sectors($vtsnum,$vtsi_table_name) [string range $::raw_sectors($vtsnum,$vtsi_table_name) 0 [expr $vtsi_mat_idx + 0x55]]$subpicattribs[string range $::raw_sectors($vtsnum,$vtsi_table_name) [expr $vtsi_mat_idx + 0x56 + ($numsubpics*6)] end]

	# cleanup
	unset ::imported_sectors ::imported_pgcs ::imported_menus

	# Reinit GUI
	#refresh_current_pgc
	::main::init_tree_menu_no_move
	::main::dvd_modified true false
	show_msg "VTS $vtsnum $type replaced OK."

	# allow user to reload buttons informations
	if {$type == "Menus"} {
		::readdvd::reload_buttons_infos $vtsnum
	}

	return
}


#############################################################################
## New (create or replace) the VMGM_PGCI_UT Title Menu PGC
proc new_title_menu {quiet} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set utiltitle "New VMGM Title Menu PGC"
	if {!$quiet && !$::config(expert_mode)} {
		set txt "This utility will create the VMGM Title menu PGC."
		set txt "$txt\nNo VOB will be associated with the menu."
		set txt "$txt\n"
		set txt "$txt\nContinue ?"
		set r [tk_messageBox -title $utiltitle -icon question -message $txt \
				-parent . -type yesno -default no]
		if {$r == "no"} return
	}

	### create or replace the VMGM menu
	if {$::menus(0,numlus)>0 && !$::config(expert_mode)} {
		set txt "WARNING: There is already a VMGM Title Menu!"
		set txt "$txt\nIf you continue, it will be REPLACED by the new one!"
		set txt "$txt\n"
		set txt "$txt\nContinue anyway ?"
		set r [tk_messageBox -title $utiltitle -icon warning -message $txt \
				-parent . -type yesno -default no]
		if {$r == "no"} return
	}
	set ::raw_sectors(0,VMGM_PGCI_UT) \
			[readFile $::config(apppath)/lib/data/vmgm_pgci_ut_wo_vob.sec binary]

	parse_PGCI_UT ::raw_sectors(0,VMGM_PGCI_UT) 0 ::menus
	::main::init_tree_menu
	::main::dvd_modified true false
	::main::display_pgc menu 0 1 1

	# if there is an old VOB, delete it or move it to backup
	delete_vob_with_confirm "VIDEO_TS.VOB" true
}

#############################################################################
## New (create or replace) a VTSM_PGCI_UT Root Menu PGC
proc new_root_menu {vtsnum quiet} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}

	set utiltitle "New VTSM Root Menu PGC"
	if {!$quiet && !$::config(expert_mode)} {
		set txt "This utility will create the VTSM $vtsnum Root menu PGC."
		set txt "$txt\nNo VOB will be associated with the menu."
		set txt "$txt\n"
		set txt "$txt\nContinue ?"
		set r [tk_messageBox -title $utiltitle -icon question -message $txt \
				-parent . -type yesno -default no]
		if {$r == "no"} return
	}

	### create or replace the VTS menu
	if {$::menus($vtsnum,numlus)>0 && !$::config(expert_mode)} {
		set txt "WARNING: There is already a VTSM $vtsnum menu!"
		set txt "$txt\nIf you continue, it will be REPLACED by the new one!"
		set txt "$txt\n"
		set txt "$txt\nContinue anyway ?"
		set r [tk_messageBox -title $utiltitle -icon warning -message $txt \
				-parent . -type yesno -default no]
		if {$r == "no"} return
	}

	set ::raw_sectors($vtsnum,VTSM_PGCI_UT) \
			[readFile $::config(apppath)/lib/data/vmgm_pgci_ut_wo_vob.sec binary]
	set_int ::raw_sectors($vtsnum,VTSM_PGCI_UT) 0x18 0x83 1	;# change Title menu to Root menu

	parse_PGCI_UT ::raw_sectors($vtsnum,VTSM_PGCI_UT) $vtsnum ::menus
	::main::init_tree_menu
	::main::dvd_modified true false
	::main::display_pgc menu $vtsnum 1 1

	# if there is an old VOB, delete it or move it to backup
	delete_vob_with_confirm "VTS_[format %02d $vtsnum]_0.VOB" true
}

proc new_menu {quiet} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set vtsnum [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
	if {$vtsnum == 0} {
		new_title_menu $quiet
	} else {
		new_root_menu $vtsnum $quiet
	}

	set ::buttons::butinfo($vtsnum,errorcode) 1
	set ::buttons::butinfo($vtsnum,errormsg) "Can't open menu VOB file."
	set ::buttons::butinfo($vtsnum,numvobids) 0
	catch {
		unset ::buttons::butinfo($vtsnum,1,numcellids)
		unset ::buttons::butinfo($vtsnum,1,1,navpack)
		unset ::buttons::butinfo($vtsnum,1,1,numbuts)
		unset ::buttons::butinfo($vtsnum,1,1,numgroups)
		unset ::buttons::butinfo($vtsnum,1,1,forceselectbutn)
		unset ::buttons::butinfo($vtsnum,1,1,forceactionbutn)
		unset ::buttons::butinfo($vtsnum,1,1,1,grouptype)
		unset ::buttons::butinfo($vtsnum,1,1,1,1,autoaction)
		unset ::buttons::butinfo($vtsnum,1,1,1,1,coords)
		unset ::buttons::butinfo($vtsnum,1,1,1,1,uldr)
		unset ::buttons::butinfo($vtsnum,1,1,1,1,command)
		unset ::buttons::butinfo($vtsnum,1,1,2,grouptype)
		unset ::buttons::butinfo($vtsnum,1,1,3,grouptype)
	}
}

#############################################################################
## Import a whole Menu
proc import_menu {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set utiltitle "Import Menu"
	set vtsnum [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
	if {$vtsnum == 0} {
		set current "VMGM"
	} else {
		set current "VTSM $vtsnum"
	}

	if {$::menus($vtsnum,numlus)>0 && !$::config(expert_mode)} {
		set txt "WARNING: There is already a $current menu!"
		set txt "$txt\nIf you continue, it will be REPLACED by the new one!"
		set txt "$txt\n"
		set txt "$txt\nContinue anyway ?"
		set r [tk_messageBox -title $utiltitle -icon warning -message $txt \
				-parent . -type yesno -default no]
		if {$r == "no"} return
	}

	set sourceifofile [tk_getOpenFile -initialdir $::config(importmenuinitialdir) \
		-filetypes {{"IFO files" {*.IFO}}} \
		-title "Select the IFO file with the menu to import"]
	if {$sourceifofile == ""} {return}
	set sourceifofile [file nativename $sourceifofile]
	set ::config(importmenuinitialdir) [file dirname $sourceifofile]

	set sourcevtsnum [readdvdfile $sourceifofile ::imported_sectors ::imported_pgcs ::imported_menus ::imported_ttns]
	if {$sourcevtsnum == -1} return
	if {$sourcevtsnum == 0} {
		set sourcetype "VMG"
		set sourceMAT "VMGM_MAT"
		set sourcetext "VMGM"
	} else {
		set sourcetype "VTS"
		set sourceMAT "VTSI_MAT"
		set sourcetext "VTSM $sourcevtsnum"
	}

	if {$::imported_sectors($sourcevtsnum,${sourcetype}M_PGCI_UT) == ""} {
		show_warning -title $utiltitle -icon warning -parent . \
				-message "Sorry, there is no menu defined in file\n\"$sourceifofile\""
		return
	}
	set menustreamsdefs [string range $::imported_sectors($sourcevtsnum,$sourceMAT) 0x100 0x1FF]
	if {$vtsnum == 0} {
		set destvtstype "VMG"
		set destvobfilename [file join $::dvddir "VIDEO_TS.VOB"]
		set ::raw_sectors(0,VMGM_MAT) \
				[string range $::raw_sectors(0,VMGM_MAT) 0 0xFF]$menustreamsdefs[string range $::raw_sectors(0,VMGM_MAT) 0x200 end]
	} else {
		set destvtstype "VTS"
		set destvobfilename [file join $::dvddir "VTS_[format %02i $vtsnum]_0.VOB"]
		set ::raw_sectors($vtsnum,VTSI_MAT) \
				[string range $::raw_sectors($vtsnum,VTSI_MAT) 0 0xFF]$menustreamsdefs[string range $::raw_sectors($vtsnum,VTSI_MAT) 0x200 end]
	}
	set ::raw_sectors($vtsnum,${destvtstype}M_PGCI_UT) \
			$::imported_sectors($sourcevtsnum,${sourcetype}M_PGCI_UT)
	set ::raw_sectors($vtsnum,${destvtstype}M_C_ADT) \
			$::imported_sectors($sourcevtsnum,${sourcetype}M_C_ADT)
	set ::raw_sectors($vtsnum,${destvtstype}M_VOBU_ADMAP) \
			$::imported_sectors($sourcevtsnum,${sourcetype}M_VOBU_ADMAP)


	parse_PGCI_UT ::raw_sectors($vtsnum,${destvtstype}M_PGCI_UT) $vtsnum ::menus
	::main::init_tree_menu
	::main::display_pgc menu $vtsnum 1 1
	::main::dvd_modified true false

	set vob_copied false
	if {$::imported_sectors($sourcevtsnum,${sourcetype}M_VOBU_ADMAP) != ""} {
		set sourcevobfilename "[file rootname $sourceifofile].VOB"
		if {$destvobfilename != $sourcevobfilename} {
			if [file exists $sourcevobfilename] {
				set txt "Copy the Menu VOB file ?"
				set r [tk_messageBox -title $utiltitle -icon question -message $txt \
						-parent . -type yesno -default yes]
				if {$r == "yes"} {
					# if there is an old VOB, delete it or move it to backup
					delete_vob_with_confirm "VTS_[format %02i $vtsnum]_0.VOB" false

					show_msg "Copying $sourcetext VOB to $current VOB.  Please wait."
					update idletasks
					file copy -force $sourcevobfilename $destvobfilename
					show_msg ""
					set vob_copied true
				}
			}
		}
	} else {
		# if there is an old VOB, delete it or move it to backup
		delete_vob_with_confirm "VTS_[format %02i $vtsnum]_0.VOB" true
	}

	unset ::imported_sectors ::imported_pgcs ::imported_menus

	# allow user to reload buttons informations
	if $vob_copied {
		::readdvd::reload_buttons_infos $vtsnum
	}

	return
}

#############################################################################
## Remove the whole Menu
proc remove_menu {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set utiltitle "Remove Menu"
	set vtsnum [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
	if {$vtsnum == 0} {
		set current "VMGM"
		set vtstype "VMG"
		set vobfilename [file join $::dvddir "VIDEO_TS.VOB"]
	} else {
		set current "VTSM $vtsnum"
		set vtstype "VTS"
		set vobfilename [file join $::dvddir "VTS_[format %02i $vtsnum]_0.VOB"]
	}

	if {$::menus($vtsnum,numlus)>0} {
		if {!$::config(expert_mode)} {
			set txt "This utility will remove the whole $current Menu."
			set txt "$txt\n"
			set txt "$txt\nAre you sure ?"
			set r [tk_messageBox -title $utiltitle -icon warning -message $txt \
					-parent . -type yesno -default no]
			if {$r == "no"} return
		}
	} else {
		show_warning -title $utiltitle -icon error -parent . \
				-message "There is no menu for the $current!"
		return
	}

	set ::raw_sectors($vtsnum,${vtstype}M_PGCI_UT) ""
	set ::raw_sectors($vtsnum,${vtstype}M_C_ADT) ""
	set ::raw_sectors($vtsnum,${vtstype}M_VOBU_ADMAP) ""
	for {set i 1} {$i <= $::menus($vtsnum,numlus)} {incr i} {
		set ::menus($vtsnum,$i,numpgcs) 0
	}
	set ::menus($vtsnum,numlus) 0

	set ::buttons::butinfo($vtsnum,errorcode) 1
	set ::buttons::butinfo($vtsnum,errormsg) "Can't open menu VOB file."
	set ::buttons::butinfo($vtsnum,numvobids) 0
	catch {
		unset ::buttons::butinfo($vtsnum,1,numcellids)
		unset ::buttons::butinfo($vtsnum,1,1,navpack)
		unset ::buttons::butinfo($vtsnum,1,1,numbuts)
		unset ::buttons::butinfo($vtsnum,1,1,numgroups)
		unset ::buttons::butinfo($vtsnum,1,1,forceselectbutn)
		unset ::buttons::butinfo($vtsnum,1,1,forceactionbutn)
		unset ::buttons::butinfo($vtsnum,1,1,1,grouptype)
		unset ::buttons::butinfo($vtsnum,1,1,1,1,autoaction)
		unset ::buttons::butinfo($vtsnum,1,1,1,1,coords)
		unset ::buttons::butinfo($vtsnum,1,1,1,1,uldr)
		unset ::buttons::butinfo($vtsnum,1,1,1,1,command)
		unset ::buttons::butinfo($vtsnum,1,1,2,grouptype)
		unset ::buttons::butinfo($vtsnum,1,1,3,grouptype)
	}

	::main::init_tree_menu
	::main::dvd_modified true false
	if {[lindex $::editor_cmds_from 0] == "menu"} {
		::main::display_pgc pgc 0 0
	}

	# if there is an old VOB, delete it or move it to backup
	delete_vob_with_confirm [file tail $vobfilename] true
}

#############################################################################
## Change the current menu type (Title, Root, Audio...) of the current PGC.
#############################################################################
## Change the current menu type (Title, Root, Audio...) of the current PGC.
proc set_menu_type {vtsnum lunum pgcnum type} {
	set numpgcs $::menus($vtsnum,$lunum,numpgcs)
	set oldmask $::menus($vtsnum,$lunum,$pgcnum,mask)

	switch $type {
		0 {set masktext ""}
		2 {set masktext "TitleM"}
		3 {set masktext "RootM"}
		4 {set masktext "SubPicM"}
		5 {set masktext "AudioM"}
		6 {set masktext "AngleM"}
		7 {set masktext "ChapterM"}
	}
	set newmask [expr ($oldmask & 0XF0) + $type]
	set ::menus($vtsnum,$lunum,$pgcnum,mask) $newmask
	set ::menus($vtsnum,$lunum,$pgcnum,masktext) $masktext

	if {$type > 0} {
		for {set pgc 1} {$pgc <= $::menus($vtsnum,$lunum,numpgcs)} {incr pgc} {
			if {$pgc == $pgcnum} continue
			if {[expr $::menus($vtsnum,$lunum,$pgc,mask) & 0X0F] == $type} {
				set ::menus($vtsnum,$lunum,$pgc,mask) \
						[expr $::menus($vtsnum,$lunum,$pgc,mask) & 0XF0]
				set ::menus($vtsnum,$lunum,$pgc,masktext) ""
			}
		}
	}
	::main::init_tree_menu
	::main::dvd_modified true false

	return
}
proc set_menu_type_GUI {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set utiltitle "Menu type"
	set pgctype [lindex $::editor_cmds_from 0]
	set vtsnum [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
	if {$pgctype != "menu"} {
		bell
		return
	}
	set lunum  [lindex [split [lindex $::editor_cmds_from 1] ","] 1]
	set pgcnum [lindex [split [lindex $::editor_cmds_from 1] ","] 2]
	set oldmask $::menus($vtsnum,$lunum,$pgcnum,mask)
	set ::utils::menutype_type [expr $oldmask & 0X0F]

	if [winfo exists .menutype] {destroy .menutype}
	set w .menutype
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w $utiltitle
	wm transient $w .
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	wm protocol $w WM_DELETE_WINDOW {set ::utils::menutype_done 0}
	bind $w <Escape> "set ::utils::menutype_done 0"

	label $w.lab -text "Menu type for PGC\n[$::widget(main_cmd_lab) cget -text]"
	pack  $w.lab
	if {$vtsnum == 0} {
		radiobutton $w.b1 -text "Title menu"   -variable ::utils::menutype_type -value 2 -anchor w
		pack $w.b1 -fill x
	} else {
		radiobutton $w.b1 -text "Root menu"    -variable ::utils::menutype_type -value 3 -anchor w
		pack $w.b1 -fill x
		radiobutton $w.b2 -text "SubPic menu"  -variable ::utils::menutype_type -value 4 -anchor w
		pack $w.b2 -fill x
		radiobutton $w.b3 -text "Audio menu"   -variable ::utils::menutype_type -value 5 -anchor w
		pack $w.b3 -fill x
		radiobutton $w.b4 -text "Angle menu"   -variable ::utils::menutype_type -value 6 -anchor w
		pack $w.b4 -fill x
		radiobutton $w.b5 -text "Chapter menu" -variable ::utils::menutype_type -value 7 -anchor w
		pack $w.b5 -fill x
	}
	radiobutton $w.b0 -text "None" -variable ::utils::menutype_type -value 0 -anchor w
	pack $w.b0 -fill x

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -anchor e
	button $w.fb.ok -text "OK" -width 10 -command {set ::utils::menutype_done 1}
	pack   $w.fb.ok -side left -padx 2
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::utils::menutype_done 0}
	pack   $w.fb.cancel -side left -padx 2

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr [winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)]
	set wy [expr [winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	wm geometry $w +$wx+$wy
	wm deiconify $w
	grab set $w

	tkwait variable ::utils::menutype_done
	grab release $w
	destroy $w

	if {$::utils::menutype_done} {
		set first $lunum
		set last  $lunum
		if {$::menus($vtsnum,numlus) > 1} {
			set txt "There are several LUs in this VTSM."
			set txt "$txt\n\nDo you want to modify the PGC type\nin each LU? (recommended)"
			set r [tk_messageBox -title $utiltitle -icon question -message $txt \
					-parent . -type yesno -default yes]
			if {$r == "yes"} {
				set first 1
				set last  $::menus($vtsnum,numlus)
			}
		}

		if [winfo exists .mf.f2.selector.a.lb] {
			set selectoroldpos [lindex [.mf.f2.selector.a.lb yview] 0]
		}

		for {set lu $first} {$lu <= $last} {incr lu} {
			if [catch {set_menu_type $vtsnum $lu $pgcnum $::utils::menutype_type}] {
				tk_messageBox -title $utiltitle -icon error -parent . -type ok \
						-message "ERROR: Can't find the PGC $pgcnum in LU $lu!"
			}
		}

		::main::display_pgc menu $vtsnum $lunum $pgcnum
		if [winfo exists .mf.f2.selector.a.lb] {
			.mf.f2.selector.a.lb yview moveto $selectoroldpos
		}
	}
	return
}

#############################################################################
## Swap two LUs
proc swap_lus {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	if {[lindex $::editor_cmds_from 0] != "menu"} {
		bell
		return
	}
	set vtsnum [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
	if {$::menus($vtsnum,numlus) < 2} {
		bell
		return
	}

	set ::utils::swaplus_1 [lindex [split [lindex $::editor_cmds_from 1] ","] 1]
	if {$::utils::swaplus_1 == 1} {
		set ::utils::swaplus_2 2
	} else {
		set ::utils::swaplus_2 1
	}

	set w .swaplus
	if [winfo exists $w] {destroy $w}
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w "Swap Menu Language Units"
	wm transient $w .
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	wm protocol $w WM_DELETE_WINDOW {set ::utils::swaplus_done 0}
	bind $w <Escape> "set ::utils::swaplus_done 0"

	label $w.lab -text "The first LU is the default on players\nwithout default language options."
	pack  $w.lab

	frame $w.f
	pack  $w.f -fill x -expand 1 -padx 2
	frame $w.f.f1 -bd 2 -relief groove
	pack  $w.f.f1 -side left -fill x -expand 1 -padx 2
	frame $w.f.f2 -bd 2 -relief groove
	pack  $w.f.f2 -side left -fill x -expand 1 -padx 2
	for {set lu 1} {$lu <= $::menus($vtsnum,numlus)} {incr lu} {
		radiobutton $w.f.f1.b$lu -text "LU $lu ($::menus($vtsnum,$lu,lang))" \
				-variable ::utils::swaplus_1 -value $lu -anchor w
		pack $w.f.f1.b$lu -fill x -expand 1
		radiobutton $w.f.f2.b$lu -text "LU $lu ($::menus($vtsnum,$lu,lang))" \
				-variable ::utils::swaplus_2 -value $lu -anchor w
		pack $w.f.f2.b$lu -fill x -expand 1
	}

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -anchor e
	button $w.fb.ok -text "OK" -width 10 -command {set ::utils::swaplus_done 1}
	pack   $w.fb.ok -side left -padx 2
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::utils::swaplus_done 0}
	pack   $w.fb.cancel -side left -padx 2

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [winfo reqwidth $w] [winfo reqheight $w]
	set wx [expr [winfo rootx .]+(([winfo width  .]-[winfo reqwidth  $w])/2)]
	set wy [expr [winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	wm geometry $w +$wx+$wy
	wm deiconify $w
	grab set $w

	set done false
	while {!$done} {
		set ::utils::swaplus_done -1
		tkwait variable ::utils::swaplus_done
		if {!$::utils::swaplus_done} {
			grab release $w
			destroy $w
			return
		}
		if {$::utils::swaplus_1 == $::utils::swaplus_2} {
			bell
		} else {
			set done true
		}
	}
	grab release $w
	destroy $w

	set lu_1 $::utils::swaplus_1
	set lu_2 $::utils::swaplus_2

	show_msg "Swapping LU $lu_1 ($::menus($vtsnum,$lu_1,lang)) with LU $lu_2 ($::menus($vtsnum,$lu_2,lang))..."

	if {$vtsnum == 0} {
		set table "VMGM_PGCI_UT"
	} else {
		set table "VTSM_PGCI_UT"
	}
	set idx1 [expr $lu_1 * 8]
	set idx2 [expr $lu_2 * 8]
	set luheader_1 [string range $::raw_sectors($vtsnum,$table) $idx1 [expr $idx1+7]]
	set luheader_2 [string range $::raw_sectors($vtsnum,$table) $idx2 [expr $idx2+7]]
	set ::raw_sectors($vtsnum,$table) [string replace $::raw_sectors($vtsnum,$table) \
			$idx1 [expr $idx1+7] $luheader_2]
	set ::raw_sectors($vtsnum,$table) [string replace $::raw_sectors($vtsnum,$table) \
			$idx2 [expr $idx2+7] $luheader_1]

	for {set pgc 1} {$pgc <= $::menus($vtsnum,$lu_1,numpgcs)} {incr pgc} {
		set tmp($pgc)				$::menus($vtsnum,$lu_1,$pgc)
		set tmp($pgc,mask)			$::menus($vtsnum,$lu_1,$pgc,mask)
		set tmp($pgc,masktext)		$::menus($vtsnum,$lu_1,$pgc,masktext)
		set tmp($pgc,header)		$::menus($vtsnum,$lu_1,$pgc,header)
		set tmp($pgc,numprograms)	$::menus($vtsnum,$lu_1,$pgc,numprograms)
		set tmp($pgc,programmap)	$::menus($vtsnum,$lu_1,$pgc,programmap)
		set tmp($pgc,numcells)		$::menus($vtsnum,$lu_1,$pgc,numcells)
		set tmp($pgc,cellpbtable)	$::menus($vtsnum,$lu_1,$pgc,cellpbtable)
		set tmp($pgc,cellpostable)	$::menus($vtsnum,$lu_1,$pgc,cellpostable)
	}
	for {set pgc 1} {$pgc <= $::menus($vtsnum,$lu_2,numpgcs)} {incr pgc} {
		set ::menus($vtsnum,$lu_1,$pgc)					$::menus($vtsnum,$lu_2,$pgc)
		set ::menus($vtsnum,$lu_1,$pgc,mask)			$::menus($vtsnum,$lu_2,$pgc,mask)
		set ::menus($vtsnum,$lu_1,$pgc,masktext)		$::menus($vtsnum,$lu_2,$pgc,masktext)
		set ::menus($vtsnum,$lu_1,$pgc,header)			$::menus($vtsnum,$lu_2,$pgc,header)
		set ::menus($vtsnum,$lu_1,$pgc,numprograms)		$::menus($vtsnum,$lu_2,$pgc,numprograms)
		set ::menus($vtsnum,$lu_1,$pgc,programmap)		$::menus($vtsnum,$lu_2,$pgc,programmap)
		set ::menus($vtsnum,$lu_1,$pgc,numcells)		$::menus($vtsnum,$lu_2,$pgc,numcells)
		set ::menus($vtsnum,$lu_1,$pgc,cellpbtable)		$::menus($vtsnum,$lu_2,$pgc,cellpbtable)
		set ::menus($vtsnum,$lu_1,$pgc,cellpostable)	$::menus($vtsnum,$lu_2,$pgc,cellpostable)
	}
	for {set pgc 1} {$pgc <= $::menus($vtsnum,$lu_1,numpgcs)} {incr pgc} {
		set ::menus($vtsnum,$lu_2,$pgc)					$tmp($pgc)
		set ::menus($vtsnum,$lu_2,$pgc,mask)			$tmp($pgc,mask)
		set ::menus($vtsnum,$lu_2,$pgc,masktext)		$tmp($pgc,masktext)
		set ::menus($vtsnum,$lu_2,$pgc,header)			$tmp($pgc,header)
		set ::menus($vtsnum,$lu_2,$pgc,numprograms)		$tmp($pgc,numprograms)
		set ::menus($vtsnum,$lu_2,$pgc,programmap)		$tmp($pgc,programmap)
		set ::menus($vtsnum,$lu_2,$pgc,numcells)		$tmp($pgc,numcells)
		set ::menus($vtsnum,$lu_2,$pgc,cellpbtable)		$tmp($pgc,cellpbtable)
		set ::menus($vtsnum,$lu_2,$pgc,cellpostable)	$tmp($pgc,cellpostable)
	}

	set tmpvar $::menus($vtsnum,$lu_1,lang)
	set ::menus($vtsnum,$lu_1,lang) $::menus($vtsnum,$lu_2,lang)
	set ::menus($vtsnum,$lu_2,lang) $tmpvar

	set tmpvar $::menus($vtsnum,$lu_1,numpgcs)
	set ::menus($vtsnum,$lu_1,numpgcs) $::menus($vtsnum,$lu_2,numpgcs)
	set ::menus($vtsnum,$lu_2,numpgcs) $tmpvar

	# refresh
	::main::init_tree_menu_no_move
	::main::dvd_modified true false
	show_msg ""
	return
}

#############################################################################
## Copy all commands from the current LU to all other LUs in same domain
proc copy_lu_to_lus {} {
	set utiltitle "Copy current LU PGCs commands to all other LUs"
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	if {[lindex $::editor_cmds_from 0] != "menu"} {
		bell
		return
	}
	set vtsnum [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
	if {$vtsnum == 0} {set domain "VMGM"} else {set domain "VTSM $vtsnum"}

	if {$::menus($vtsnum,numlus) < 2} {
		bell
		return
	}

	set source_lu [lindex [split [lindex $::editor_cmds_from 1] ","] 1]
	set numpgcs $::menus($vtsnum,1,numpgcs)

	if {!$::config(expert_mode)} {
		set txt "This utility will copy all commands of all PGCs"
		set txt "$txt\nof the current LU ($domain, LU $source_lu) to ALL other LUs."
		set txt "$txt\n\nThis allow to edit the commands in only one LU"
		set txt "$txt\nand then update the other LUs automatically."
		set txt "$txt\n\nNote that only the PGC commands are copied,"
		set txt "$txt\nnot the PGC parameters nor the PGC menu type."
		set txt "$txt\n\nContinue?"
		set r [tk_messageBox -icon question -title $utiltitle -parent . \
				-type okcancel -default ok -message $txt]
		if {$r == "cancel"} return
	}
	# sanity checks
	if {!$::menus($vtsnum,lus_are_identical)} {
		for {set lu 2} {$lu <= $::menus($vtsnum,numlus)} {incr lu} {
			if {$::menus($vtsnum,$lu,numpgcs) != $numpgcs} {
				set txt "Error: there are not the same number of PGCs"
				set txt "$txt\nin the different LUs of the current domain"
				set txt "$txt\n($domain)."
				set txt "$txt\n\nSorry!"
				tk_messageBox -icon error -title $utiltitle -parent . \
						-type ok -message $txt
				return
			}
		}
		set txt "Warning:\nWhen reading the DVD, PgcEdit has detected"
		set txt "$txt\nthat there were different commands"
		set txt "$txt\nin the various PGCs of the menu LUs,"
		set txt "$txt\nin the current domain ($domain)."
		set txt "$txt\n\nAre you sure you want to continue?"
		set r [tk_messageBox -icon warning -title $utiltitle -parent . \
				-type okcancel -default cancel -message $txt]
		if {$r == "cancel"} return
	}

	for {set lu 1} {$lu <= $::menus($vtsnum,numlus)} {incr lu} {
		if {$lu != $source_lu} {
			show_msg "Copying LU $source_lu ($::menus($vtsnum,$source_lu,lang)) to LU $lu ($::menus($vtsnum,$lu,lang))..."
			update idletasks
			for {set pgc 1} {$pgc <= $::menus($vtsnum,$source_lu,numpgcs)} {incr pgc} {
				set ::menus($vtsnum,$lu,$pgc) $::menus($vtsnum,$source_lu,$pgc)
			}
		}
	}
	set ::menus($vtsnum,lus_are_identical) true
	::main::dvd_modified true false
	show_msg "LU $source_lu ($::menus($vtsnum,$source_lu,lang)) commands copied OK."
	return
}


#############################################################################
## Create a new PGC in the current domain
proc new_pgc {quiet multiple} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	set utiltitle "Create New PGC"

	set pgctype [lindex $::editor_cmds_from 0]
	set vtsnum [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
	# NTSC or PAL 0 length time
	if [info exists ::raw_sectors(1,VTSI_MAT)] {
		if {[expr [get_int ::raw_sectors(1,VTSI_MAT) 0x200 2] & 0x3000] == 0} {
			set cellpbtime0 192
		} else {
			set cellpbtime0 64
		}
	} else {
		if {[expr [get_int ::raw_sectors(0,VMGM_MAT) 0x100 2] & 0x3000] == 0} {
			set cellpbtime0 192
		} else {
			set cellpbtime0 64
		}
	}

	if {$pgctype == "menu"} {
		set lunum [lindex [split [lindex $::editor_cmds_from 1] ","] 1]
		set numpgcs $::menus($vtsnum,$lunum,numpgcs)
		set pgcnum [expr $numpgcs + 1]
	} else {
		if {$vtsnum == 0} {
			set txt "Can't add a new PGC to the First Play PGC!"
			set txt "$txt\nPlease select another PGC."
			show_warning -title $utiltitle -icon error -message $txt \
					-parent .
			return false
		}
		set numpgcs $::pgcs($vtsnum,numpgcs)
		set pgcnum [expr $numpgcs + 1]
	}

	if {! $quiet} {
		set txt "This utility will create a new, dummy PGC"
		set txt "$txt\nat the end of the current LU, or one PGC"
		set txt "$txt\nat the end of each LU of the current VTS"
		set txt "$txt\nor at the end of the current Title domain."
		set txt "$txt\n"
		set txt "$txt\nShould I create the following PGC?"
		if {$pgctype == "menu"} {
			if {$vtsnum == 0} {
				set txt "$txt\nVMGM"
			} else {
				set txt "$txt\nVTSM $vtsnum"
			}
			set txt "$txt , LU $lunum , $pgcnum"
			set txt "$txt\n\n(If there are several LUs, you will be prompted"
			set txt "$txt\nlater for multiple PGC creation at once)"
		} else {
			set txt "$txt\nVTST $vtsnum , $pgcnum"
			set txt "$txt\n\nThe dummy PGC will be 'attached' to the current"
			set txt "$txt\nTitle (ie. it will inherit the current Title and"
			set txt "$txt\nTTN numbers.)"
		}
		if {!$::config(expert_mode)} {
			set r [tk_messageBox -title $utiltitle -icon question -message $txt \
					-parent . -type yesno -default yes]
			if {$r == "no"} {return false}
		}
	}

	if {$pgctype == "menu"} {
		set first $lunum
		set last  $lunum
		if {$::menus($vtsnum,numlus) > 1} {
			if {! $quiet} {
				set txt "There are several LUs in this VTSM."
				set txt "$txt\n\nDo you want to create a new PGC in each LU?\n(recommended)"
				set r [tk_messageBox -title $utiltitle -icon question -message $txt \
						-parent . -type yesno -default yes]
			} else {
				if $multiple {set r "yes"} else {set r "no"}
			}
			if {$r == "yes"} {
				set first 1
				set last  $::menus($vtsnum,numlus)
			}
		}

		for {set lu $first} {$lu <= $last} {incr lu} {
			set pgcn [expr $::menus($vtsnum,$lu,numpgcs) + 1]
			set ::menus($vtsnum,$lu,$pgcn,mask) 0
			set ::menus($vtsnum,$lu,$pgcn,masktext) ""
			set ::menus($vtsnum,$lu,$pgcn) {{* pre} {00 00 00 00 00 00 00 00} {* post} {* cell}}
			set ::menus($vtsnum,$lu,$pgcn,header) [string repeat \000 236]
			set_int ::menus($vtsnum,$lu,$pgcn,header) 4 $cellpbtime0 4
			set ::menus($vtsnum,$lu,$pgcn,numprograms) 0
			set ::menus($vtsnum,$lu,$pgcn,programmap) ""
			set ::menus($vtsnum,$lu,$pgcn,numcells) 0
			set ::menus($vtsnum,$lu,$pgcn,cellpbtable) ""
			set ::menus($vtsnum,$lu,$pgcn,cellpostable) ""

			# update V**M_PGCI_UT table
			if {$vtsnum == 0} {
				set tablename "VMGM_PGCI_UT"
			} else {
				set tablename "VTSM_PGCI_UT"
			}
			set luidx [get_int ::raw_sectors($vtsnum,$tablename) [expr ($lu*8) + 4] 4]
			set idx [expr ($::menus($vtsnum,$lu,numpgcs) * 8) + $luidx + 8]
			set part1 [string range $::raw_sectors($vtsnum,$tablename) 0 [expr $idx-1]]
			set part2 [string range $::raw_sectors($vtsnum,$tablename) $idx end]
			set ::raw_sectors($vtsnum,$tablename) $part1[string repeat \000 8]$part2

			incr ::menus($vtsnum,$lu,numpgcs)
		}
		::main::init_tree_menu_no_move
		::main::display_pgc menu $vtsnum $lunum $pgcnum
	} else {
		set ::pgcs($vtsnum,$pgcnum,mask) 0
		set ::pgcs($vtsnum,$pgcnum,masktext) ""
		set ::pgcs($vtsnum,$pgcnum) {{* pre} {00 00 00 00 00 00 00 00} {* post} {* cell}}
		set ::pgcs($vtsnum,$pgcnum,header) [string repeat \000 236]
		set_int ::pgcs($vtsnum,$pgcnum,header) 4 $cellpbtime0 4
		set ::pgcs($vtsnum,$pgcnum,numprograms) 0
		set ::pgcs($vtsnum,$pgcnum,programmap) ""
		set ::pgcs($vtsnum,$pgcnum,numcells) 0
		set ::pgcs($vtsnum,$pgcnum,cellpbtable) ""
		set ::pgcs($vtsnum,$pgcnum,cellpostable) ""
		set ::pgcs($vtsnum,$pgcnum,pttlist) ""

		if [info exists ::pgcs([lindex $::editor_cmds_from 1],TTN)] {
			set ttnnum $::pgcs([lindex $::editor_cmds_from 1],TTN)
		} else {
			set ttnnum 1
		}
		set ::pgcs($vtsnum,$pgcnum,TTN) $ttnnum	;# bug fixed

		# update VTS_PGCI table
		set idx [expr ($::pgcs($vtsnum,numpgcs) * 8) + 8]
		set part1 [string range $::raw_sectors($vtsnum,VTS_PGCI) 0 [expr $idx-1]]
		set part2 [string range $::raw_sectors($vtsnum,VTS_PGCI) $idx end]
		set ::raw_sectors($vtsnum,VTS_PGCI) $part1[string repeat \000 8]$part2
		set_int ::raw_sectors($vtsnum,VTS_PGCI) $idx $ttnnum 1

		incr ::pgcs($vtsnum,numpgcs)

		::main::init_tree_menu_no_move
		::main::display_pgc pgc $vtsnum $pgcnum
	}
	::main::dvd_modified true false
	return true
}

proc clone_pgc {} {
	if {![info exists ::dvddir]} {bell; return}

	set cur_type  [lindex $::editor_cmds_from 0]
	set cur_index [lindex $::editor_cmds_from 1]
	set cur_vts   [lindex [split $cur_index ","] 0]

	if {! $::config(expert_mode)} {
		show_msg ""
		set txt "This utility will clone the current PGC.\n\n"
		if {$cur_type == "pgc"} {
			if {$::pgcs($cur_index,numprograms) > 0} {
				set txt "${txt}A new TTN number will be assigned to the new PGC.\n"
				set txt "${txt}Optionally, a new Title will also be assigned.\n"
			} else {
				set txt "${txt}As the current PGC is a dummy PGC, the new PGC\n"
				set txt "${txt}will inherit the current TTN number.\n"
			}
		}
		set txt "${txt}\nContinue?"
		set r [tk_messageBox -title "Clone PGC" -icon question \
				-message $txt -parent . -type okcancel -default ok]
		if {$r == "cancel"} return
	}


	if {$cur_type == "menu"} {
		set cur_lu [lindex [split $cur_index ,] 1]
		set cur_numpgcs $::menus($cur_vts,$cur_lu,numpgcs)
		set cur_pgcnum [lindex [split $cur_index ,] 2]
		set new_pgcnum [expr $cur_numpgcs + 1]
		set new_index $cur_vts,$cur_lu,$new_pgcnum

		if {$::menus($cur_vts,numlus) > 1} {
			set txt "There are several LUs in this VTSM."
			set txt "$txt\n\nDo you want to clone the PGC in each LU?\n(recommended)"
			set r [tk_messageBox -title "Clone PGC" -icon question -message $txt \
					-parent . -type yesnocancel -default yes]
			if {$r == "yes"} {
				set first 1
				set last  $::menus($cur_vts,numlus)
			} else {
				set ::menus($cur_vts,lus_are_identical) false
				set first $cur_lu
				set last  $cur_lu
			}
			if {$r == "cancel"} {
				return
			}
		} else {
			set first $cur_lu
			set last  $cur_lu
		}
	} else {
		if {$cur_vts == 0} {bell; return}
		set cur_numpgcs $::pgcs($cur_vts,numpgcs)
		set cur_pgcnum [lindex [split $cur_index ,] 1]
		set new_pgcnum [expr $cur_numpgcs + 1]
		set new_index $cur_vts,$new_pgcnum
	}

	show_msg "Creating New PGC $new_pgcnum..."

	# copy PGC
	if {$cur_type == "menu"} {
		for {set lu $first} {$lu <= $last} {incr lu} {
			if {! [info exists ::menus($cur_vts,$lu,$cur_pgcnum)]} {
				tk_messageBox -title "Clone PGC" -icon warning -parent . -type ok \
						-message "WARNING!\nPGC $cur_pgcnum do not exist in LU $lu.\nIt will not be cloned!"
				set ::menus($cur_vts,lus_are_identical) false
				break
			}
			incr ::menus($cur_vts,$lu,numpgcs)
			set newpgc $::menus($cur_vts,$lu,numpgcs)
	
			set ::menus($cur_vts,$lu,$newpgc) \
					$::menus($cur_vts,$lu,$cur_pgcnum)
			set ::menus($cur_vts,$lu,$newpgc,header) \
					$::menus($cur_vts,$lu,$cur_pgcnum,header)
			set ::menus($cur_vts,$lu,$newpgc,numprograms) \
					$::menus($cur_vts,$lu,$cur_pgcnum,numprograms)
			set ::menus($cur_vts,$lu,$newpgc,numcells) \
					$::menus($cur_vts,$lu,$cur_pgcnum,numcells)
			set ::menus($cur_vts,$lu,$newpgc,cellpbtable) \
					$::menus($cur_vts,$lu,$cur_pgcnum,cellpbtable)
			set ::menus($cur_vts,$lu,$newpgc,cellpostable) \
					$::menus($cur_vts,$lu,$cur_pgcnum,cellpostable)
			set ::menus($cur_vts,$lu,$newpgc,programmap) \
					$::menus($cur_vts,$lu,$cur_pgcnum,programmap)
	
			set ::menus($cur_vts,$lu,$newpgc,mask)			0
			set ::menus($cur_vts,$lu,$newpgc,masktext)		""
			set ::menus($cur_vts,$lu,$newpgc,numbuttons) \
					$::menus($cur_vts,$lu,$cur_pgcnum,numbuttons)
	
			######### TODO:
			# update V**M_PGCI_UT table
			if {$cur_vts == 0} {
				set tablename "VMGM_PGCI_UT"
			} else {
				set tablename "VTSM_PGCI_UT"
			}
			set luidx [get_int ::raw_sectors($cur_vts,$tablename) [expr ($cur_lu*8) + 4] 4]
			set idx [expr ($::menus($cur_vts,$cur_lu,numpgcs) * 8) + $luidx + 8]
			set origentry [string range $::raw_sectors($cur_vts,$tablename) $idx [expr $idx+7]]
			set_int origentry 0 0 1	;# clear meny type
	
			set luidx [get_int ::raw_sectors($cur_vts,$tablename) [expr ($cur_lu*8) + 4] 4]
			set idx [expr ($::menus($cur_vts,$lu,numpgcs) * 8) + $luidx + 8]
			set part1 [string range $::raw_sectors($cur_vts,$tablename) 0 [expr $idx-1]]
			set part2 [string range $::raw_sectors($cur_vts,$tablename) $idx end]
			set ::raw_sectors($cur_vts,$tablename) $part1$origentry$part2
		}
	} else {
		set ::pgcs($cur_vts,numpgcs) $new_pgcnum
		if {$::pgcs($cur_index,numprograms) > 0} {
			set dummypgc false
		} else {
			set dummypgc true
		}

		set ::pgcs($new_index)				$::pgcs($cur_index)
		set ::pgcs($new_index,header)		$::pgcs($cur_index,header)
		set ::pgcs($new_index,numprograms)	$::pgcs($cur_index,numprograms)
		set ::pgcs($new_index,numcells)		$::pgcs($cur_index,numcells)
		set ::pgcs($new_index,cellpbtable)	$::pgcs($cur_index,cellpbtable)
		set ::pgcs($new_index,cellpostable) $::pgcs($cur_index,cellpostable)
		set ::pgcs($new_index,programmap)	$::pgcs($cur_index,programmap)

		set ::pgcs($new_index,pttlist)		$::pgcs($cur_index,pttlist)

		if $dummypgc {
			set ttn $::pgcs($cur_index,TTN)
		} else {
			set ttn 0
			for {set i 1} {$i < $new_pgcnum} {incr i} {
				if {[info exists ::pgcs($cur_vts,$i,TTN)] && $::pgcs($cur_vts,$i,TTN) > $ttn} {
					set ttn $::pgcs($cur_vts,$i,TTN)
				}
			}
			incr ttn
		}
		set ::pgcs($new_index,TTN) $ttn

		# update VTS_PGCI table
		set idx [expr ($new_pgcnum * 8)]
		set part1 [string range $::raw_sectors($cur_vts,VTS_PGCI) 0 [expr $idx-1]]
		set part2 [string range $::raw_sectors($cur_vts,VTS_PGCI) $idx end]
		set ::raw_sectors($cur_vts,VTS_PGCI) $part1[string repeat \000 8]$part2
		set_int ::raw_sectors($cur_vts,VTS_PGCI) $idx [expr $ttn+128] 1

		### build chapter table
		if {! $dummypgc} {
			set donewtitle [tk_messageBox -title "Clone PGC" -icon question \
					-message "Create a new Title number?\n(This will add an entry in the VMG_TT_SRPT table)" \
					-parent . -type yesnocancel -default yes]
			if {$donewtitle == "cancel"} {
				return
			}

			set ptttable $::raw_sectors($cur_vts,VTS_PTT_SRPT)
			set pttnumtitles [get_int ptttable 0 2]
			set pttheader [string range $ptttable 0 [expr ($pttnumtitles*4) + 7]]
			set pttrest [string range $ptttable [expr ($pttnumtitles*4) + 8] end]
	
			# get start and end index in the PTT table for the old PTTs.
			set startidx [get_int pttheader [expr ($::pgcs($cur_index,TTN) * 4)+4] 4]
			if {[expr ($::pgcs($cur_index,TTN) * 4)+8] >= [string length $pttheader]} {
				set endidx end
			} else {
				set endidx [expr [get_int pttheader [expr ($::pgcs($cur_index,TTN) * 4)+8] 4] -1]
			}
			set newptts [string range $ptttable $startidx $endidx]
	
			# add entry to the PTT table header
			set idx [string length $pttheader]
			set pttheader $pttheader\000\000\000\000
			set_int pttheader $idx [expr [string length $pttheader]+[string length $pttrest]] 4
			for {set i 0} {$i < $pttnumtitles} {incr i} {
				set idx [expr ($i * 4)+8]
				set_int pttheader $idx [expr [get_int pttheader $idx 4] +4] 4
			}
			set ptttable $pttheader$pttrest$newptts
			set_int ptttable 0 $::pgcs($new_index,TTN) 2
			set_int ptttable 4 [expr [string length $ptttable]-1] 4
			set ::raw_sectors($cur_vts,VTS_PTT_SRPT) $ptttable
	
			if {$donewtitle == "yes"} {
				# Add the new title to the VMG global table of titles
				set data $::raw_sectors(0,VMG_TT_SRPT)
				set srptnumtitles [get_int data 0 2]
				set data [string range $data 0 [expr ($srptnumtitles*12)+7]]
				incr srptnumtitles
				set_int data 0 $srptnumtitles 2
				set oldtitle ""
				if {[info exists ::pgcs($cur_index,Title)] && $::pgcs($cur_index,Title) != ""} {
					set oldtitle $::pgcs($cur_index,Title)
				}
				if {$oldtitle == ""} {
					for {set p 1} {$p <= $::pgcs($cur_vts,numpgcs)} {incr p} {
						if {$::pgcs($cur_vts,$p,TTN) == $::pgcs($cur_index,TTN) &&
								[info exists ::pgcs($cur_vts,$p,Title)] && $::pgcs($cur_vts,$p,Title) != ""} {
							set oldtitle $::pgcs($cur_vts,$p,Title)
							break
						}
					}
				}
				if {$oldtitle != ""} {
					set newentry [string range $::raw_sectors(0,VMG_TT_SRPT) \
							[expr ($oldtitle*12)-4] [expr ($oldtitle*12)+8]]
				} else {
					set newentry \024\001\000\000\000\000\000\000\000\000\000\000
					set_int newentry 6 $cur_vts 1
				}
				set_int newentry 2 [llength $::pgcs($new_index,pttlist)] 2
				set_int newentry 7 $::pgcs($new_index,TTN) 1
				set data $data$newentry
				set_int data 4 [expr [string length $data]-1] 4
				set ::raw_sectors(0,VMG_TT_SRPT) $data
				set ::pgcs($new_index,Title) $srptnumtitles
			}
		}
	}
	# refresh GUI
	::main::dvd_modified true false
	::main::init_tree_menu

	if {$cur_type == "menu"} {
		::main::display_pgc menu $cur_vts $cur_lu $new_pgcnum
	} else {
		::main::display_pgc pgc $cur_vts $new_pgcnum
	}
	show_msg "PGC cloned OK."
	return
}

#############################################################################
## Procedure:  new_DVD
## Create a new DVD wuth an empty VMG
proc {new_DVD} {} {
	set utiltitle "New DVD"

	if {!$::config(expert_mode)} {
		set txt "This utility will create a new VIDEO_TS.IFO file"
		set txt "$txt\nin the specified folder."
		set txt "$txt\nThe new DVD will have no VTST Titles, nor menus."
		set txt "$txt\nYou must therefore import titles to the new DVD,"
		set txt "$txt\nwith 'Import VTST Titles', and modify the PGCs"
		set txt "$txt\ncommands to suit your needs."
		set txt "$txt\n\nContinue?"
		set r [tk_messageBox -title $utiltitle -icon question -message $txt \
				-parent . -type yesno -default yes]
		if {$r == "no"} {return -1}
	}

	if {$::dvd_modified && !$::config(expert_mode)} {
		set r [tk_messageBox -message "The DVD has been modified.\n\nAre you sure?" \
				-icon warning -title "PgcEdit" -type yesno -default no]
		if {$r == "no"} {return 0}
	}

	set destdir [file normalize \
			[tk_chooseNewDirectory -initialdir $::config(initialdir) \
			-title "$utiltitle: Select or create the EMPTY target folder..."]]
	if {$destdir == ""} {return -1}
	if [file exists [file join $destdir "VIDEO_TS.IFO"]] {
		set r [tk_messageBox -title $utiltitle -icon question \
				-message "File [file join $destdir "VIDEO_TS.IFO"] already exists!\nOverwrite ?" \
				-parent . -type yesno -default no]
		if {$r == "no"} {return -1}
	}

	# copy the new VIDEO_TS.IFO to the destination folder
	set tempdata [readFile [file join $::config(apppath) "lib/data/new_dvd_video_ts.ifo"] binary]
	set err [saveFile [file join $destdir "VIDEO_TS.IFO"] binary $tempdata]
	if {$err != 0} {return $err}
	set err [saveFile [file join $destdir "VIDEO_TS.BUP"] binary $tempdata]
	if {$err != 0} {return $err}

	set ::dvd_modified false
	return [::main::Read_New_DVD $destdir]
}
proc copy_from_data_dir {source destination} {
	set data [readFile $::config(apppath)/lib/data/$source binary]
	saveFile [$destination binary $data]
}

#############################################################################
## Procedure:  set_provider_ID
## Show/Change the provider ID
proc {set_provider_ID} {} {
	if [info exists ::raw_sectors(0,VMGM_MAT)] {
		# Get old Provider ID (if any)
		set oldid [get_provider_ID]
		# get new ID
		set newid [::stringreq::stringreq -window .getid -def $oldid -ok "Set ID" \
				-callback ::utils::set_provider_ID_callback -width 32 \
				-text "Enter the DVD Provider ID:" -title "PgcEdit" -geometry w \
				-cancelreturn "***CANCELLED***"]
		if {$newid != "***CANCELLED***"} {
			set newid [string trim $newid " "]
			set newid "$newid[string repeat \000 [expr 32-[string length $newid]]]"
			set ::raw_sectors(0,VMGM_MAT) \
					[string replace $::raw_sectors(0,VMGM_MAT) 0X40 0X5F $newid]
			::main::set_window_title
			::main::dvd_modified true false
		}
	} else {
		bell
	}
}
proc {set_provider_ID_callback} {str} {
	if {[string length $str] > 32} {
		bell
		return 0
	} else {
		return 1
	}
}
proc {get_provider_ID} {} {
	if [info exists ::raw_sectors(0,VMGM_MAT)] {
		set id [string range $::raw_sectors(0,VMGM_MAT) 0X40 0X5F]
		set stopchar [string first \000 $id]
		if {$stopchar != -1} {
			set id [string range $id 0 [expr $stopchar - 1]]
		}
		return [string trim $id " "]
	} else {
		return ""
	}
}

#############################################################################
## Procedure:  set_General_Name
## Show/Set/Change the DVD-Text General Name
proc {set_General_Name_GUI} {} {
	set utiltitle "Set DVD-Text General Name and Movie Name"
	if {![info exists ::dvddir]} {
		bell
		return
	}
	if {$::raw_sectors(0,VMG_TXTDT_MG) != ""} {
		set newsec false
		set newmsg "VMG_TXTDT_MG (DVD-TEXT data sector) already exists."
		set oldid [get_General_Name]
		set maxlen [get_General_Name_max_length]
		if {$maxlen < 1} {
			set text "Can't find the DVD-TEXT General Name or Movie Name.\n\n"
			set text "${text}If you really want to proceed, you need to remove\n"
			set text "${text}the current DVD-TEXT data sector and retry..."
			tk_messageBox -title $utiltitle -icon warning -message $text
			return
		}
	} else {
		set newsec true
		set newmsg "The VMG_TXTDT_MG (DVD-TEXT data sector) will be created."
		set ::raw_sectors(0,VMG_TXTDT_MG) \
				[readFile $::config(apppath)/lib/data/vmg_txtdt_mg_default.sec binary]
		set oldid ""
		set maxlen 30
	}

	# get new ID
	set newid [::stringreq::stringreq -window .getid -def $oldid -ok "Set Name" \
			-callback "::utils::set_general_name_callback $maxlen" -width $maxlen \
			-text "$newmsg\n\nEnter the DVD-TEXT General Name\n(Max $maxlen characters)" \
			-title $utiltitle -geometry w \
			-cancelreturn "***CANCELLED***"]
	if {$newid != "***CANCELLED***"} {
		set_General_Name $newid $maxlen
		::main::set_window_title
		::main::dvd_modified true false
	} elseif $newsec {
		set ::raw_sectors(0,VMG_TXTDT_MG) ""
	}
}
proc {set_general_name_callback} {maxlen str} {
	if {[string length $str] > $maxlen} {
		bell
		return 0
	} else {
		return 1
	}
}
# return one of the Volume General Name, Volume Movie Name,
# Title 1 General Name or Title 1 Movie Name
proc {get_General_Name} {} {
	if {![info exists ::dvddir]} {return ""}
	if {$::raw_sectors(0,VMG_TXTDT_MG) == ""} {return ""}

	set name ""
	set numentries [get_int ::raw_sectors(0,VMG_TXTDT_MG) 0xE8 2]
	set type 0
	set volume_name_idx  0
	set volume_movie_idx 0
	set title_name_idx   0
	set title_movie_idx  0
	for {set i 0} {$i < $numentries} {incr i} {
		set idx [expr 0xEC + ($i * 4)]
		set entry [get_int ::raw_sectors(0,VMG_TXTDT_MG) $idx 1]
		if {$entry < 48} {set type $entry; continue}
		incr idx 2
		if {$type == 1 && $entry == 48} {
			set volume_name_idx [get_int ::raw_sectors(0,VMG_TXTDT_MG) $idx 2]
			continue
		}
		if {$type == 1 && $entry == 57} {
			set volume_movie_idx [get_int ::raw_sectors(0,VMG_TXTDT_MG) $idx 2]
			continue
		}
		if {$type == 2 && $entry == 48 && $title_name_idx == 0} {
			set title_name_idx [get_int ::raw_sectors(0,VMG_TXTDT_MG) $idx 2]
			continue
		}
		if {$type == 2 && $entry == 57 && $title_movie_idx == 0} {
			set title_movie_idx [get_int ::raw_sectors(0,VMG_TXTDT_MG) $idx 2]
			continue
		}
	}
	set nameidx 0
	if {$title_movie_idx > 0}  {set nameidx $title_movie_idx}
	if {$title_name_idx > 0}   {set nameidx $title_name_idx}
	if {$volume_movie_idx > 0} {set nameidx $volume_movie_idx}
	if {$volume_name_idx > 0}  {set nameidx $volume_name_idx}
	if {$nameidx == 0} {return ""}

	incr nameidx 0xE8
	set name [string range $::raw_sectors(0,VMG_TXTDT_MG) $nameidx [expr $nameidx + 32]]
	set stopchar [string first \000 $name]
	if {$stopchar != -1} {
		set name [string range $name 0 [expr $stopchar - 1]]
	}
	set stopchar [string first \011 $name]
	if {$stopchar != -1} {
		set name [string range $name 0 [expr $stopchar - 1]]
	}
	set name [string trim $name " "]
	return $name
}

proc {get_General_Name_max_length} {} {
	if {![info exists ::dvddir]} {return 0}

	set numentries [get_int ::raw_sectors(0,VMG_TXTDT_MG) 0xE8 2]
	set type 0
	set volume_name_idx  0
	set volume_movie_idx 0
	for {set i 0} {$i < $numentries} {incr i} {
		set idx [expr 0xEC + ($i * 4)]
		set entry [get_int ::raw_sectors(0,VMG_TXTDT_MG) $idx 1]
		if {$entry < 48} {set type $entry; continue}
		incr idx 2
		if {$type == 1 && $entry == 48} {
			set volume_name_idx [get_int ::raw_sectors(0,VMG_TXTDT_MG) $idx 2]
			continue
		}
		if {$type == 1 && $entry == 57} {
			set volume_movie_idx [get_int ::raw_sectors(0,VMG_TXTDT_MG) $idx 2]
			continue
		}
	}
	if {$volume_movie_idx == 0 && $volume_name_idx == 0} {bell; return 0}
	set max1 30
	set max2 30
	if {$volume_movie_idx > 0} {
		set max1 [get_General_Name_max_length_sub [expr $volume_movie_idx + 0xE8]]
	}
	if {$volume_name_idx > 0}  {
		set max2 [get_General_Name_max_length_sub [expr $volume_name_idx + 0xE8]]
	}

	if {$max1 <= $max2} {
		return $max1
	} else {
		return $max2
	}
}
proc get_General_Name_max_length_sub {nameidx} {
	set name [string range $::raw_sectors(0,VMG_TXTDT_MG) $nameidx [expr $nameidx + 30]]
	set stopchar1 [string first \000 $name]
	if {$stopchar1 == -1} {set stopchar1 30}
	set stopchar2 [string first \011 $name]
	if {$stopchar2 == -1} {set stopchar2 30}
	if {$stopchar1 < $stopchar2} {
		set stopchar $stopchar1
	} else {
		set stopchar $stopchar2
	}
	for {set i [expr $stopchar+1]} {$i < 30} {incr i} {
		if {[string range ${name} $i $i] != "\000"} {
			return [expr $i-1]
		}
	}
	return 30
}

proc {set_General_Name} {name maxlen} {
	if {![info exists ::dvddir]} {return 0}

	set numentries [get_int ::raw_sectors(0,VMG_TXTDT_MG) 0xE8 2]
	set type 0
	set volume_name_idx  0
	set volume_movie_idx 0
	for {set i 0} {$i < $numentries} {incr i} {
		set idx [expr 0xEC + ($i * 4)]
		set entry [get_int ::raw_sectors(0,VMG_TXTDT_MG) $idx 1]
		if {$entry < 48} {set type $entry; continue}
		incr idx 2
		if {$type == 1 && $entry == 48} {
			set volume_name_idx [get_int ::raw_sectors(0,VMG_TXTDT_MG) $idx 2]
			continue
		}
		if {$type == 1 && $entry == 57} {
			set volume_movie_idx [get_int ::raw_sectors(0,VMG_TXTDT_MG) $idx 2]
			continue
		}
	}
	if {$volume_movie_idx == 0 && $volume_name_idx == 0} {bell; return 0}
	if {$volume_movie_idx > 0} {
		set_general_name_at_address $name [expr $volume_movie_idx + 0xE8] $maxlen
	}
	if {$volume_name_idx > 0}  {
		set_general_name_at_address $name [expr $volume_name_idx + 0xE8] $maxlen
	}
	return 1
}
proc set_general_name_at_address {name address maxlen} {
	set name $name\011
	incr maxlen
	while {[string length $name] < $maxlen} {set name $name\000}

	set data1 [string range $::raw_sectors(0,VMG_TXTDT_MG) 0 [expr $address-1]]
	set data2 [string range $::raw_sectors(0,VMG_TXTDT_MG) [expr $address+$maxlen] end]
	set ::raw_sectors(0,VMG_TXTDT_MG) $data1$name$data2
}

#############################################################################
## Remove the DVD-TEXT datas (the whole sector is erased)
proc {remove_dvd_text_data} {} {
	if {![info exists ::dvddir]} {
		bell
		return
	}
	if {$::raw_sectors(0,VMG_TXTDT_MG) != ""} {
		if {!$::config(expert_mode)} {
			set msg "This will remove the whole VMG_TXTDT_MG\n(DVD-TEXT data sector), not only the\nGeneral Name!\n\nAre you sure ?"
			set r [tk_messageBox -title "Remove DVD-TEXT data" -icon question -message $msg \
					-parent . -type okcancel]
		} else {
			set r "ok"
		}
		if {$r == "ok"} {
			set ::raw_sectors(0,VMG_TXTDT_MG) ""
			::main::set_window_title
			::main::dvd_modified true false
		}
	} else {
		show_warning -title "Remove DVD-TEXT data" -icon error -parent . \
				-message "There is no VMG_TXTDT_MG (DVD-TEXT\ndata sector) in this DVD."
	}
}

#############################################################################
## Delete unused VOB file or move to backup or (optionally) leave it.
## Args: the VOB filename, without path.
##       leave_ok boolean: Add the "Leave it there" button for no action.
proc delete_vob_with_confirm {filename leave_ok} {
	set filename [file tail $filename]
	set vobfile [file join $::dvddir $filename]

	if [file exists $vobfile] {
		set txt "Warning: there is an old $filename"
		set txt "$txt\nin your DVD folder."
		set txt "$txt\n\nWhat do you want to do?"
		if {$leave_ok} {
			set r [tk_dialog .confirm "PgcEdit" $txt "warning" 0 \
					" Backup file " " Delete file " "Leave it there"]
		} else {
			set r [tk_dialog .confirm "PgcEdit" $txt "warning" 0 \
					" Backup file " " Delete file "]
		}
		switch $r {
			0 {
				set backupdir [file join $::dvddir "PgcEdit_backup"]
				if {! [file exists $backupdir]} {
					file mkdir $backupdir
				}
				if [file exists $backupdir] {
					set doit true
					if [file exists [file join $backupdir $filename]] {
						set txt "Warning: There is already a $filename"
						set txt "$txt\nin the backup folder.\n\nOverwrite?"
						set txt "$txt\n\nIf you answer No, the current $filename"
						set txt "$txt\nwill be erased without backup."
						set r [tk_messageBox -title "PgcEdit" -icon warning \
								-message $txt -parent . -type yesno -default no]
						if {$r == "no"} {
							set doit false
						}
					}
					if $doit {
						file rename -force $vobfile $backupdir
					} else {
						file delete $vobfile
					}
				}
			}
			1 {
				file delete $vobfile
			}
		}
	}
}


#########################################################
# Build Calls cross references table
proc build_crossrefs {only_vts} {
	show_msg "Building cross references..."
	if {![info exists ::crossrefs(pgcs,0,0,numrefs)]} {
		for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
			for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
				for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
					set ::crossrefs(menus,$vts,$lu,$pgc,numrefs) 0
				}
			}
			for {set pgc 0} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				set ::crossrefs(pgcs,$vts,$pgc,numrefs) 0
			}
		}
	}
	for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
		if {$only_vts != -1 && $vts != 0 && $vts != $only_vts} continue
		if [info exists ::crossrefs_calculated($vts)] continue
		if {$vts == 0} {set name "VMGM"} else {set name "VTSM $vts"}
		show_msg "Building cross references: $name..."
		for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
			for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
				add_link_refs "menu" $vts $lu $pgc

				set numcmds [llength $::menus($vts,$lu,$pgc)]
				set linenum 0
				set section "pre"
				for {set c 1} {$c < $numcmds} {incr c} {
					set cmd [lindex $::menus($vts,$lu,$pgc) $c]
					if {[lindex $cmd 0] == "*"} {
						set linenum 0
						set section [lindex $cmd 1]
					} else {
						incr linenum
						set r [::main::highlight_jump_cmd menu $vts,$lu,$pgc [string toupper \[$cmd\]] [expr $c+1] false false]
						#set debugout "$debugout\nmenus $vts $lu $pgc $section $linenum \[$cmd\] : $r"
						add_crossref $r [list "menus" $vts $lu $pgc $section $linenum $cmd]
					}
				}

				if {$::tcl_platform(platform) != "macintosh"} {
					set numcells $::menus($vts,$lu,$pgc,numcells)
					set cellpostable $::menus($vts,$lu,$pgc,cellpostable)
					for {set cell 0} {$cell < $numcells} {incr cell} {
						set idx [expr $cell * 4]
						set vobid [get_int cellpostable $idx 2]
						incr idx 3
						set cellid [get_int cellpostable $idx 1]
						if [info exists ::buttons::butinfo($vts,$vobid,$cellid,numbuts)] {
							set numbuts $::buttons::butinfo($vts,$vobid,$cellid,numbuts)
							if {$numbuts > 0} {
								set numgroups $::buttons::butinfo($vts,$vobid,$cellid,numgroups)
								for {set butnum 1} {$butnum <= $numbuts} {incr butnum} {
									for {set group 1} {$group <= $numgroups} {incr group} {
										set cmd $::buttons::butinfo($vts,$vobid,$cellid,$group,$butnum,command)
										set r [::main::highlight_jump_cmd menu $vts,$lu,$pgc \[$cmd\] [expr $c+1] false false]
										add_crossref $r [list "menu_buttons" $vts $lu $pgc [expr $cell+1] $group $butnum $cmd]
									}
								}
							}
						}
					}
				}
			}
		}
		if {$vts == 0} {
			show_msg "Building cross references: FP-PGC..."
			set numcmds [llength $::pgcs($vts,0)]
			set linenum 0
			set section "pre"
			for {set c 1} {$c < $numcmds} {incr c} {
				set cmd [lindex $::pgcs($vts,0) $c]
				if {[lindex $cmd 0] == "*"} {
					set linenum 0
					set section [lindex $cmd 1]
				} else {
					incr linenum
					set r [::main::highlight_jump_cmd pgc $vts,0 [string toupper \[$cmd\]] [expr $c+1] false false]
					add_crossref $r [list "pgcs" $vts 0 0 $section $linenum $cmd]
				}
			}
		} else {
			show_msg "Building cross references: VTST $vts..."
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				add_link_refs "pgc" $vts 0 $pgc

				set numcmds [llength $::pgcs($vts,$pgc)]
				set linenum 0
				set section "pre"
				for {set c 1} {$c < $numcmds} {incr c} {
					set cmd [lindex $::pgcs($vts,$pgc) $c]
					if {[lindex $cmd 0] == "*"} {
						set linenum 0
						set section [lindex $cmd 1]
					} else {
						incr linenum
						set r [::main::highlight_jump_cmd vts $vts,$pgc [string toupper \[$cmd\]] [expr $c+1] false false]
						add_crossref $r [list "pgcs" $vts 0 $pgc $section $linenum $cmd]
					}
				}
			}
		}
		set ::crossrefs_calculated($vts) true
	}
	show_msg ""
}

proc add_crossref {targetlist refslist} {
	if {$targetlist != ""} {
		set targettype		[lindex $targetlist 0]
		set targetvts		[lindex $targetlist 1]
		set targetlu		[lindex $targetlist 2]
		set targetpgc		[lindex $targetlist 3]

		# remove the links to same PGC
		set reftype [lindex $refslist 0]
		if {$reftype == "pgcs"} {
			set reftype pgc
		} else {
			set reftype menu
		}
		if {	$targettype	== $reftype
				&& $targetvts == [lindex $refslist 1]
				&& $targetpgc == [lindex $refslist 3]} {
			return
		}

		if {$targettype == "menu"} {
			if {$targetlu == "all"} {
				for {set tlu 1} {$tlu <= $::menus($targetvts,numlus)} {incr tlu} {
					catch {incr ::crossrefs(menus,$targetvts,$tlu,$targetpgc,numrefs)}
					set ::crossrefs(menus,$targetvts,$tlu,$targetpgc,$::crossrefs(menus,$targetvts,$tlu,$targetpgc,numrefs)) $refslist
				}
			} else {
				catch {incr ::crossrefs(menus,$targetvts,$targetlu,$targetpgc,numrefs)}
				set ::crossrefs(menus,$targetvts,$targetlu,$targetpgc,$::crossrefs(menus,$targetvts,$targetlu,$targetpgc,numrefs)) $refslist
			}
		} else {
			catch {incr ::crossrefs(pgcs,$targetvts,$targetpgc,numrefs)}
			set ::crossrefs(pgcs,$targetvts,$targetpgc,$::crossrefs(pgcs,$targetvts,$targetpgc,numrefs)) $refslist
		}
	}
}
proc add_link_refs {type vts lu pgc} {
	if {$type == "menu"} {
		set header $::menus($vts,$lu,$pgc,header)
	} else {
		set header $::pgcs($vts,$pgc,header)
	}

	set nextpgcn [get_int header 0x9C 2]
	if {$nextpgcn > 0 && $nextpgcn != $pgc} {
		if {$type == "menu"} {
			catch {incr ::crossrefs(menus,$vts,$lu,$nextpgcn,numrefs)}
			set ::crossrefs(menus,$vts,$lu,$nextpgcn,$::crossrefs(menus,$vts,$lu,$nextpgcn,numrefs)) [list "nextpgcn" menu $vts $lu $pgc]
		} else {
			catch {incr ::crossrefs(pgcs,$vts,$nextpgcn,numrefs)}
			set ::crossrefs(pgcs,$vts,$nextpgcn,$::crossrefs(pgcs,$vts,$nextpgcn,numrefs)) [list "nextpgcn" pgc $vts $lu $pgc]
		}
	}

	set prevpgcn [get_int header 0x9E 2]
	if {$prevpgcn > 0 && $prevpgcn != $pgc} {
		if {$type == "menu"} {
			catch {incr ::crossrefs(menus,$vts,$lu,$prevpgcn,numrefs)}
			set ::crossrefs(menus,$vts,$lu,$prevpgcn,$::crossrefs(menus,$vts,$lu,$prevpgcn,numrefs)) [list "prevpgcn" menu $vts $lu $pgc]
		} else {
			catch {incr ::crossrefs(pgcs,$vts,$prevpgcn,numrefs)}
			set ::crossrefs(pgcs,$vts,$prevpgcn,$::crossrefs(pgcs,$vts,$prevpgcn,numrefs)) [list "prevpgcn" pgc $vts $lu $pgc]
		}
	}

	set gouppgcn [get_int header 0xA0 2]
	if {$gouppgcn > 0 && $gouppgcn != $pgc && $gouppgcn != 65535} {
		if {$type == "menu"} {
			catch {incr ::crossrefs(menus,$vts,$lu,$gouppgcn,numrefs)}
			set ::crossrefs(menus,$vts,$lu,$gouppgcn,$::crossrefs(menus,$vts,$lu,$gouppgcn,numrefs)) [list "gouppgcn" menu $vts $lu $pgc]
		} else {
			catch {incr ::crossrefs(pgcs,$vts,$gouppgcn,numrefs)}
			set ::crossrefs(pgcs,$vts,$gouppgcn,$::crossrefs(pgcs,$vts,$gouppgcn,numrefs)) [list "gouppgcn" pgc $vts $lu $pgc]
		}
	}
}
#########################################################
# Show Calls cross references table
proc show_crossrefs {} {
	if {! [info exists ::dvddir]} {
		bell
		return ""
	}

	unset -nocomplain ::crossrefs_calculated
	unset -nocomplain ::crossrefs
	build_crossrefs -1

	set unreferenced ""
	set txt "Calls cross references"
	set txt "$txt\nDVD-TEXT General Name: \"[::utils::get_General_Name]\""
	set txt "$txt\nProvider ID: \"[::utils::get_provider_ID]\""
	set txt "$txt\nNumber of VTS: $::pgcs(numvts)\n"

	if {$::tcl_platform(platform) == "macintosh"} {
		set txt "$txt\nWARNING:  Menu buttons commands not availables for Macintosh, except OS X.\n"
		set txt "$txt\nThe cross references are not complete!\n"
	}

	show_msg "Dumping cross references: FP-PGC"
	set txt "$txt\n################################ First-Play PGC ################################"
	set txt "$txt\n[::main::pgc_label "" pgc 0 0]:"
	set txt "$txt\n    Entry PGC."
	for {set ref 1} {$ref <= $::crossrefs(pgcs,0,0,numrefs)} {incr ref} {
		set txt $txt[display_ref $::crossrefs(pgcs,0,0,$ref)]
	}
	set txt "$txt\n"

	for {set vts 0} {$vts <= $::pgcs(numvts)} {incr vts} {
		if {$vts == 0} {set name "VMGM"} else {set name "VTSM $vts"}
		show_msg "Dumping cross references: $name..."
		set label "[string repeat "#" [expr (78 - [string length $name]) / 2]] $name "
		set label $label[string repeat "#" [expr 80 - [string length $label]]]
		if {$::menus($vts,numlus) > 0} {
			set txt "$txt\n$label"
			for {set lu 1} {$lu <= $::menus($vts,numlus)} {incr lu} {
				for {set pgc 1} {$pgc <= $::menus($vts,$lu,numpgcs)} {incr pgc} {
					set found 0
					set txt "$txt\n[::main::pgc_label "" menu $vts $lu $pgc]:"
					if {$::menus($vts,$lu,$pgc,masktext) != ""} {
						set txt "$txt\n    Callable with remote control $::menus($vts,$lu,$pgc,masktext)enu button."
						set found 1
					}
					if {$::crossrefs(menus,$vts,$lu,$pgc,numrefs) > 0} {
						for {set ref 1} {$ref <= $::crossrefs(menus,$vts,$lu,$pgc,numrefs)} {incr ref} {
							set txt $txt[display_ref $::crossrefs(menus,$vts,$lu,$pgc,$ref)]
						}
						set found 1
					}
					if {! $found} {
						set txt "$txt\n    *** UNREFERENCED! ***"
						set unreferenced "$unreferenced[::main::pgc_label $::config(fixed_width_font) menu $vts $lu $pgc]\n"
						set lab [::main::pgc_label $::config(global_font) menu $vts $lu $pgc]
						.mf.f2.selector.a.lb itemconfigure [lsearch $::main::labelslist $lab] -background #FFE0E0
					}
					set txt "$txt\n"
				}
			}
		}
		if {$vts > 0} {
			show_msg "Dumping cross references: VTST $vts..."
			set label "[string repeat "#" [expr (78 - [string length "VTST $vts"]) / 2]] VTST $vts "
			set label $label[string repeat "#" [expr 80 - [string length $label]]]
			set txt "$txt\n$label"
			for {set pgc 1} {$pgc <= $::pgcs($vts,numpgcs)} {incr pgc} {
				set txt "$txt\n[::main::pgc_label "" pgc $vts $pgc]:"
				if {$::crossrefs(pgcs,$vts,$pgc,numrefs) > 0} {
					for {set ref 1} {$ref <= $::crossrefs(pgcs,$vts,$pgc,numrefs)} {incr ref} {
						set txt $txt[display_ref $::crossrefs(pgcs,$vts,$pgc,$ref)]
					}
				} else {
					set txt "$txt\n    *** UNREFERENCED! ***"
					set unreferenced "$unreferenced[::main::pgc_label $::config(fixed_width_font) pgc $vts $pgc]\n"
					set lab [::main::pgc_label $::config(global_font) pgc $vts $pgc]
					.mf.f2.selector.a.lb itemconfigure [lsearch $::main::labelslist $lab] -background #FFE0E0

				}
				set txt "$txt\n"
			}
		}
	}
	show_msg "Dumping cross references: Unreferenced PGCs..."
	set txt "$txt\n############################### Unreferenced PGCs ##############################"
	if {$unreferenced == ""} {
		set txt "${txt}\nNone."
	} else {
		set txt "$txt\n[string trimright $unreferenced "\n"]"
	}
	::main::showtext .crossrefs "Calls cross references" $txt
	show_msg ""
	return
}
proc display_ref {ref} {
	switch [lindex $ref 0] {
		"menus" {
			set txt "\n    [::main::pgc_label "" menu [lindex $ref 1] [lindex $ref 2] [lindex $ref 3]], [lindex $ref 4] command [lindex $ref 5]:"
			set txt "$txt\n	[::disasm::disasm [lindex $ref 6]]"
		}
		"pgcs" {
			set txt "\n    [::main::pgc_label "" pgc [lindex $ref 1] [lindex $ref 3]], [lindex $ref 4] command [lindex $ref 5]:"
			set txt "$txt\n	[::disasm::disasm [lindex $ref 6]]"
		}
		"menu_buttons" {
			set txt "\n    [::main::pgc_label "" menu [lindex $ref 1] [lindex $ref 2] [lindex $ref 3]], cell [lindex $ref 4], group [lindex $ref 5], button [lindex $ref 6]:"
			set txt "$txt\n	[::disasm::disasm [lindex $ref 7]]"
		}
		"nextpgcn" {
			if {[lindex $ref 1] == "pgc"} {
				set txt "\n    [::main::pgc_label "" [lindex $ref 1] [lindex $ref 2] [lindex $ref 4]]:  NextPGC link."
			} else {
				set txt "\n    [::main::pgc_label "" [lindex $ref 1] [lindex $ref 2] [lindex $ref 3] [lindex $ref 4]]:  NextPGC link."
			}
		}
		"prevpgcn" {
			if {[lindex $ref 1] == "pgc"} {
				set txt "\n    [::main::pgc_label "" [lindex $ref 1] [lindex $ref 2] [lindex $ref 4]]:  PrevPGC link."
			} else {
				set txt "\n    [::main::pgc_label "" [lindex $ref 1] [lindex $ref 2] [lindex $ref 3] [lindex $ref 4]]:  PrevPGC link."
			}
		}
		"gouppgcn" {
			if {[lindex $ref 1] == "pgc"} {
				set txt "\n    [::main::pgc_label "" [lindex $ref 1] [lindex $ref 2] [lindex $ref 4]]:  GoUpPGC link."
			} else {
				set txt "\n    [::main::pgc_label "" [lindex $ref 1] [lindex $ref 2] [lindex $ref 3] [lindex $ref 4]]:  GoUpPGC link."
			}
		}
	}
	return $txt
}

#########################################################
# Jump to crossref
proc pgc_crossrefs {} {
	if {! [info exists ::dvddir]} {
		bell
		return ""
	}

	set type [lindex $::editor_cmds_from 0]
	set indexes [split [lindex $::editor_cmds_from 1] ","]
	set vts [lindex $indexes 0]
	if {$type == "menu"} {
		set lu [lindex $indexes 1]
		set pgc [lindex $indexes 2]
		set returncmd "::main::display_pgc menu $vts $lu $pgc"
	} else {
		set lu 0
		set pgc [lindex $indexes 1]
		set returncmd "::main::display_pgc pgc $vts $pgc"
	}

	set w .pgcxrefs
	if {![winfo exists $w]} {
		unset -nocomplain ::crossrefs_calculated
		unset -nocomplain ::crossrefs

		toplevel $w -class Dialog
		wm withdraw $w
		wm transient $w .
		if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
		wm protocol $w WM_DELETE_WINDOW {set ::utils::pgcxrefs_done 0}
		bind $w <Escape> "set ::utils::pgcxrefs_done 0"
	
		frame  $w.f -borderwidth 2 -relief groove
		pack   $w.f -expand 1 -fill both -padx 4 -pady 3
		canvas $w.f.c -yscrollcommand "$w.f.vs set"
		pack   $w.f.c -side left -expand 1 -fill both
		scrollbar $w.f.vs -command "$w.f.c yview" -orient vert
		pack   $w.f.vs -side left -fill y
	} else {
		set ::config(pgcxrefs_win_geometry) [wm geometry $w] 
		destroy $w.f.c.f1
		label  $w.f.c.tmp -text "Building cross references.\nPlease wait..."
		pack   $w.f.c.tmp
		update idletasks
	}

	if {$type=="pgc"} {
		wm title $w [::main::pgc_label {} $type $vts $pgc]
	} else {
		wm title $w [::main::pgc_label {} $type $vts $lu $pgc]
	}

	frame $w.f.c.f1
	if {$vts == 0} {
		build_crossrefs -1
	} else {
		build_crossrefs $vts
	}

	if {$type=="pgc"} {
		if {$vts==0} {
			pack [label $w.f.c.f1.lab1 -text "Entry PGC."] -anchor w -padx 2
		}
		if {$::crossrefs(pgcs,$vts,$pgc,numrefs)>0} {
			set old [pgc_ref $::crossrefs(pgcs,$vts,$pgc,1)]
			set old [string range $old 0 [string first , $old]]
			for {set ref 1} {$ref <= $::crossrefs(pgcs,$vts,$pgc,numrefs)} {incr ref} {
				set txt [pgc_ref $::crossrefs(pgcs,$vts,$pgc,$ref)]
				if {[string range $txt 0 [string first , $txt]]!=$old} {
					pack [frame $w.f.c.f1.pad$ref -height 6]
					set old [string range $txt 0 [string first , $txt]]
				}
				pack [button $w.f.c.f1.but$ref -text $txt \
					-command "::utils::gotoxref \"$::crossrefs(pgcs,$vts,$pgc,$ref)\"" \
					-justify left -anchor w -pady 0  -background [button_color \
					$::crossrefs(pgcs,$vts,$pgc,$ref)]] -anchor w -fill x -ipadx 2
			}
		} elseif {! [winfo exists $w.f.c.f1.lab1]} {
			pack [label $w.f.c.f1.lab1 -text "*** UNREFERENCED! ***"]
		}
	} else {
		if {$::menus($vts,$lu,$pgc,masktext) != ""} {
			pack [label $w.f.c.f1.lab1 -text \
					"Callable with remote control $::menus($vts,$lu,$pgc,masktext)enu button."] \
					-anchor w -padx 2
		}
		if {$::crossrefs(menus,$vts,$lu,$pgc,numrefs)>0} {
			set old [pgc_ref $::crossrefs(menus,$vts,$lu,$pgc,1)]
			set old [string range $old 0 [string first , $old]]
			for {set ref 1} {$ref <= $::crossrefs(menus,$vts,$lu,$pgc,numrefs)} {incr ref} {
				set txt [pgc_ref $::crossrefs(menus,$vts,$lu,$pgc,$ref)]
				if {[string range $txt 0 [string first , $txt]]!=$old} {
					pack [frame $w.f.c.f1.pad$ref -height 6]
					set old [string range $txt 0 [string first , $txt]]
				}
				pack [button $w.f.c.f1.but$ref -text $txt \
					-command "::utils::gotoxref \"$::crossrefs(menus,$vts,$lu,$pgc,$ref)\"" \
					-justify left -anchor w -pady 0 -background [button_color \
					$::crossrefs(menus,$vts,$lu,$pgc,$ref)]] -anchor w -fill x -ipadx 2
			}
		} elseif {! [winfo exists $w.f.c.f1.lab1]} {
			pack [label $w.f.c.f1.lab1 -text "*** UNREFERENCED! ***"] -padx 2
		}
	}

	pack  $w.f.c.f1 -anchor nw -fill both -expand 1
	$w.f.c create window 0 0 -window $w.f.c.f1 -anchor nw

	# inform scrollbar of the height of the canvas's internal window
	update
	$w.f.c configure -scrollregion "0 0 0 [winfo height $w.f.c.f1]" \
			-width [winfo width $w.f.c.f1]

	# buttons
	if {! [winfo exists $w.fb]} {
		frame  $w.fb
		pack   $w.fb -padx 2 -pady 3 -fill x
		button $w.fb.return -text "Return to target" -command $returncmd
		pack   $w.fb.return -side left -padx 2
		button $w.fb.new -text "New Target" -width 10 -command {
				set ::utils::pgcxrefs_done -1
				after 100 {set ::utils::pause 1}
				tkwait variable ::utils::pause
				::utils::pgc_crossrefs
			}
		pack   $w.fb.new -side left -padx 2
		button $w.fb.refresh -text "Refresh" -width 10 -command {
				set ::utils::pgcxrefs_done -1
				after 100 {set ::utils::pause 1}
				tkwait variable ::utils::pause
				unset -nocomplain ::crossrefs_calculated
				unset -nocomplain ::crossrefs
				::utils::pgc_crossrefs
			}
		pack   $w.fb.refresh -side left -padx 2
		button $w.fb.cancel -text "Close" -width 10 -command {set ::utils::pgcxrefs_done 0}
		pack   $w.fb.cancel -side right -padx 2
	} else {
		catch {destroy $w.f.c.tmp}
		$w.fb.return configure -command $returncmd
	}

	update
	set width [winfo reqwidth $w]
	if {$width < 300} {set width 300}
	wm minsize $w $width [winfo reqheight $w]
	wm maxsize $w $width 2000
	if [info exists ::config(pgcxrefs_win_geometry)] {
		wm geometry $w $::config(pgcxrefs_win_geometry)
	} else {
		set wx [expr [winfo rootx .]+(([winfo width  .]-$width)/2)]
		set wy [expr [winfo rooty .]+(([winfo height .]-[winfo reqheight $w])/2)-25]
		if {$wx < 10} {set wx 10}
		if {$wy < 10} {set wy 10}
		wm geometry $w +$wx+$wy
	}
	wm deiconify $w
	#grab set $w

	tkwait variable ::utils::pgcxrefs_done
	if {$::utils::pgcxrefs_done != -1 && [winfo exists $w]} {
		set ::config(pgcxrefs_win_geometry) [wm geometry $w] 
		destroy $w
	}
	return
}

proc button_color {ref} {
	switch [lindex $ref 0] {
		menus {return #ddeedd}
		pgcs  {return #ddddee}
		menu_buttons {return #eedddd}
		default {return #eeeeee}
	}
}

proc pgc_ref {ref} {
	switch [lindex $ref 0] {
		"menus" {
			set txt "[::main::pgc_label "" menu [lindex $ref 1] [lindex $ref 2] [lindex $ref 3]], [lindex $ref 4] command [lindex $ref 5]"
			set txt "$txt\n[::disasm::disasm [lindex $ref 6]]"
		}
		"pgcs" {
			set txt "[::main::pgc_label "" pgc [lindex $ref 1] [lindex $ref 3]], [lindex $ref 4] command [lindex $ref 5]"
			set txt "$txt\n[::disasm::disasm [lindex $ref 6]]"
		}
		"menu_buttons" {
			set txt "[::main::pgc_label "" menu [lindex $ref 1] [lindex $ref 2] [lindex $ref 3]], cell [lindex $ref 4], group [lindex $ref 5], button [lindex $ref 6]"
			set txt "$txt\n[::disasm::disasm [lindex $ref 7]]"
		}
		"nextpgcn" {
			if {[lindex $ref 1] == "pgc"} {
				set txt "[::main::pgc_label "" [lindex $ref 1] [lindex $ref 2] [lindex $ref 4]],  NextPGC link."
			} else {
				set txt "[::main::pgc_label "" [lindex $ref 1] [lindex $ref 2] [lindex $ref 3] [lindex $ref 4]],  NextPGC link."
			}
		}
		"prevpgcn" {
			if {[lindex $ref 1] == "pgc"} {
				set txt "[::main::pgc_label "" [lindex $ref 1] [lindex $ref 2] [lindex $ref 4]],  PrevPGC link."
			} else {
				set txt "[::main::pgc_label "" [lindex $ref 1] [lindex $ref 2] [lindex $ref 3] [lindex $ref 4]],  PrevPGC link."
			}
		}
		"gouppgcn" {
			if {[lindex $ref 1] == "pgc"} {
				set txt "[::main::pgc_label "" [lindex $ref 1] [lindex $ref 2] [lindex $ref 4]],  GoUpPGC link."
			} else {
				set txt "[::main::pgc_label "" [lindex $ref 1] [lindex $ref 2] [lindex $ref 3] [lindex $ref 4]],  GoUpPGC link."
			}
		}
	}
	return $txt
}

proc gotoxref {ref} {
	#destroy .pgcxrefs
	#set ::utils::pgcxrefs_done 1
	switch [lindex $ref 0] {
		"menus" {
			::main::display_pgc menu [lindex $ref 1] [lindex $ref 2] [lindex $ref 3]
			switch [lindex $ref 4] {
				pre {
					set linenum [lindex $ref 5]
				}
				post {
					set cmds $::menus([lindex $ref 1],[lindex $ref 2],[lindex $ref 3])
					set linenum [expr [num_type_cmds pre cmds] + [lindex $ref 5] +1]
				}
				cell {
					set cmds $::menus([lindex $ref 1],[lindex $ref 2],[lindex $ref 3])
					set linenum [expr [num_type_cmds pre cmds] + [num_type_cmds post cmds] + [lindex $ref 5] +2]
				}
			}
			::main::set_listbox_selection $linenum
		}
		"pgcs" {
			::main::display_pgc pgc [lindex $ref 1] [lindex $ref 3]
			switch [lindex $ref 4] {
				pre {
					set linenum [lindex $ref 5]
				}
				post {
					set cmds $::pgcs([lindex $ref 1],[lindex $ref 3])
					set linenum [expr [num_type_cmds pre cmds] + [lindex $ref 5] +1]
				}
				cell {
					set cmds $::pgcs([lindex $ref 1],[lindex $ref 3])
					set linenum [expr [num_type_cmds pre cmds] + [num_type_cmds post cmds] + [lindex $ref 5] +2]
				}
			}
			::main::set_listbox_selection $linenum
		}
		"menu_buttons" {
			::main::display_pgc menu [lindex $ref 1] [lindex $ref 2] [lindex $ref 3]
			::menubuttons::display_menu_buttons [lindex $ref 1] [lindex $ref 2] [lindex $ref 3] [expr [lindex $ref 6]*1024] info [lindex $ref 4]
		}
		"nextpgcn" -
		"prevpgcn" -
		"gouppgcn" {
			if {[lindex $ref 1] == "pgc"} {
				::main::display_pgc pgc [lindex $ref 2] [lindex $ref 4]
			} else {
				::main::display_pgc menu [lindex $ref 2] [lindex $ref 3] [lindex $ref 4]
			}
			::main::action_Edit_PGC
		}
	}
}



}	;# end of namespace ::utils
